
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : tmhtml.scm
;; DESCRIPTION : conversion of TeXmacs trees into Html trees
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven, David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initialization
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define tmhtml-env (make-ahash-table))
(define tmhtml-css-env (make-ahash-table))

(define (tmhtml-initialize)
  (set! tmhtml-env (tmhtml-env-make)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Manipulation of the environment
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-env-make)
  ;; Create a tmhtml-env.
  (make-ahash-table))

(define (tmhtml-env-list env var)
  (let ((r (ahash-ref env var)))
    (if r r (list '()))))

(define (tmhtml-env-get env var)
  ;; Get the value of a variable in the current tmhtml environment.
  (let ((val (tmhtml-env-list env var)))
    (if (null? (cdr val)) #f
	(car val))))

(define (tmhtml-env-set env var val)
  ;; Set the value of a tmhtml variable, creating a new context.
  ;; This function is undoable by tmhtml-env-reset.
  (ahash-set! env var (cons val (tmhtml-env-list env var))))

(define (tmhtml-env-assign env var val)
  ;; Set the value of a tmhtml variable, modifying the current context.
  (let (l (tmhtml-env-list env var))
    (set-car! l val)
    (ahash-set! env var l)))

(define (tmhtml-env-reset env var)
  ;; Restore a tmhtml variable to it value before the last tmthtml-env-set.
  ;; For a given variable there must be exactly the same number of
  ;; tmhtml-env-reset and tmhtml-env-set.
  (let ((val (tmhtml-env-list env var)))
    (if (not (null? (cdr val)))
	(ahash-set! env var (cdr val)))))

(define-macro (tmhtml-env-with env var val . body)
  ;; Macro to pair syntaxically a tmhtml-env-set and a tmhtml-env-reset.
  (let ((sym (gentemp)))
    `(begin
       (tmhtml-env-set ,env ,var ,val)
       (let ((,sym (begin ,@body)))
	 (tmhtml-env-reset ,env ,var)
	 ,sym))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SXHTML utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (sxhtml-heading? x)
  ;; Is the node x a HTML heading element?
  (and (pair? x) (in? (car x) '(h:h1 h:h2 h:h3 h:h4 h:h5 h:h6))))

(define (sxhtml-list? x)
  ;; Is the node x a HTML list element?
  (and (pair? x) (in? (car x) '(h:ul h:ol h:dl))))

(define (sxhtml-label? x)
  ;; Is the node x a h:a element with an id attribute?
  (and (pair? x)
       (eq? 'h:a (car x))
       (pair? (cadr x))
       (eq? '@ (caadr x))
       (assq 'id (cdadr x))))

(define (sxhtml-glue-label x label)
  ;; Set the id attribute of element x from the id of element label.
  (sxml-set-attr x (sxml-named-attr label 'id)))

(define (sxhtml-table? x)
  ;; Is the node x a h:table element?
  (func? x 'h:table))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SXML utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Fundamental acessors
(define sxml-name car)
(define sxml-attr-list! cdadr)

(define (sxml-has-attr-list? e)
  ;; Has the element e an attribute node?
  (and (pair? (cdr e))
       (pair? (cadr e))
       (eq? '@ (caadr e))))

(define (sxml-element-head e)
  ;; Element name and attributes (if present).
  (if (sxml-has-attr-list? e)
      (list (car e) (cadr e))
      (list (car e))))

(define (sxml-content e)
  ;; Complement function of sxml-element-head.
  (if (sxml-has-attr-list? e) (cddr e) (cdr e)))

(define (sxml-set-name e name)
  ;; Set the name of the sxml element e.
  ;; Name is the new element name as a string.
  (cons (string->symbol name) (cdr e)))

(define (sxml-set-content e content)
  ;; Replace the subnodes of e with the node-set content.
  (append (sxml-element-head e) content))	  
  
(define (sxml-prepend e l)
  ;; Prepend a node set l to the content of an element e.
  (append (sxml-element-head e) l
	  (sxml-content e)))

(define (sxml-attr-list? e)
  ;; Return the attribute list of element e or #f.
  (and (sxml-has-attr-list? e)
       (sxml-attr-list! e)))

(define (sxml-attr-list e)
  ;; Return the attribute list of element e or empty list.
  (or (sxml-attr-list? e) '()))

(define (sxml-named-attr obj attr-name)
  ;; Named attribute of element e or #f.
  (and-let* ((l (sxml-attr-list? obj)))
	    (assq attr-name l)))

(define (sxml-attr obj attr-name)
  ;; Value of a named attribute of element e or #f.
  (and-let* ((x (sxml-named-attr obj attr-name)))
	    (cadr x)))

(define (sxml-set-attr e attr)
  ;; Set an attribute of an element e. Attr is a list (symbol? string?).
  ;; Create the attribute list or the attribute if necessary.
  (let ((attr-name (car attr)))
    `(,(sxml-name e)
      (@ ,attr ,@(list-filter (sxml-attr-list e)
			      (lambda (x) (not (eq? x attr-name)))))
      ,@(sxml-content e))))

(define (sxml-set-attrs e attrs)
  ;; Set several attributes of an element e. Attrs is a list of attributes.
  ;; Create the attribute list or attributes if necessary.
  (let rec ((e e) (attrs attrs))
    (if (null? attrs) e
	(rec (sxml-set-attr e (car attrs)) (cdr attrs)))))

(define-macro (sxml-find-name-separator len)
  ;; optimized (string-rindex name #\:)
  ;; returns position of a separator between namespace-id and LocalName
  ;; (copied from sxml-tools)
  `(let rpt ((pos (1- ,len))) 
     (cond
       ((negative? pos) #f) 	
       ((char=? #\: (string-ref name pos)) pos)
       (else (rpt (1- pos))))))

(define (sxml-ncname obj)
  ;; Returns Local Part of Qualified Name (Namespaces in XML production [6])
  ;; for given obj, which is ":"-separated suffix of its Qualified Name
  ;; If a name of a node given is NCName (Namespaces in XML production [4]),
  ;; then it is returned as is.
  ;; Please note that while SXML name is a symbol this function returns a
  ;; string.
  ;; (copied from sxml-tools)
  (let* ((name (symbol->string (car obj)))
	 (len (string-length name)))
    (cond
      ((sxml-find-name-separator len)
       => (lambda (pos) 
	    (substring name (+ pos 1) len)))
      (else name))))

(define (sxml-name->ns-id sxml-name)
  ;; Returns namespace-id part of given name, or #f if it's LocalName
  ;; (copied from sxml-tools)
  (let* ((name (symbol->string sxml-name)))
    (cond
      ((sxml-find-name-separator (string-length name))
       => (lambda (pos) 
	    (substring name  0 pos)))
      (else #f))))

(define (sxml-strip-ns-prefix prefix x)
  ;; Remove a given namespace prefix wherever it appears in element names.
  ;; Prefix must be the ns-prefix as a string, and x a document fragment.
  (let rec ((x x))
    (if (string? x) x
	(sxml-set-content
	 (if (== prefix (sxml-name->ns-id (sxml-name x)))
	     (sxml-set-name x (sxml-ncname x))
	     x)
	 (map rec (sxml-content x))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; TeXmacs scheme tree utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-first-data doc-item)
  ;; Given a doc-item, return its first data node. That is the first node found
  ;; by preorder traversal which is neither a 'concat' nor a 'paragraph'.
  ;;
  ;; If traversal finds an empty paragraph-item, return "".
  ;;
  ;; Assume the doc-item is partially concat and paragraph simplified.
  (let rec ((x doc-item))
    (if (string? x) x
	(if (or (eq? 'concat (car x))
		(eq? 'paragraph (car x)))
	    (rec (cadr x))	      ; assume at least one para/concat-item
	    x))))

(define (tmhtml-remove-first-data doc-item)
  ;; Given a doc-item, remove its first data node. Simplify concats and
  ;; preserve paragraph structure. If the first data node is an empty
  ;; paragraph-item, do nothing.
  ;;
  ;; Assume the doc-item is partially concat and paragraph simplified.
  (let rec ((x doc-item))
    (cond ((string? x) "")
	  ((eq? 'paragraph (car x))	; assume at least one para-item
	   `(paragraph ,(rec (cadr x)) ,@(cddr x)))
	  ((eq? 'concat (car x))	; assume at least one concat-item
	   (if (null? (cddr x)) ""	; simplify null concat
	       `(concat ,@(cddr x))))
	  (else ""))))

(define (tmhtml-list-map proc pred doc-body)
  ;; General mapping over LaTeX-style lists
  ;; proc: mark item -> x. Mapping function
  ;; pred: doc-item -> ?. Is the doc-item an list item start?
  ;; doc-body: List body.
  ;;
  ;; Each item is the list of doc-items between the start of the doc-body,
  ;; doc-items whose first data is a marker and the end of the doc-body. The
  ;; proc is applied with the marker of the current item as first parameter,
  ;; and the item data as second parameter. If the first doc-item does not
  ;; start by a mark, the first application is done with mark=#f.
  (define (marked? x) (pred (tmhtml-first-data x)))
  (if (null? doc-body) '()
      (let rec ((doc-body (cdr doc-body))
		(mark (if (marked? (car doc-body))
			  (tmhtml-first-data (car doc-body))
			  #f))
		(accum (list (if (marked? (car doc-body))
				 (tmhtml-remove-first-data (car doc-body))
				 (car doc-body)))))
	(cond ((null? doc-body)
	       (list (proc mark (reverse accum))))
	      ((marked? (car doc-body))
	       (cons (proc mark (reverse accum))
		     (rec (cdr doc-body)
			  (tmhtml-first-data (car doc-body))
			  (list (tmhtml-remove-first-data (car doc-body))))))
	      (else (rec (cdr doc-body) mark
			 (cons (car doc-body) accum)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Postprocessing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-post-simplify-nodes l)
  ;; Catenate adjacent string nodes and remove empty string nodes
  (let rec ((l l))
    (cond ((null? l) '())
	  ((and (string? (car l)) (string-null? (car l)))
	   (rec (cdr l)))
	  ((null? (cdr l)) l)
	  ((and (string? (car l)) (string? (cadr l)))
	   (rec (cons (string-append (car l) (cadr l)) (cddr l))))
	  (else (cons (car l) (rec (cdr l)))))))

(define (tmhtml-post-simplify-element e)
  ;; Simplify the nodes of the element content
  (list (append (sxml-element-head e)
		(tmhtml-post-simplify-nodes (sxml-content e)))))

(define (tmhtml-post-heading l)
  ;; Post-process the converted result of a concat containing a section title.
  ;;
  ;; The first label after the section is changed to an 'id' attribute in the
  ;; heading element, if it has not already an 'id' attribute.
  ;;
  ;; NOTE: assumes the heading is the first node of the set.
  (let ((heading (car l)))
    (if (sxml-attr heading 'id) l
	(receive (labels rest) (list-partition (cdr l) sxhtml-label?)
	  (if (null? labels) l
	      (cons (sxml-prepend (sxhtml-glue-label heading (car labels))
				  (cdr labels))
		    rest))))))

(define (tmhtml-post-table l)
  ;; Post process the converted result of a concat containing a table.
  ;;
  ;; If a label is adjacent to the table, use it to set the table id. If there
  ;; are several labels adjacent to the table, leave all but one label
  ;; untouched. There is no guarantee on which label is glued.
  (define (glue-label-to-table x knil)
    (cond ((null? knil) (list x))
	  ((and (sxhtml-label? x)
		(sxhtml-table? (car knil))
		(not (sxml-attr (car knil) 'id)))
	   (cons (sxhtml-glue-label (car knil) x)
		 (cdr knil)))
	  ((and (sxhtml-table? x)
		(not (sxml-attr x 'id))
		(sxhtml-label? (car knil)))
	   (cons (sxhtml-glue-label x (car knil))
		 (cdr knil)))
	  (else (cons x knil))))
  (list-fold-right glue-label-to-table '() l))

(define (tmhtml-post-paragraphs l)
  ;; Post process a collection of h:p elements
  ;;
  ;; If a h:p contains a h:hN, remove the h:p node and prepend the rest of the
  ;; contents to the next h:p. If the next element, after post processing is
  ;; not a h:p, create an intermediate h:p to hold the data.
  ;;
  ;; If a h:p contains a list element, remove the enclosing h:p. The TeXmacs
  ;; editor ensures that an <item-list> or <desc-list> is the only element
  ;; contained in its enclosing <doc-item>.
  ;;
  ;; NOTE: assumes the heading is at the start of a paragraph. That is
  ;; consistent with the fact that (as of 2003-02-04) the only converted
  ;; invisible markup is <label> and correct usage requires it to be after the
  ;; section title.
  (let rec ((in l) (out '()) (trail #f))
    (let* ((para (and (pair? in) (car in)))
	   (cont (and para (sxml-content para)))
	   (first (and cont (pair? cont) (car cont)))
	   (next (lambda (o t) (rec (cdr in) o t)))
	   (flush (lambda () (if trail `((h:p ,@trail) ,@out) out)))
	   (accept (lambda () (if trail (sxml-prepend para trail) para)))
	   (give (lambda () (and (pair? (cdr cont)) (cdr cont)))))
      ;; invariant: (xor prev prev-trail)
      (cond ((null? in) (reverse (flush)))
	    ((or (null? cont) (string? first))
	     (next (cons (accept) out) #f))
	    ((sxhtml-heading? first)
	     ;; tmhtml-post-heading should be called by concat handler
	     (next (cons first (flush)) (give)))
	    ((sxhtml-list? first)
	     ;; texmacs editor ensures there is not trail after a list
	     (next (append cont (flush)) #f))
	    ((and (sxhtml-table? first) (null? (cdr cont)))
	     ;; if table is not alone, we cannot help but produce bad html
	     ;; if table is alone, drop the enclosing <h:p>
	     (next (append cont (flush)) #f))
	    (else (next (cons (accept) out) #f))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Strings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-string s) s)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Simple text
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-noop l) '())

(define (tmhtml-file l)
  (let* ((doc (car l))
	 (styles (cdadr l))
	 (lang (caddr l))
	 (tmpath (cadddr l)))
    `(h:html (h:body (@ ("align" "justify"))
		     ,@(tmhtml doc)))))

(define (tmhtml-document l)
  (tmhtml-post-paragraphs (map (lambda (x) (cons 'h:p (tmhtml x)))
			       l)))

(define (tmhtml-paragraph l)
  (let rec ((l l))
    (if (null? l) '()
	(let ((h (tmhtml (car l)))
	      (r (rec (cdr l))))
	  (cond ((null? h) r)		; correct when r is null too
		((null? r) h)
		(else `(,@h (h:br) ,@r)))))))

(define (tmhtml-concat l)
  (tmhtml-post-simplify-nodes
   (let ((l (tmhtml-list l)))
     (cond ((null? l) '())
	   ((string? (car l)) l)
	   ((sxhtml-heading? (car l)) (tmhtml-post-heading l))
	   ((list-any sxhtml-table? l) (tmhtml-post-table l))
	   (else l)))))

(define (tmhtml-format l)
  (let ((s (car l)))
    (cond ((== s "new line") '((h:br)))
	  ((== s "next line") '((h:br)))
	  (else '()))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Local and global environment changes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(drd-table tmhtml-with-cmd%
  (("font family" "tt") (h:tt))
  (("font series" "bold") (h:b))
  (("font shape" "italic") (h:i)))

(drd-table tmhtml-with-cmd% ; deprecated
  ("font size" tmhtml-with-font-size)
  (("color" "black") (h:font (@ ("color" "black"))))
  (("color" "grey") (h:font (@ ("color" "grey"))))
  (("color" "white") (h:font (@ ("color" "white"))))
  (("color" "red") (h:font (@ ("color" "red"))))
  (("color" "blue") (h:font (@ ("color" "blue"))))
  (("color" "yellow") (h:font (@ ("color" "black"))))
  (("color" "magenta") (h:font (@ ("color" "magenta"))))
  (("color" "orange") (h:font (@ ("color" "orange"))))
  (("color" "green") (h:font (@ ("color" "green"))))
  (("color" "brown") (h:font (@ ("color" "brown"))))
  (("paragraph mode" "left") (h:div (@ ("align" "left"))))
  (("paragraph mode" "justify") (h:div (@ ("align" "justify"))))
  (("paragraph mode" "center") (h:center)))

(drd-table tmhtml-with-cmd% ; netscape4
  (("nr columns" "1") (h:multicol (@ ("cols" "1"))))
  (("nr columns" "2") (h:multicol (@ ("cols" "2"))))
  (("nr columns" "3") (h:multicol (@ ("cols" "3"))))
  (("nr columns" "4") (h:multicol (@ ("cols" "4"))))
  (("nr columns" "5") (h:multicol (@ ("cols" "5")))))

(define (tmhtml-with-font-size val)
  (let* ((x (* (string->number val) 100))
	 (s (cond ((< x 1) #f) ((< x 55) "-4") ((< x 65) "-3")
		  ((< x 75) "-2") ((< x 95) "-1") ((< x 115) "0")
		  ((< x 135) "+1") ((< x 155) "+2") ((< x 185) "+3")
		  ((< x 225) "+4") ((< x 500) "+5") (else #f))))
    (and s `(h:font (@ (size ,s))))))

(define (tmhtml-with-one var val arg-result)
  ;; Internal routine for tmhtml-with.
  ;;
  ;; Convert a with-binding. Val is the value being bound to var. Arg-result is
  ;; the body of the with element (including subsequent bindings) after
  ;; conversion.
  (let ((w (or (drd-ref tmhtml-with-cmd% (list var val))
	       (let ((h (drd-ref tmhtml-with-cmd% var)))
		 (and h ((eval h) val))))))
    (if w
	(list (append w arg-result))
	arg-result)))

(define (tmhtml-with l)
  ;; Handler for 'with' primitive.
  ;;
  ;; Apply tmhtml-with-one for each binding in the node, and update the
  ;; environment. Earlier bindings will produce outer nodes.
  (let rec ((l l))
    ;; sanity check, only true if 'with' has an even number of
    ;; parameters, which should never happen.
    (if (null? l) '()
	(if (null? (cdr l))
	    (tmhtml (car l)) ;; all binding have been handled, convert body
	    ;; WARNING: bad conversion if variable or value is not a string.
	    (let* ((var (force-string (car l)))
		   (val (force-string (cadr l)))
		   (next (cddr l)))
	      (tmhtml-env-with tmhtml-env var val
	         (tmhtml-with-one var val (rec next))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Other primitives
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-label l)
  ;; WARNING: bad conversion if ID is not a string.
  `((h:a (@ (id ,(force-string (car l)))))))

;(define (tmhtml-reference l)
;  (list 'ref (force-string (car l))))

;(define (tmhtml-pageref l)
;  (list 'pageref (force-string (car l))))

(define (tmhtml-hyperlink l)
  ;; WARNING: bad conversion if URI is not a string.
  `((h:a (@ (href ,(force-string (car l))))
	 ,@(tmhtml-list (cdr l)))))

(define (tmhtml-specific l)
  (if (== (car l) "html") (tmhtml (cadr l)) '()))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Standard markup
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(drd-table tmhtml-stdmarkup%
  ;;; Sectioning
  (chapter (h:h1))
  (section (h:h2))
  (subsection (h:h3))
  (subsubsection (h:h4))
  ;; paragraph and subparagraph are intented to be used at the start
  ;; of a paragraph. So they cannot be converted to 'h5' and 'h6
  (paragraph (h:strong (@(class "paragraph"))))
  (subparagraph (h:strong (@(class "subparagraph"))))
  ;; Lists
  ((:or itemize itemize-minus itemize-dot itemize-arrow) tmhtml-itemize)
  ((:or enumerate enumerate-numeric enumerate-roman enumerate-Roman
	enumerate-alpha enumerate-Alpha)
   tmhtml-enumerate)
  ((:or description description-compact description-dash
	description-align description-long)
   tmhtml-description)
  ;; Phrase elements
  (strong (h:strong))
  (em (h:em))
  (dfn (h:dfn))
  (code* (h:code))
  (samp (h:samp)) ; WARNING: semantic documentation does not match HTML4
  (kbd (h:kbd))
  (var (h:var))
  (abbr (h:abbr))
  (acronym (h:acronym))
  ;; Presentation
  (tt (h:tt)))

;;    (name (h:name)) ; not in HTML4
;;    (person (h:person)))) ; not in HTML4

(define (tmhtml-list-document list-doc)
  ;; Convert a list-document to an list of <h:li> elements.
  (define (item->li mark item)
    (cond ((null? item) '(h:li))
	  ((null? (cdr item)) `(h:li ,@(tmhtml (car item))))
	  (else `(h:li ,@(tmhtml `(document ,@item))))))
  (if (null? (cdr list-doc)) '((h:li))
      (tmhtml-list-map item->li
		       (lambda (x) (== x '(item)))
		       (cdr list-doc))))

(define (tmhtml-itemize args)
  `((h:ul ,@(tmhtml-list-document (car args)))))

(define (tmhtml-enumerate args)
  `((h:ol ,@(tmhtml-list-document (car args)))))

(define (tmhtml-desc-document desc-doc)
  (define (item->dt-dd mark item)
    (let ((html-item (if (null? (cdr item))
			 (tmhtml (car item))
			 (tmhtml `(document ,@item)))))
      (append
       (if mark `((h:dt ,@(tmhtml (cadr mark)))) '())
       (cond ((and (null? html-item) mark) '())
	     ((null? html-item) '((h:dd)))
	     (else `((h:dd ,@html-item)))))))
  (if (null? (cdr desc-doc)) '((h:dd))
      (apply append (tmhtml-list-map item->dt-dd
				     (lambda (x) (func? x 'item* 1))
				     (cdr desc-doc)))))

(define (tmhtml-description args)
  `((h:dl ,@(tmhtml-desc-document (car args)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Tables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(drd-table tmhtml-tables%
  (block ,(lambda (l) (tmhtml-table "l" #t l)))
  (block* ,(lambda (l) (tmhtml-table "c" #t l)))
  (tabular ,(lambda (l) (tmhtml-table "l" #f l)))
  (tabular* ,(lambda (l) (tmhtml-table "c" #f l))))

(define (tmhtml-table-cols p)
  (map (lambda (x)
	 `(h:col (@ (align ,(cadr (assoc x '(("l" "left")
					     ("c" "center")
					     ("r" "right"))))))))
       (p 'cols 'halign)))

(define (tmhtml-table-contents p)
  (define (cell x)
    `(h:td ,@(tmhtml x)))
  (define (row l)
    `(h:tr ,@(map cell l)))
  (map row (p 'rows 'content)))

(define (tmhtml-table-make p)
  `((h:table
     ,@(if (p 'global 'border) '((@ (border "1"))) '())
     ,@(tmhtml-table-cols p)
     (h:tbody ,@(tmhtml-table-contents p)))))

;; WARNING: use tmtex table parsing, which should be moved to a common module.
(define (tmhtml-table halign border x)
  ;; assert (= 1 (length x))
  ;; assert (or (func? (car x) 'tformat) (func? (car x) 'table))
  (tmhtml-table-make
   (tmtable-parser `(tformat ,@(tmtable-cell-halign halign)
			     ,@(tmtable-block-borders border)
			     ,(car x)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pictures
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; WARNING: Should also test that width and height are not magnifications.
;; Currently, magnifications make string->tmlength return #f.

(define (tmhtml-dimension-attr s name handlers)
  (let ((w (and-let* ((tmlen (string->tmlength s))
		      ((not (tmlength-null? tmlen)))
		      (proc (assoc (tmlength-unit tmlen) handlers)))
		     ((second proc) (tmlength-value tmlen)))))
    (if w `((,name ,w)) '())))

(define (exact-number->string x)
  (number->string (inexact->exact x)))
(define (number->percent x)
  (string-append (exact-number->string (* 100 x)) "%"))

(define (tmhtml-postscript l)
  (let ((s (first l)) (w (second l)) (h (third l)))
    (if (not (string? s)) '()	; only convert linked images
	`((h:img (@ (src ,s)
		    ,@(tmhtml-dimension-attr
		       w 'width `((par ,number->percent)
				  (px ,exact-number->string)))
		    ,@(tmhtml-dimension-attr
		       h 'height `((px ,exact-number->string)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main conversion routines
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-list l)
  (append-map tmhtml l))

(define (tmhtml-dispatch htable l)
  (let ((x (drd-ref ,htable (car l))))
    (cond ((not x) #f)
	  ((symbol? x) ((eval x) (cdr l)))
	  ((procedure? x) (x (cdr l)))
	  (else (tmhtml-post-simplify-element
		 (append x (tmhtml-list (cdr l))))))))

(define (tmhtml-expand l)
  ;; Explicit expansions are converted and handled as implicit expansions.
  (tmhtml-implicit-expand (cons (string->symbol (car l))
				(cdr l))))

(define (tmhtml-implicit-expand l)
  (or (tmhtml-dispatch 'tmhtml-stdmarkup% l)
      (tmhtml-dispatch 'tmhtml-tables% l)
      '()))

(define (tmhtml x)
  ;; Main conversion function.
  ;; Takes a TeXmacs tree in Scheme notation and produce a SXML node-set.
  ;; All handler functions have a similar prototype.
  (if (string? x)
      (if (string-null? x) '()
	  (list (tmhtml-string x)))		; handle string nodes
      (or (tmhtml-dispatch 'tmhtml-primitives% x)
	  (tmhtml-implicit-expand x))))

(define (tmhtml-decorate root)
  "Add metadata to a XML document to prepare it for serialization"
  `(*TOP*
    ;; x-texmacs is internal encoding of TeXmacs.
    ;; It is Cork encoding with TeXmacs universal symbols.
    (*PI* xml "version=\"1.0\" encoding=\"x-texmacs\"")
    ,(sxml-set-attrs
      root '((xmlns:h "http://www.w3.org/1999/xhtml")
	     (xmlns:m "http://www.w3.org/1998/Math/MathML")
	     (xmlns:x "http://www.texmacs.org/2002/extensions")))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Dispatching
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(drd-dispatcher tmhtml-primitives%
  (document tmhtml-document)
  (paragraph tmhtml-paragraph)
  (surround tmhtml-noop)
  (concat tmhtml-concat)
  (format tmhtml-format)
  ((:or hspace vspace* vspace space htab split move resize float
	repeat datoms dlines dpages dbox)
   tmhtml-noop)
  ((:or group left mid right big lprime rprime below above
	lsub lsup rsub rsup frac sqrt wide neg
	tree old_matrix old_table old_mosaic old_mosaic_item)
   tmhtml-noop)
  (table (lambda (l) (tmhtml `(tabular (table ,@l)))))
  (tformat (lambda (l) (tmhtml `(tabular (tformat ,@l)))))
  ((:or twith cwith tmarker row cell sub_table) tmhtml-noop)

  (assign tmhtml-noop)
  (with tmhtml-with)
  ((:or set reset) tmhtml-noop)
  ((:or var_expand expand apply) tmhtml-expand)
  ((:or begin end include macro func env eval) tmhtml-noop)
  (value tmhtml-expand)
  (arg tmhtml-noop)
  ((:or backup quote delay hold release) tmhtml-noop)
  
  ((:or or xor and not plus minus times over div mod merge length range
	number date translate is_tuple look_up equal unequal less lesseq
	greater greatereq if case while extern authorize)
   tmhtml-noop)
  ((:or inactive symbol latex hybrid tuple collection associate) tmhtml-noop)
  (label tmhtml-label)
  (reference tmhtml-noop)
  (pageref tmhtml-noop)
  (write tmhtml-noop)
  (specific tmhtml-specific)
  (hlink tmhtml-hyperlink)
  ((:or action tag meaning) tmhtml-noop)
  ((:or switch fold exclusive progressive superposed) tmhtml-noop)
  ((:or graphics point line arc bezier) tmhtml-noop)
  (postscript tmhtml-postscript)

  (!file tmhtml-file))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-produce x)
  ;; Convert a TeXmacs tree x to a HTML tree.
  ;;
  ;; If we are converting a document, the tree x is:
  ;;   (file! <doc> <style> <lan> <tmpath>)
  ;; with:
  ;;   <doc>: edit tree of the document being converted.
  ;;   <style>: TUPLE of style and package names used by the document. If
  ;;     there is only a style (or no style at all) a TUPLE is introduced by
  ;;     the caller.
  ;;   <lan>: the language of the document (from init env).
  ;;   <tmpath>: TEXMACS_PATH with trailing slashes removed.
  ;;
  ;; If we are converting a simple snippet (exporting the clipboard) the tree x
  ;; is simply the snippet being exported.
  (tmhtml-initialize)
  (htmlout-produce (sxml-strip-ns-prefix "h" (tmhtml x))))
