
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : tmtable.scm
;; DESCRIPTION : tools for converting tables from and to other formats
;; COPYRIGHT   : (C) 2002  David Allouche, Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Generic table parser
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmtable-parser x)
  ;; Internal state
  (let* ((contents			; table data
	  (let rec ((x x))
	    (if (func? x 'tformat) (rec (cAr x))
		;; assert (func? x 'table)
		(map (lambda (row) (map cadr (cdr row)))
		     (cdr x)))))
	 (nrcols			; number of columns
	  (length (car contents)))
	 (cols-halign			; column-wide horizontal alignment 
	  (make-vector nrcols "l"))
	 (global-border #f))		; all cell borders visible?
  
    ;; Public getters
    (define (global name)
      (cond ((eq? name 'border) global-border)
	    ((eq? name 'nrcols) nrcols)))
    (define (cols name)
      (if (eq? name 'halign) (vector->list cols-halign)))
    (define (rows name)
      (if (eq? name 'content) contents))
  
    ;; Public dispatcher
    (define (table-parser/dispatch scope . args)
      (apply (cond ((eq? scope 'global) global)
		   ((eq? scope 'cols) cols)
		   ((eq? scope 'rows) rows)) args))
  
    ;; Private
    (define (set-cols-halign! j1 j2 val)
      ;; Set the horizontal alignement property for a range of table rows.
      (if (<= j1 j2)
	  (begin
	    (vector-set! cols-halign j1 val)
	    (set-cols-halign! (1+ j1) j2 val))))
  
    (define (parse-tformat-item x)
      ;; Modify the internal state according to a table format entry. 
      (if (and (func? x 'cwith 6) (== (cadr x) "1") (== (caddr x) "-1"))
	  (let* ((J1 (string->number (cadr (cddr x))))
		 (J2 (string->number (caddr (cddr x))))
		 (n nrcols)
		 (j1 (cond ((< J1 0) (+ J1 n)) ((> J1 0) (- J1 1)) (else 0)))
		 (j2 (cond ((< J2 0) (+ J2 n)) ((> J2 0) (- J2 1)) (else 0)))
		 (var (cadr (cddddr x)))
		 (val (caddr (cddddr x))))
	    (cond ((== var "cell halign")
		   (if (in? val '("l" "c" "r"))
		       (set-cols-halign! (max 0 j1) (min j2 (- n 1)) val)))
		  ((in? var '("cell lborder" "cell rborder"
			      "cell tborder" "cell bborder"))
		   (set! global-border #t))))))

    ;; (trace-variables contents nrcols cols-halign global-border)
    ;; (set-trace-level! global cols rows)
    ;; (set-trace-level! set-cols-halign! parse-tformat-item)
    ;; (set-trace-level! table-parser/dispatch)

    (do ((x x (cAr x)))
	((func? x 'table))
      (for-each parse-tformat-item (cDdr x)))

    ;; evaluate to dispatcher closure
    table-parser/dispatch))

(define (tmtable-block-borders x)
  (if x '((cwith "1" "-1" "1" "1" "cell lborder" "1ln")
	  (cwith "1" "1" "1" "-1" "cell tborder" "1ln")
	  (cwith "1" "-1" "1" "-1" "cell bborder" "1ln")
	  (cwith "1" "-1" "1" "-1" "cell rborder" "1ln"))
      '()))

(define (tmtable-cell-halign x)
  `((cwith "1" "-1" "1" "-1" "cell halign" ,x)))
