
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : debug.scm
;; DESCRIPTION : debugging tools
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven, David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Output
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (display* . l)
  "Display all objects in @l."
  (for-each display l))

(define (display-err x)
  "Display @x to the error port."
  (display x (current-error-port)))

(define (display-err* . l)
  "Display all objects in @l to the error port."
  (for-each display-err l))

(define (tm-display-error . l)
  (apply display-err* `("TeXmacs] " ,@l "\n")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Assertions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (check-arg-type pred arg caller)
  (if (pred arg) arg
      (scm-error 'wrong-type-arg caller
		 "Wrong type argument: ~S" (list arg) '())))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Safe evaluation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; NOTE: there should be a more explicit error reporting for users 

(define (apply-with lst lam)
  ;; Syntactic sugar for multiple return values as a list."
  (apply lam lst))

(define (safe-apply action args)
  (catch #t (lambda ()
	      (let ((v (apply action args)))
		(list #f v)))
	 (lambda args (list #t #f))))

(define (safe-apply-error action args)
  (catch #t
	 (lambda ()
	   (let ((v (apply action args)))
	     (list #f v)))
	 (lambda args
	   (tm-display-error "Guile error in " action)
	   (list #t #f))))

(define (safe-eval-list lst)
  (catch #t (lambda ()
	      (let ((v (eval lst)))
		(if (list? v)
		    (list #f v)
		    (list #t v))))
	 (lambda args (list #t #f))))

(define (safe-eval-bool lst)
  (catch #t (lambda ()
	      (let ((v (eval lst)))
		(if (boolean? v)
		    (list #f v)
		    (list #t v))))
	 (lambda args (list #t #f))))
  
(define (eval-bool-or-false lst)
  (apply-with (safe-eval-bool lst)
	      (lambda (error? bool)
		(and (not error?) bool))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Regression testing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (regression-test-equal group test res-in result exp-in expected)
  (if (not (equal? result expected))
      (begin
	(newline)
	(display* "Expected in - " (object->string exp-in) "\n")
	(display* "Result   in - " (object->string res-in) "\n")
	(display* "Expected - " (object->string expected) "\n")
	(display* "Result   - " (object->string result) "\n")
	(display* "Do not match!\n")
	(display* "Regression failure: " group " / " test)
	(error "Regression failure:" group test))))

(define (regression-test-nequal group test res-in result exp-in expected)
  (if (equal? result expected)
      (begin
	(newline)
	(display* "Expected Not in - " (object->string exp-in) "\n")
	(display* "Result       in - " (object->string res-in) "\n")
	(display* "Expected Not - " (object->string expected) "\n")
	(display* "Result       - " (object->string result) "\n")
	(display* "Unwanted match! Regression failure!\n")
	(display* "Regression failure: " group " / " test)
	(error "Regression failure:" group test))))

(define-macro (regression-test-group group-desc group-id
				     result-cmd expected-cmd . body)
  (let* ((make-command (lambda (cmd)
			 (if (equal? cmd ':none)
			     (lambda (x) x)
			     (lambda (x) (list cmd x)))))
	 (make-result (make-command result-cmd))
	 (make-expected (make-command expected-cmd))
	 (tests
	  (let rec ((n 1) (l body))	; process body items
	    (define (make-test e?)
	      (let* ((t (car l))
		     (test-desc (cadr t))
		     (result-in (caddr t))
		     (expected-in (cadddr t))
		     (result (make-result result-in))
		     (expected (make-expected expected-in)))
		;; Display messages and run test.
		`((display ,(string-append "  -- " test-desc "\n"))
		  (,(if e? 'regression-test-equal 'regression-test-nequal)
		   ,group-id ,test-desc
		   ,result-in ,result ,expected-in ,expected)
		  ,@(rec (1+ n) (cdr l))))) ; rest of the body
	    (cond ((null? l) `(,(1- n))) ; evaluate to number of tests
		  ;; Improper list or unexpect atom. Nevermind.
		  ((not (pair? l)) l)
		  ((not (pair? (car l)))
		   (cons (car l) (rec n (cdr l))))
		  ;; Test case.
		  ((equal? 'test (caar l)) (make-test #t))
		  ((equal? 'test-fails (caar l)) (make-test #f))
		  ;; Non-test form, preserve.
		  (else (cons (car l) (rec n (cdr l))))))))
    `(begin
       (display ,(string-append "Test group: " group-desc " [" group-id "]\n"))
       ,@tests)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Test suite library
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-macro (regtest-table-library)
  ;; basic shorthands for input of texmacs tables
  `(begin
     (define (cell x) `(cell ,x))
     (define (row l) `(row ,@(map cell l)))
     (define (table ll) `(table ,@(map row ll)))
     (define (tformat pp ll) `(tformat ,@pp ,(table ll)))
     (define (colwith i var val) `(cwith "1" "-1" ,i ,i ,var ,val))
     (define (rowwith i var val) `(cwith ,i ,i "1" "-1" ,var ,val))
     (define (allwith var val) `(cwith "1" "-1" "1" "-1" ,var ,val))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Debugging
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define tm-last-backtrace #f)

(define (backtrace-dumper key . args)
  ;; Internal function for backtrack
  (let ((stack (make-stack #t backtrace-dumper)))
    (set! tm-last-backtrace stack)
    (display-backtrace stack (current-error-port))
    (apply display-error stack (current-error-port) args)
    (apply throw (cons key args))))

(define-macro (backtrack . body)
  ;; Evaluate body and print a full stack backtrace if an error occurs.
  ;; Copied from:
  ;;   http://mail.gnu.org/archive/html/guile-user/2001-01/msg00126.html
  ;; For this to work, the user init file should start with:
  ;;   (debug-enable 'backtrace 'debug)
  ;; NOTE: this is a hack, the whole error reporting system of TeXmacs should
  ;;   be redesigned (well... not to say "designed to start with").
  ;; WARNING: has a problem coexisting with wrap-trace-level
  `(lazy-catch #t (lambda ()  ,@body) backtrace-dumper))

(define (wrap-catch proc)
  ;; Wrap a procedure in a closure which displays and passes exceptions.
  (lambda args
    (catch #t
	   (lambda () (apply proc args))
	   (lambda err
	     (tm-display-error "Guile error: " (list err))
	     (apply throw err)))))

(define (wrap-catch-list expr)
  ;; Similar to wrap-catch for a scheme expression in list form.
  `(catch #t
	  (lambda () ,expr)
	  (lambda err
	    (tm-display-error "Guile error: " (list err))
	    (apply throw err))))

(define trace-level 0)
(define (trace-indent)
  ;; Produce the string to be used to indent trace output.
  (let rec ((n trace-level) (s '()))
    (if (equal? 0 n) (apply string-append s)
	(rec (1- n) (cons "| " s)))))

(define (trace-display . args)
  ;; As display but also print trace indentation.
  (display (trace-indent))
  (for-each (lambda (a)
	      (display (if (string? a) a (object->string a)))
	      (display " "))
	    args)
  (newline))

(define-macro (trace-variables . vars)
  ;; Use trace-display to show the name and value of some variables.
  (define (trace-one-variable v)
    `(trace-display (string-append ,(symbol->string v) ": "
				   (object->string ,v))))
  `(begin ,@(map trace-one-variable vars)))
				     

;;   Trace levels
;; Display parameters and return value of a function.
;; Increase the trace indentation to show the call hierarchy.
;; Do not preserve tail recursion.

(define (wrap-trace name lam)
  (lambda args
    (trace-display
     (if (null? args)
	 (string-append "[" name "]")
	 (apply string-append
		`("[" ,name
		  ,@(map (lambda (x) (string-append " " (object->string x)))
			 args) "]"))))
    (set! trace-level (1+ trace-level))
    (catch #t
	   (lambda ()
	     (let ((res (apply lam args)))
	       (set! trace-level (1- trace-level))      
	       (trace-display (object->string res))
	       res))
	   (lambda err
	     (set! trace-level (1- trace-level))
	     (apply throw err)))))

(define-macro (set-trace-level! . names)
  ;; Make each function a trace-level. Functions can be set multiple
  ;; times, only the first application is effective.
  ;; Parameters are function names
  `(begin
     ,@(map (lambda (name)
	      `(if (not (procedure-property ,name 'trace-wrapped))
		   (begin
		     (set! ,name (wrap-trace ,(symbol->string name) ,name))
		     (set-procedure-property! ,name 'trace-wrapped #t))))
	    names)))

;;   Trace points
;; Display parameters of a function when it is called.
;; Preserve tail recursion.

(define (wrap-trace-point lam msg)
  (lambda args
    (trace-display (string-append "[" msg " " (object->string args) "]"))
    (apply lam args)))

(define-macro (set-trace-point! name . opt)
  ;; Make one trace point.
  ;; Care must be taken of net setting the same function multiple times.
  (let ((msg (if (null? opt)
		 (symbol->string name)
		 (car opt))))
    `(set! ,name (wrap-trace-point ,name ,msg))))
