
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : def-menu.scm
;; DESCRIPTION : Definition of menus
;; COPYRIGHT   : (C) 1999  Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu grammar
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; ATTENTION: in menu-pre-entry, we also perform some convenience
;; rewritings in order to let menus match the grammar below.
;; The convience rewriting allow the use of ..., (check :string? :1),
;; (shortcut :string?) and multiple actions in menu entries.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-grammar
  (:menu-label (:or
    :string?
    (text :tuple? :string?)
    (icon :string?)
    (balloon :menu-label :string?)))
  (:menu-wide-label (:or
    :menu-label
    (check :menu-wide-label :string? :1)
    (shortcut :menu-wide-label :string?)))
  (:menu-entry (:menu-wide-label :1))
  (:menu-symbol (symbol :string? :*))
  (:menu-item-list (:repeat :menu-item))
  (:menu-item (:or
    ---
    |
    (group :string?)
    :menu-entry
    :menu-symbol
    (-> :menu-label :menu-item-list)
    (=> :menu-label :menu-item-list)
    (tile :integer? :menu-item-list)
    (if :1 :menu-item-list)
    (when :1 :menu-item-list)
    (link :1)
    (promise :1)
    (:menu-item-list)))
  (:menu (:or
    (:menu-item-list)
    (ver-menu :menu-item-list)
    (link :1)
    (promise :1))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Preparation of menu entries
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-promise x)
  (list 'unquote `(lambda () ,x)))

(define (menu-format-error which)
  (display* "TeXmacs] bad menu format in " which "\n")
  which)

(define (menu-label-add-dots l)
  (cond ((match? l ':string?) (string-append l "..."))
	((match? l '(text :tuple? :string?))
	 `(text ,(cadr l) ,(string-append (caddr l) "...")))
	((match? l '(icon :string?)) l)
	(else `(,(car l) ,(menu-label-add-dots (cadr l)) ,(caddr l)))))

(define (menu-pre-entry p)
  (if (= (length p) 2)
      p ;`(,(car p) ,(make-promise (cadr p)))
      (begin
	(display* "Reminder: " p "\n")
	(cond ((null? (cdr p)) (menu-format-error p))
	      ;; convenience rewritings
	      ((match? (cdr p) '(... :1 :*))
	       (menu-pre-entry `(,(menu-label-add-dots (car p)) ,@(cddr p))))
	      ((match? (cdr p) '(:string? :1 :*))
	       (menu-pre-entry `((shortcut ,(car p) ,(cadr p)) ,@(cddr p))))
	      ((match? (cdr p) '((check :2) :1 :*))
	       (menu-pre-entry `((check ,(car p) ,@(cdadr p)) ,@(cddr p))))
	      (else (menu-pre-entry `(,(car p) (begin ,@(cdr p)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Preparation of menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (menu-pre-list l)
  (map menu-pre l))

(define (menu-pre p)
  (if (pair? p)
      (cond ((string? (car p)) (menu-pre-entry p))
	    ((symbol? (car p))
	     (with result (ahash-ref menu-pre-table (car p))
	       (if (or (not result) (not (match? (cdr p) (car result))))
		   (menu-pre-list p)
		   ((cadr result) p))))
	    ((match? (car p) ':menu-wide-label) (menu-pre-entry p))
	    (else (menu-pre-list p)))
      (cond ((== p '---) p)
	    ((== p '|) p)
	    ((== p '()) p)
	    (else (menu-format-error p)))))

(define-table menu-pre-table
  (group (:string) ,id)
  (symbol (:string :*) ,id)
  (link (:1) ,id)
  (-> (:menu-label :*)
    ,(lambda (p) `(-> ,(cadr p) ,@(menu-pre-list (cddr p)))))
  (=> (:menu-label :*)
    ,(lambda (p) `(=> ,(cadr p) ,@(menu-pre-list (cddr p)))))
  (tile (:integer? :*)
    ,(lambda (p) `(tile ,(cadr p) ,@(menu-pre-list (cddr p)))))
  (if (:1 :*)
    ,(lambda (p) `(if ,(make-promise (cadr p)) ,@(menu-pre-list (cddr p)))))
  (when (:1 :*)
    ,(lambda (p) `(when ,(make-promise (cadr p)) ,@(menu-pre-list (cddr p)))))
  (promise (:1)
    ,(lambda (p) `(promise ,(make-promise (cadr p))))))

(ahash-set! menu-pre-table 'unquote `((:1) ,id))
(ahash-set! menu-pre-table 'unquote-splicing `((:1) ,id))

(define (menu-prepare p)
  (cond ((not (tuple? p)) p)
	((tuple? p 'ver-menu) `(ver-menu ,@(menu-pre-list (cdr p))))
	((tuple? p 'link 1) p)
	((tuple? p 'promise 1) `(promise ,(make-promise (cadr p))))
	(else (menu-pre-list p))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Manipulating menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define menu-bindings-table (make-ahash-table))

(define (menu-match? what label)
  (or (== what label)
      (match? label `(text :1 ,what))
      (and (match? label '((:or balloon check shortcut) :1 :*))
	   (menu-match? what (cadr label)))))

(define (menu-set-in! menu name what)
  (cond ((or (null? menu) (not (pair? menu))) #f)
	;((and (null? (cdr name))
	;      (match? (car menu) '(:menu-wide-label :*))
	;      (menu-match? (car name) (caar menu)))
	; (set-car! (cdar menu) what))
	((and (match? (car menu) '((:or -> =>) :menu-label :*))
	      (menu-match? (car name) (cadar menu)))
	 (if (null? (cdr name))
	     (begin (set-cdr! (cdar menu) what) what)
	     (menu-set-in! (cddar menu) (cdr name) what)))
	((match? (car menu) '(link :1))
	 (let ((ok (menu-set! (cons (cadar menu) name) what)))
	   (if ok ok (menu-set-in! (cdr menu) name what))))
	(else (menu-set-in! (cdr menu) name what))))

(define (menu-set! name menu)
  (cond ((null? name) (menu-set! 'texmacs-menu menu))
	((not (list? name)) (menu-set! (list name) menu))
	((symbol? (car name))
	 (if (null? (cdr name))
	     (ahash-set! menu-bindings-table (car name) menu)
	     (let ((old-menu (ahash-ref menu-bindings-table (car name))))
	       (if old-menu (menu-set-in! old-menu (cdr name) menu) #f))))
	((string? (car name))
	 (menu-set! (cons 'texmacs-menu name) menu))
	(else #f)))

(define (menu-get-in menu name)
  (cond ((null? name) menu)
	((or (null? menu) (not (pair? menu))) #f)
	;((and (null? (cdr name))
	;      (match? (car menu) '(:menu-wide-label :*))
	;      (menu-match? (car name) (caar menu)))
	; (cadar menu))
	((and (match? (car menu) '((:or -> =>) :menu-label :*))
	      (menu-match? (car name) (cadar menu)))
	 (menu-get-in (cddar menu) (cdr name)))
	((match? (car menu) '(link :1))
	 (let ((submenu (menu-get (cons (cadar menu) name))))
	   (if submenu submenu (menu-get-in (cdr menu) name))))
	(else (menu-get-in (cdr menu) name))))

(define (menu-get name)
  (cond ((null? name) (menu-get 'texmacs-menu))
	((not (list? name)) (menu-get (list name)))
	((symbol? (car name))
	 (let ((menu (ahash-ref menu-bindings-table (car name))))
	   (if menu (menu-get-in menu (cdr name)) #f)))
	((string? (car name))
	 (menu-get (cons 'texmacs-menu name)))
	(else #f)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Definition of menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-macro (menu-dynamic . body)
  (list 'quasiquote (menu-prepare body)))

(define-macro (menu-bind name . body)
  `(menu-set! ',name ,(list 'quasiquote (menu-prepare body))))

(define-macro (menu-extend name . body)
  `(menu-set! ',name (append (menu-get ',name)
			     ,(list 'quasiquote (menu-prepare body)))))
