
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : base.scm
;; DESCRIPTION : frequently used Scheme subroutines
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Booleans
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (xor-sub l)
  (cond ((null? l) #f)
	((car l) (not (xor-sub (cdr l))))
	(else (xor-sub (cdr l)))))

(tm-define (xor . l)
  (:type ((tuple bool) -> bool))
  (:synopsis "Exclusive or of all elements in @l.")
  (xor-sub l))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Strings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; NOTE: guile-1.6.0 implements SRFI-13 (string library) in C.

(tm-define (char->string c)
  (:type (char -> string))
  (:synopsis "Convert @c to a string")
  (list->string (list c)))

(tm-define (string-tail s n)
  (:type (string int -> string))
  (:synopsis "Return all but the first @n chars of @s.")
  (substring s n (string-length s)))

(tm-define (char-in-string? c s)
  (:type (char string -> bool))
  (:synopsis "Test whether @c occurs in @s")
  (not (== (string-index s c) #f)))

(tm-define (string-starts? s what)
  (:type (string string -> bool))
  (:synopsis "Test whether @s starts with @what.")
  (let ((n (string-length s))
	(k (string-length what)))
    (and (>= n k) (== (substring s 0 k) what))))

(tm-define (string-ends? s what)
  (:type (string string -> bool))
  (:synopsis "Test whether @s ends with @what.")
  (let ((n (string-length s))
	(k (string-length what)))
    (and (>= n k) (== (substring s (- n k) n) what))))

(tm-define (string-replace s what by)
  (:type (string string string -> string))
  (:synopsis "Systematically replace @what by @by in @s")
  (list->string
   (list-replace
    (string->list s)
    (string->list what)
    (string->list by))))

(tm-define (force-string s)
  (:type (object -> string))
  (:synopsis "Return @s if @s is a string and the empty string otherwise")
  (if (string? s) s ""))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tm-define (compose g f)
  (:type ((B -> C) (A -> B) -> (A -> C)))
  (:synopsis "Compose the functions @f and @g")
  (lambda x (g (apply f x))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Dictionaries
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (fill-dictionary-entry d key im)
  (if (not (null? key))
      (begin
	(ahash-set! d (car key) im)
	(fill-dictionary-entry d (cdr key) im))))

(tm-define (fill-dictionary d l)
  (:type (forall Key Im
	   ((alias (ahash-table Key Im)) (list (cross Key Im)) ->)))
  (:synopsis "Fill hash table @d with list of entries @l")
  (:warning "Depreciated")
  (if (not (null? l))
      (begin
	(let* ((r (reverse (car l))))
	  (fill-dictionary-entry d (cdr r) (car r)))
	(fill-dictionary d (cdr l)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Objects
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tm-define (string->object s)
  (:type (string -> object))
  (:synopsis "Parse @s and build scheme object")
  (call-with-input-string s read))

(tm-define (func? x f . opts)
  (:type (object symbol -> bool)
	 (object symbol int -> bool))
  (:synopsis "Is @x a list with first object @f?"
	     "Optionally test the length of @x.")
  (let ((n (length opts)))
    (cond ((= n 0) (and (list? x) (not (null? x)) (== (car x) f)))
	  ((= n 1)
	   (let ((nn (car opts)))
             (and (list? x) (not (null? x))
                  (== (car x) f) (= (length x) (+ nn 1)))))
	  (else (error "Too many arguments.")))))

(tm-define (tuple? x . opts)
  (:type (object -> bool)
	 (object symbol -> bool)
	 (object symbol int -> bool))
  (:synopsis "Equivalent to @list? without options"
	     "Equivalent to @func? otherwise")
  (if (null? opts)
      (list? x)
      (apply func? (cons x opts))))
