
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : general.scm
;; DESCRIPTION : general scheme routines for TeXmacs
;; COPYRIGHT   : (C) 2001  Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Subroutines
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tm-subtree p)
  (subtree (the-buffer) p))

(define (tm-debug)
  (display* (tree->object (the-buffer)) "\n"))

(define (tm-start p)
  (cursor-start (the-buffer) p))

(define (tm-end p)
  (cursor-end (the-buffer) p))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mode related
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (always?) #t)
(define (in-text?) (== (get-env "mode") "text"))
(define (in-math?) (== (get-env "mode") "math"))
(define (in-prog?) (== (get-env "mode") "prog"))
(define (in-session?) (inside? "session"))
(define (not-in-session?) (not (inside? "session")))
(define (in-math-in-session?) (and (in-math?) (in-session?)))
(define (in-math-not-in-session?) (and (in-math?) (not-in-session?)))
(define (in-plugin-with-converters?)
  (plugin-supports-math-input-ref (get-env "prog language")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Language related
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (in-czech?) (and (in-text?) (== (get-env "language") "czech")))
(define (in-dutch?) (and (in-text?) (== (get-env "language") "dutch")))
(define (in-english?) (and (in-text?) (== (get-env "language") "english")))
(define (in-finnish?) (and (in-text?) (== (get-env "language") "finnish")))
(define (in-french?) (and (in-text?) (== (get-env "language") "french")))
(define (in-german?) (and (in-text?) (== (get-env "language") "german")))
(define (in-hungarian?)
  (and (in-text?) (== (get-env "language") "hungarian")))
(define (in-italian?) (and (in-text?) (== (get-env "language") "italian")))
(define (in-polish?) (and (in-text?) (== (get-env "language") "polish")))
(define (in-portugese?)
  (and (in-text?) (== (get-env "language") "portugese")))
(define (in-romanian?)
  (and (in-text?) (== (get-env "language") "romanian")))
(define (in-russian?) (and (in-text?) (== (get-env "language") "russian")))
(define (in-spanish?) (and (in-text?) (== (get-env "language") "spanish")))
(define (in-swedish?) (and (in-text?) (== (get-env "language") "swedish")))
(define (in-ukrainian?)
  (and (in-text?) (== (get-env "language") "ukrainian")))

(define (in-cyrillic?) (or (in-russian?) (in-ukrainian?)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Keyboard related
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (like-emacs?) (== (get-preference "look and feel") "emacs"))
(define (like-windows?) (== (get-preference "look and feel") "windows"))
(define (like-old?) (== (get-preference "look and feel") "old style"))
(define (like-old-text?) (and (like-old?) (in-text?)))
(define (like-old-math?) (and (like-old?) (in-math?)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; General commands
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (load-expr file)
  (read (open-file (decode-file-name file) OPEN_READ)))
 
(define (save-expr file value)
  (write value (open-file (decode-file-name file) OPEN_WRITE))
  (flush-all-ports))

(define (get-env s)
  (tree->string (get-env-tree s)))

(define (preview-with-ghostview)
  (print-to-file "$TEXMACS_HOME_PATH/system/preview.ps")
  (cond ((file-exists? "$PATH" "ggv")
	 (shell "ggv $TEXMACS_HOME_PATH/system/preview.ps &"))
	((file-exists? "$PATH" "ghostview")
	 (shell "ghostview $TEXMACS_HOME_PATH/system/preview.ps &"))
	((file-exists? "$PATH" "gv")
	 (shell "gv $TEXMACS_HOME_PATH/system/preview.ps &"))
	(else (set-message
	       "Error: ghostview does not seem to be installed on your system"
	       "preview"))))

(define (choose-file-and-print-page-selection start end)
  (choose-file "Print page selection to file" "postscript"
	       `(lambda (name) (print-pages-to-file name ,start ,end))))

(define (conditional-kill-buffer confirm)
  (if (yes? confirm) (kill-buffer)))

(define (safely-kill-buffer)
  (if (buffer-unsaved?)
      (interactive
       '("The buffer has not been saved. Really close it?")
       'conditional-kill-buffer)
      (kill-buffer)))

(define (safely-kill-window)
  (if (<= (get-nr-windows) 1) (safely-quit-TeXmacs) (kill-window)))

(define (conditional-quit-TeXmacs confirm)
  (if (yes? confirm) (quit-TeXmacs)))

(define (safely-quit-TeXmacs)
  (if (exists-unsaved-buffer?)
      (interactive
       '("There are unsaved files. Really quit?")
       'conditional-quit-TeXmacs)
      (quit-TeXmacs)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Loading and saving
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (conditional-save-buffer file fm confirm)
  (if (yes? confirm) (texmacs-save-buffer file fm)))

(define (secure-save-buffer file fm)
  (if (file-exists? "." file)
      (interactive '("File already exists. Overwrite existing file?")
		   `(lambda (confirm)
		      (conditional-save-buffer ,file ,fm confirm)))
      (texmacs-save-buffer file fm)))

(define (save-buffer . l)
  (cond ((= (length l) 0)
	 (if (no-name?)
	     (interactive '("Save as:") 'save-buffer)
	     (texmacs-save-buffer (get-name-buffer) "generic")))
	((= (length l) 1) (secure-save-buffer (car l) "generic"))
	((= (length l) 2) (secure-save-buffer (car l) (cadr l)))
	(else (secure-save-buffer
	       (unique-file-name (car l) (cadr l)) (caddr l)))))

(define (conditional-load-buffer-path dir file fm where confirm)
  (if (yes? confirm)
      (texmacs-load-buffer dir (string-append file "~") fm where #t)
      (texmacs-load-buffer dir file fm where #f)))

(define (load-buffer-path-bis dir file fm where)
  (if (and (file-exists? dir file)
	   (file-exists? dir (string-append file "~")))
      (let ((f1 (unique-file-name dir file))
	    (f2 (unique-file-name dir (string-append file "~"))))
	(if (newer? f2 f1)
	    (interactive
	     '("Load more recent autosave file?")
	     `(lambda (confirm)
		(conditional-load-buffer-path ,dir ,file ,fm ,where confirm)))
	    (texmacs-load-buffer dir file fm where #f)))
      (texmacs-load-buffer dir file fm where #f)))

(define (load-buffer-path . l)
  (cond ((= (length l) 2)
	 (load-buffer-path-bis (car l) (cadr l) "generic" 0))
	((and (= (length l) 3) (string? (caddr l)))
	 (load-buffer-path-bis (car l) (cadr l) (caddr l) 0))
	((and (= (length l) 3) (integer? (caddr l)))
	 (load-buffer-path-bis (car l) (cadr l) "generic" (caddr l)))
	(else (load-buffer-path-bis
	       (car l) (cadr l) (caddr l) (cadddr l)))))

(define (load-buffer . l)
  (cond ((= (length l) 1)
	 (load-buffer-path-bis "$TEXMACS_FILE_PATH" (car l) "generic" 0))
	((and (= (length l) 2) (string? (cadr l)))
	 (load-buffer-path-bis "$TEXMACS_FILE_PATH" (car l) (cadr l) 0))
	((and (= (length l) 2) (integer? (cadr l)))
	 (load-buffer-path-bis
	  "$TEXMACS_FILE_PATH" (car l) "generic" (cadr l)))
	(else (load-buffer-path-bis
	       "$TEXMACS_FILE_PATH" (car l) (cadr l) (caddr l)))))

(define (load-texmacs-buffer s) (load-buffer s "TeXmacs"))
(define (save-texmacs-buffer s) (save-buffer s "TeXmacs"))
(define (load-latex-buffer s) (load-buffer s "latex"))
(define (save-latex-buffer s) (save-buffer s "latex"))
(define (load-html-buffer s) (load-buffer s "html"))
(define (save-html-buffer s) (save-buffer s "html"))
(define (load-scheme-buffer s) (load-buffer s "scheme"))
(define (save-scheme-buffer s) (save-buffer s "scheme"))
(define (load-verbatim-buffer s) (load-buffer s "verbatim"))
(define (save-verbatim-buffer s) (save-buffer s "verbatim"))
(define (load-in-new-window s) (load-buffer s 1))
(define (load-browse-buffer s)
  (if (help-buffer?) (load-buffer s "help") (load-buffer s)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Loading help buffers
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define help-file-path
  "$TEXMACS_DOC_PATH:$TEXMACS_HOME_PATH/doc:$TEXMACS_PATH/doc")

(define (help-file-exists? s)
  (file-exists? help-file-path s))

(define (get-help-name s)
  (if (or (string-ends? s ".tex") (string-ends? s ".tm") (file-exists? "." s))
      s
      (let* ((lan (get-output-language))
	     (suf (cond ((== lan "czech") ".cs.tm")
			((== lan "dutch") ".nl.tm")
			((== lan "finnish") ".fi.tm")
			((== lan "french") ".fr.tm")
			((== lan "german") ".de.tm")
			((== lan "hungarian") ".hu.tm")
			((== lan "italian") ".it.tm")
			((== lan "polish") ".pl.tm")
			((== lan "portuguese") ".pt.tm")
			((== lan "romanian") ".ro.tm")
			((== lan "russian") ".ru.tm")
			((== lan "spanish") ".es.tm")
			((== lan "swedish") ".sv.tm")
			((== lan "ukrainian") ".uk.tm")
			(else ".en.tm")))
	     (dir help-file-path))
	(cond ((file-exists? dir (string-append s suf))
	       (unique-file-name dir (string-append s suf)))
	      ((and (not (== suf ".en.tm"))
		    (file-exists? dir (string-append s ".en.tm")))
	       (unique-file-name dir (string-append s ".en.tm")))
	      (else "")))))

(define (load-help-buffer-sub s type)
  (let ((name (get-help-name s)))
    (if (== name "")
	(set-message (string-append "Error: help file#'" s "'#not found")
		     "load help file")
	(cond ((== type "normal")
	       (let ((doc (load-file-as-tree "." name "help")))
		 (if (== (tree->object doc) "error")
		     (set-message "Bad help file" "load help file")
		     (set-help-buffer name doc))))
	      ((== type "article") (tmdoc-expand-help name 'tmdoc-title))
	      ((== type "book") (tmdoc-expand-help name 'title))))))

(define (load-help-buffer s) (load-help-buffer-sub s "normal"))
(define (load-help-article s) (load-help-buffer-sub s "article"))
(define (load-help-book s) (load-help-buffer-sub s "book"))

(define (load-help-online s)
  (load-help-buffer
   (string-append "http://www.gnu.org/software/texmacs-doc/" s)))

(define (update-help-online)
  (system "cd $TEXMACS_HOME_PATH; wget ftp://ftp.texmacs.org/pub/TeXmacs/doc/TeXmacs-doc.tar.gz -O TeXmacs-doc.tar.gz; gunzip TeXmacs-doc.tar.gz; tar -xvf TeXmacs-doc.tar; rm -f TeXmacs-doc.tar"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initialization commands
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (init-default . args)
  (for-each init-default-one args))

(define (init-page-margins l r t b)
  (init-env "odd page margin" l)
  (init-env "even page margin" l)
  (init-env "page right margin" r)
  (init-env "page top margin" t)
  (init-env "page bottom margin" b)
  (init-text-width (length- (get-env "page width") l r)))

(define (init-screen-reduction l r t b)
  (init-env "reduction page left margin" l)
  (init-env "reduction page right margin" r)
  (init-env "reduction page top margin" t)
  (init-env "reduction page bottom margin" b))

(define (init-page-size w h)
  (init-env "page type" "user")
  (init-env "page width" w)
  (init-env "page height" h)
  (init-page-margins "5mm" "5mm" "5mm" "5mm")
  (init-screen-reduction "0cm" "0cm" "0cm" "0cm"))

(define (init-text-width s) (init-env "paragraph width" s))
(define (init-font-size s) (init-env "font base size" s))
(define (init-dpi s) (init-env "dpi" s))
(define (init-first-indent s) (init-env "first indentation" s))
(define (init-interline s) (init-env "interline space" s))
(define (init-interline-spc s) (init-env "line stretch" s))
(define (init-interpar-spc s) (init-env "interparagraph space" s))
(define (init-magn s) (init-env "magnification" s))
(define (other-shrinking-factor s) (set-shrinking-factor (string->number s)))
(define (init-page-shrink s) (init-env "page shrink" s))
(define (init-page-extend s) (init-env "page extend" s))
(define (init-page-flexibility s) (init-env "page flexibility" s))
(define (init-as-on-paper)
  (init-screen-reduction "0mm" "0mm" "0mm" "0mm")
  (init-env "show header and footer" "true"))

(define (init-language lan)
  (let ((before (in? (tree->object (get-init-tree "language"))
		     '("russian" "ukrainian")))
	(after (in? lan '("russian" "ukrainian"))))
    (if (and before (not after)) (init-default "font"))
    (init-env "language" lan)
    (if (and after (not before)) (init-env "font" "cyrillic"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Toggling properties
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; For the View menu
(define (toggle-full-screen-mode) (full-screen-mode (not (full-screen?))))
(define (toggle-visible-header) (show-header (not (visible-header?))))
(define (toggle-visible-icon-bar n)
  (show-icon-bar n (not (visible-icon-bar? n))))
(define (toggle-visible-footer) (show-footer (not (visible-footer?))))

;; For the Insert menu
(define (toggle-insertion-position what)
  (position-insertion what (not (test-insertion-position? what))))

;; For the Document menu
(define (toggle-preamble)
  (let ((new (if (string=? (get-env "preamble") "true") "false" "true")))
    (init-env "preamble" new)
    (if (== new "false")
	(begin (generate-all-aux) (update-buffer)))))

(define (toggle-visible-header-and-footer)
  (init-env "show header and footer"
	    (if (== (get-env "show header and footer") "true")
		"false" "true")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; try to obtain the papersize in this order from
;; - the environment variable PAPERCONF
;; - the contents of the file specified by the PAPERSIZE environment variable
;; - the contents of the file "/etc/papersize"
;; or else default to "a4"
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (get-default-paper-size)
  (or (getenv "PAPERSIZE")
      (let ((papersizefile (or (getenv "PAPERCONF") '"/etc/papersize")))
	(if
	 (access? papersizefile R_OK)
	 (let ((pps-port (open-input-file papersizefile)))
	   (let ((size (read-line pps-port)))
	     (begin
	       (close-input-port pps-port)
	       size)))
	 "a4"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Cursor handling
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (cursor-left)
  (if (inside? "input") (session-go-left) (go-left)))

(define (cursor-right)
  (if (inside? "input") (session-go-right) (go-right)))

(define (cursor-up)
  (if (inside? "input") (session-go-up) (go-up)))

(define (cursor-down)
  (if (inside? "input") (session-go-down) (go-down)))

(define (cursor-page-up)
  (if (inside? "input") (session-go-page-up) (go-page-up)))

(define (cursor-page-down)
  (if (inside? "input") (session-go-page-down) (go-page-down)))

(define (cursor-start-line) (go-start-line))
(define (cursor-end-line) (go-end-line))

(define (cursor-select r)
  (select-from-shift-keyboard)
  (r)
  (select-from-cursor))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Length arithmetic
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (length+ . ops)
  (define (length-add* ops res)
    (if (null? ops)
	res
	(length-add* (cdr ops) (length-add res (car ops)))))
  (if (null? ops)
      "0unit"
      (length-add* (cdr ops) (car ops))))

(define (length- . ops)
  (if (null? (cdr ops))
      (string-append "-" (car ops))
      (length+ (car ops) (length- (apply length+ (cdr ops))))))

(define (length-zero? len)
  (= 0 (length-decode len)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Miscellaneous commands
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (test-env? var val)
  (== (get-env var) val))

(define (test-init? var val)
  (== (get-init-tree var) (string->tree val)))

(define (test-default? . vals)
  (if (null? vals)
      #t
      (and (not (init-has? (car vals)))
	   (apply test-default? (cdr vals)))))

(define (real-math-font? fn)
  (or (== fn "roman") (== fn "concrete")))

(define (real-math-family? fn)
  (or (== fn "mr") (== fn "ms") (== fn "mt")))

(define (filter-maxima-in session s)
  (cond ((== s "") "0;")
	((== (car (reverse (string->list s))) #\;) s)
	((== (car (reverse (string->list s))) #\$) s)
	(else (string-append s ";"))))

(define (extern-exec s)
  (if (inside? "session")
      (connection-exec (get-env "prog language")
		       (get-env "this session")
		       (string->tree s))))

(define (not-implemented s) (set-message "Error: not yet implemented" s))

(define (kbd-select-enlarge)
  (if (selection-active-enlarging?)
      (select-enlarge)
      (begin
	(selection-cancel)
	(selection-set-start)
	(select-from-keyboard #t))))

(define (kbd-select-environment)
  (if (selection-active-enlarging?)
      (select-enlarge-environmental)
      (begin
	(selection-cancel)
	(selection-set-start)
	(select-from-keyboard #t)
	(select-enlarge-environmental))))

(define (kill-line)
  (selection-set-start)
  (go-end-line)
  (selection-set-end)
  (cut "primary"))

(define (replace-start-forward what by)
  (replace-start what by #t))

(define (copy-export format which)
  (let ((temp (selection-get-export)))
    (selection-set-export format)
    (copy which)
    (selection-set-export temp)))

(define (cut-export format which)
  (let ((temp (selection-get-export)))
    (selection-set-export format)
    (cut which)
    (selection-set-export temp)))

(define (paste-import format which)
  (let ((temp (selection-get-import)))
    (selection-set-import format)
    (paste which)
    (selection-set-import temp)))

(define (with-active-buffer-sub name cmd)
  (let ((old (get-name-buffer)))
    (switch-to-active-buffer name)
    (eval cmd)
    (switch-to-active-buffer old)))

(define-macro with-active-buffer
  (lambda l (with-active-buffer-sub (car l) (cons 'begin (cdr l)))))
