
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : menu.scm
;; DESCRIPTION : routines for generating menus
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven, David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; See def-menu.scm for the grammar of menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-empty) (widget-harray '() -1))
(define (make-menu-error . args)
  (apply tm-display-error args)
  (widget-text "Error" #f ""))
(define (make-menu-bad-format p e?)
  (make-menu-error "menu has bad format in " (object->string p)))
(define (pixel) 256) ; TODO: length arithmetic
(define (widget-empty) (widget-glue #f #f 0 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Automatic check marks
;; Should be moved to corresponding routines later
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (set-check-mark! proc mark pred)
  (set-procedure-property! proc :check-mark (list mark pred)))
(define (get-check-mark proc)
  (procedure-property proc :check-mark))

(set-check-mark! init-default "*" test-default?)
(set-check-mark! make-with "o" test-env?)
(set-check-mark! init-env "*" test-init?)

(define (test-shrinking-factor? n) (= (get-shrinking-factor) n))
(set-check-mark! toggle-full-screen-mode "v" full-screen?)
(set-check-mark! toggle-visible-header "v" visible-header?)
(set-check-mark! toggle-visible-icon-bar "v" visible-icon-bar?)
(set-check-mark! toggle-visible-footer "v" visible-footer?)
(set-check-mark! set-shrinking-factor "*" test-shrinking-factor?)

(define (test-preamble?) (string=? (get-env "preamble") "true"))
(define (test-page-medium? s) (string=? (get-env "page medium") s))
(define (test-page-type? s) (string=? (get-env "page type") s))
(define (test-page-orientation? s) (string=? (get-env "page orientation") s))
(define (visible-header-and-footer?)
  (== "true" (get-env "show header and footer")))
(define (as-on-paper?)
  (and (visible-header-and-footer?)
       (length-zero? (get-env "reduction page left margin"))
       (length-zero? (get-env "reduction page right margin"))
       (length-zero? (get-env "reduction page top margin"))
       (length-zero? (get-env "reduction page bottom margin"))))
(set-check-mark! toggle-preamble "v" test-preamble?)
(set-check-mark! set-page-medium "*" test-page-medium?)
(set-check-mark! set-page-type "*" test-page-type?)
(set-check-mark! set-page-orientation "*" test-page-orientation?)
(set-check-mark! toggle-visible-header-and-footer
		 "v" visible-header-and-footer?)
(set-check-mark! init-as-on-paper "o" as-on-paper?)

(define (selection-test-import? s) (string=? s (selection-get-import)))
(define (selection-test-export? s) (string=? s (selection-get-export)))
(define (session-test-math-input? b?) (if (session-math-input?) b? (not b?)))
(set-check-mark! selection-set-import "*" selection-test-import?)
(set-check-mark! selection-set-export "*" selection-test-export?)
(set-check-mark! session-use-math-input "*" session-test-math-input?)
(set-check-mark! toggle-insertion-position "v" test-insertion-position?)
(define (toggle-insertion-position-not s) (toggle-insertion-position s))
(define (test-insertion-position-not? s) (not (test-insertion-position? s)))
(set-check-mark! toggle-insertion-position-not
		 "v" test-insertion-position-not?)

(define (test-cell-mode? s) (string=? (get-cell-mode) s))
(set-check-mark! set-cell-mode "*" test-cell-mode?)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu labels
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-label p e? . opt)
  "Make widget for menu label @p."
  ;; Possibilities for p:
  ;;   <label> :: (balloon <label> <string>)
  ;;     Label with a popup balloon. The <string> is the balloon text.
  ;;   <label> :: (text <font desc> <string>)
  ;;     Label <string> drawn in black text of an arbitrary font.
  ;;     <font desc> :: ([family [class [series [shape [size [dpi]]]]]])
  ;;     Example default values are: family="roman", class="mr",
  ;;     series="medium", shape="normal", size=10, dpi=600.
  ;;   <label> :: <string>
  ;;     Simple menu label, its display style is controlled by tt? and e?
  ;;   <label> :: (icon <string>)
  ;;     Pixmap menu label, the <string> is the name of the pixmap.
  (let ((tt? (and (not (null? opt)) (car opt)))
	(col (color (if e? "black" "dark grey"))))
    (cond ((string? p)			; "text"
	   (widget-menu-text p col (get-input-language) tt?))
  	  ((tuple? p 'balloon 2)        ; (balloon <label> "balloon text")
  	   (make-menu-label (cadr p) e? tt?))
  	  ((tuple? p 'text 2)		; (text <font desc> "text")
	   (widget-box (cadr p) (caddr p) col #t #t))
  	  ((tuple? p 'icon 1)		; (icon "name.xpm")
  	   (widget-xpm (cadr p) #t)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Elementary menu items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-hsep)
  "Make @--- menu item."
  (widget-separator (* 2 (pixel)) (* 2 (pixel)) #f))

(define (make-menu-vsep)
  "Make @| menu item."
  (widget-separator (* 2 (pixel)) (* 2 (pixel)) #t))

(define (make-menu-group s)
  "Make @(group :string?) menu item."
  (widget-command-button-3
   (widget-empty) (make-menu-label s #f) (widget-empty)
   (object->command noop)
   #f #t))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu entries
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-entry-button e? bar? check label short command)
  (if bar?
      (widget-command-button-1 (make-menu-label label e?)
			       command #f)
      (widget-command-button-3
       (if (string=? check "")
	   (widget-empty)
	   (widget-box '()
		       (cadr (assoc check '(("v" "<checked>")
					    ("o" "<circ>")
					    ("*" "<bullet>"))))
		       (color (if e? "black" "dark grey"))
		       #t #f))
       (make-menu-label label e?)
       (if (string=? short "")
	   (widget-empty)
	   (make-menu-label short e? #t))
       command
       e? #f)))

(define (make-menu-entry-shortcut label action opt-key)
  (cond (opt-key opt-key)
	((pair? label) "")
	(else (get-inverse-binding action))))

(define (make-menu-entry-check opt-check action)
  (or opt-check
      (and (pair? action)
	   (let ((proc (eval (car action))))
	     (and (procedure? proc)
		  (let ((prop (get-check-mark proc)))
		    (and prop
			 (list (car prop)
			       (cons (cadr prop) (cdr action))))))))))

(define (make-menu-entry-eval-check opt-check)
  (if (and opt-check (eval-bool-or-false (cadr opt-check)))
      (car opt-check)
      ""))

(define (make-menu-entry-attrs label action opt-key opt-check)
  (cond ((match? label '(shortcut :1 :string?))
	 (make-menu-entry-attrs (cadr label) action (caddr label) opt-check))
	((match? label '(check :1 :string? :1))
	 (make-menu-entry-attrs (cadr label) action opt-key (cddr label)))
	(else (values label action opt-key opt-check))))

(define (make-menu-entry-sub p e? bar?)
  (receive
      (label action opt-key opt-check)
      (make-menu-entry-attrs (car p) (cAr p) #f #f)
    (make-menu-entry-button
     e? bar?
     (make-menu-entry-eval-check (make-menu-entry-check opt-check action))
     label
     (make-menu-entry-shortcut label action opt-key)
     (server-command (if e? (wrap-catch-list action) 'noop))
     )))

(define (make-menu-entry p e? bar?)
  "Make @:menu-item menu item."
  (let ((button (make-menu-entry-sub p e? bar?))
	(label (car p)))
    (if (tuple? label 'balloon 2)
	(widget-balloon button
			(widget-text (caddr label) #t (get-input-language)))
	button)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Symbol fields
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-symbol-button e? symstring opt-symobj)
  (widget-command-button-1
   (widget-box '() symstring
	       (color (if e? "black" "dark grey"))
	       #t #f)
   (object->command (if opt-symobj
			(lambda () (insert-object opt-symobj))
			(lambda () (insert-string symstring))))
   #f))

(define (make-menu-symbol p e?)
  "Make @(symbol :string? :*) menu item."
  ;; Possibilities for p:
  ;;   <menu-symbol> :: (symbol <symbol-string> [<symbol-object>] [<shortcut>])
  ;;   <symbol-object> :: (<symbol-type> <symbol-name>)
  ;;   <symbol-type> :: left | right | mid | big
  ;;   <symbol-name> :: <string>
  (with (tag symstring . opt) p
    (let ((error? #f) (opt-shortcut #f) (opt-symobj #f))
      (for-each
       (lambda (x)
	 (cond (error? (noop))
	       ((pair? x) (set! opt-symobj x))
	       ((string? x) (set! opt-shortcut x))
	       (else (set! error? #t))))
       opt)
      (if error? (make-menu-error "invalid symbol attribute in " p)
	  (let ((sh (if opt-shortcut opt-shortcut
			(get-inverse-binding (string->symbol symstring)))))
	    (if (string=? sh "")
		(make-menu-symbol-button e? symstring opt-symobj)
		(widget-balloon
		 (make-menu-symbol-button e? symstring opt-symobj)
		 (make-menu-label (string-append "Keyboard equivalent: " sh)
				  e?))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Submenus and tile menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-submenu p e?)
  "Make @((:or -> =>) :menu-label :menu-item-list) menu item."
  (with (tag label . items) p
    (let ((button
	   ((cond ((== tag '=>) widget-pulldown-button-lazy)
		  ((== tag '->) widget-pullright-button-lazy))
	    (make-menu-label label e?)
	    (object->make-widget
	     (lambda () (make-menu-widget (cons 'ver-menu items) e?))))))
      (if (tuple? label 'balloon 2)
	  (widget-balloon button
			  (widget-text (caddr label) #t (get-input-language)))
	  button))))

(define (make-menu-tile p e?)
  "Make @(tile :integer? :menu-item-list) menu item."
  (with (tag width . items) p
    (widget-tile (make-menu-items items e? #f) width)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Dynamic menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-if p e? bar?)
  "Make @(if :1 :menu-item-list) menu items."
  (with (tag pred? . items) p
    (if (pred?) (make-menu-items-list items e? bar?) '())))

(define (make-menu-when p e? bar?)
  "Make @(when :1 :menu-item-list) menu items."
  (with (tag pred? . items) p
    (make-menu-items-list items (and e? (pred?)) bar?)))

(define (make-menu-link p e? bar?)
  "Make @(link :1) menu items."
  (with linked (menu-get (cadr p))
    (if linked (make-menu-items linked e? bar?)
	(make-menu-error "bad link: " (object->string (cadr p))))))

(define (make-menu-promise p e? bar?)
  "Make @(promise :1) menu items."
  (with (error? value) (safe-execute (cadr p))
    (if (not error?) (make-menu-items value e? bar?)
	(make-menu-error "bad promise: " (object->string (cadr p))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Main routines for making menu items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-items-list l e? bar?)
  "Make menu items for each element in @l and append results."
  (append-map (lambda (p) (make-menu-items p e? bar?)) l))

(define (make-menu-items p e? bar?)
  "Make menu items @p. The items are on a bar if @bar? and greyed if not @e?."
  (cond ((== p '---) (list (make-menu-hsep)))
	((== p '|) (list (make-menu-vsep)))
	((match? p '(group :string?)) (list (make-menu-group (cadr p))))
	((match? p '(:menu-wide-label :1)) (list (make-menu-entry p e? bar?)))
	((match? p '(symbol :string? :*)) (list (make-menu-symbol p e?)))
	((match? p '(-> :menu-label :*)) (list (make-menu-submenu p e?)))
	((match? p '(=> :menu-label :*)) (list (make-menu-submenu p e?)))
	((match? p '(link :1)) (make-menu-link p e? bar?))
	((match? p '(promise :1)) (make-menu-promise p e? bar?))
	((match? p '(if :1 :*)) (make-menu-if p e? bar?))
	((match? p '(when :1 :*)) (make-menu-when p e? bar?))
	((match? p '(tile :integer? :*)) (list (make-menu-tile p e?)))
	((match? p '(:*)) (make-menu-items-list p e? bar?))
	(else (list (make-menu-bad-format p e?)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu bodies
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-hmenu p e?)
  (widget-harray (make-menu-items p e? #t) -1))

(define (make-menu-vmenu p e?)
  (widget-vmenu (make-menu-items (cdr p) e? #f)))

(define (make-menu-linked-menu p e?)
  (with linked (menu-get (cadr p))
    (if linked (make-menu-main linked e?)
	(make-menu-error "bad link: " (object->string (cadr p))))))

(define (make-menu-promised-menu p e?)
  (with (error? value) (safe-execute (cadr p))
    (if (not error?) (make-menu-main value e?)
	(make-menu-error "bad promise: " (object->string (cadr p))))))

(define (make-menu-main p e?)
  "Transform the menu @p into a widget."
  (cond ((not (tuple? p)) (make-menu-bad-format p e?))
	((tuple? p 'ver-menu) (make-menu-vmenu p e?))
	((tuple? p 'link 1) (make-menu-linked-menu p e?))
	((tuple? p 'promise 1) (make-menu-promised-menu p e?))
	(else (make-menu-hmenu p e?))))
 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Debugging
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; (set-trace-level! make-menu-entry-button)
; (set-trace-level! make-menu-entry-shortcut)
; (set-trace-level! make-menu-entry-check)
; (set-trace-level! make-menu-entry-eval-check)

; (set-trace-point! make-menu-hsep          "case: ---")
; (set-trace-point! make-menu-vsep          "case: |")
; (set-trace-point! make-menu-group         "case: group")
; (set-trace-point! make-menu-items-list    "case: list")
; (set-trace-point! make-menu-if            "case: if")
; (set-trace-point! make-menu-when          "case: when")
; (set-trace-point! make-menu-submenu       "case: =>/->")
; (set-trace-point! make-menu-tile          "case: tile")
; (set-trace-point! make-menu-link          "case: link")
; (set-trace-point! make-menu-promise       "case: promise")
; (set-trace-point! make-menu-entry         "case: item")
; (set-trace-point! make-menu-symbol        "case: symbol")

; (set-trace-level! make-menu-widget)
; (set-trace-level! make-menu-entry-sub)
; (set-trace-level! make-menu-items)

; (set-trace-point! make-menu-hmenu         "case: hmenu")
; (set-trace-point! make-menu-vmenu         "case: vmenu")
; (set-trace-point! make-menu-linked-menu   "case: link")
; (set-trace-point! make-menu-promised-menu "case: promise")
 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interface
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(tm-define (make-menu-widget p e?)
  (:type (object -> widget))
  (:synopsis "Transform a menu into a widget.")
  (:args (p "a scheme object which represents the menu")
	 (e? "greyed menu if @e? is @#f"))
  ((wrap-catch make-menu-main) p e?))
