
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : preferences.scm
;; DESCRIPTION : management of the user preferences
;; COPYRIGHT   : (C) 1999  Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The default preferences
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define default-preferences
  `(("profile" "beginner" noop)
    ("look and feel" "emacs" notify-look-and-feel)
    ("A" "default" notify-prefix-modifier)
    ("M" "default" notify-prefix-modifier)
    ("H" "default" notify-prefix-modifier)
    ("alt" "default" notify-key-prefix)
    ("meta" "default" notify-key-prefix)
    ("windows" "default" notify-key-prefix)
    ("caps-lock" "default" notify-key-prefix)
    ("language" ,(get-locale-language) notify-language)
    ("cyrillic input method" "" notify-cyrillic-input-method)
    ("printing command" "lpr" notify-printing-command)
    ("paper type" "a4" notify-paper-type)
    ("printer dpi" "600" notify-printer-dpi)
    ("header" "on" notify-header)
    ("main icon bar" "on" notify-icon-bar)
    ("context dependent icons" "on" notify-icon-bar)
    ("user provided icons" "off" notify-icon-bar)
    ("status bar" "on" notify-status-bar)
    ("shrinking factor" "5" notify-shrinking-factor)
    ("security" "prompt on scripts" notify-security)))

(define preinit-preferences
  '("A" "M" "H"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Keyboard configuration
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define prefix-modifier-table (make-ahash-table))

(define (get-default-alt-modifier)
  (let ((s (eval-system "xmodmap | grep Alt_L")))
    (cond ((string-starts? s "mod")
	   (string-append "Mod" (substring s 3 4)))
	  (else "Mod1"))))

(define (get-default-meta-modifier)
  (let ((s (eval-system "xmodmap | grep Meta_L")))
    (cond ((string-starts? s "mod")
	   (string-append "Mod" (substring s 3 4)))
	  (else (get-default-super-modifier)))))

(define (get-default-super-modifier)
  (let ((s (eval-system "xmodmap | grep Super_L")))
    (cond ((string-starts? s "mod")
	   (string-append "Mod" (substring s 3 4)))
	  (else "Mod4"))))

(define (get-default-hyper-modifier)
  (let ((s (eval-system "xmodmap | grep Hyper_L")))
    (cond ((string-starts? s "mod")
	   (string-append "Mod" (substring s 3 4)))
	  (else "Mod3"))))

(define (notify-prefix-modifier var val)
  (cond ((not (== val "default"))
	 (ahash-set! prefix-modifier-table var val)
	 (if preferences-initialization-flag
	     (set-message
	      "Restart in order to let the new look and feel take effect"
	      "configure look and feel")))
	((== var "A")
	 (notify-prefix-modifier var (get-default-alt-modifier)))
	((== var "M")
	 (notify-prefix-modifier var (get-default-meta-modifier)))
	((== var "H")
	 (notify-prefix-modifier var (get-default-hyper-modifier)))))

(define (get-prefix-modifier s)
  (ahash-ref prefix-modifier-table s))

(define (get-prefix-modifiers l)
  (list-sort (map-in-order get-prefix-modifier l) string>?))

(define (modifier-list->string l)
  (apply string-append (map-in-order (lambda (s) (string-append s "-")) l)))

(define (compute-wildcard-line l)
  (let* ((r (get-prefix-modifiers l))
	 (L (modifier-list->string l))
	 (R (modifier-list->string r)))
    (list R L #t)))

(define (compute-wildcard-lines)
  (map-in-order compute-wildcard-line
   '(("A")
     ("M")
     ("H")
     ("M" "A")
     ("H" "A")
     ("H" "M")
     ("H" "M" "A"))))

(define (xmodmap s)
  (system (string-append "xmodmap -e \"" s "\"")))

(define (notify-key-prefix var val)
  (let* ((Modx (get-prefix-modifier val)))
    ;(display* var " -> " val " -> " Modx "\n")
    (cond ((== Modx #f) (noop))
	  ((== var "alt")
	   (xmodmap (string-append "clear " Modx))
	   (xmodmap (string-append "add " Modx " = Alt_L"))
	   (xmodmap (string-append "add " Modx " = Alt_R")))
	  ((== var "meta")
	   (xmodmap (string-append "clear " Modx))
	   (xmodmap (string-append "add " Modx " = Meta_L"))
	   (xmodmap (string-append "add " Modx " = Meta_R")))
	  ((== var "windows")
	   (xmodmap "keycode 115 = Super_L")
	   (xmodmap "keycode 116 = Super_R")
	   (xmodmap (string-append "clear " Modx))
	   (xmodmap (string-append "add " Modx " = Super_L"))
	   (xmodmap (string-append "add " Modx " = Super_R")))
	  ((== var "caps-lock")
	   (xmodmap "clear Lock")
	   (xmodmap (string-append "clear " Modx))
	   (xmodmap (string-append "add " Modx " = Caps_Lock"))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Other call-backs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (notify-look-and-feel var val)
  (if preferences-initialization-flag
      (set-message "Restart in order to let the new look and feel take effect"
		   "configure look and feel")))

(define (notify-language var val)
  (set-output-language val)
  (if (and (has-view?) (== (tree->object (the-buffer)) '(document "")))
      (init-language val))
  (cond ((or (== val "russian") (== val "ukrainian"))
	 (notify-preference "cyrillic input method"))))

(define (notify-cyrillic-input-method var val)
  (cond ((== val "translit")
	 (exec-file "$TEXMACS_PATH/progs/keyboard/russian" "translit.scm"))
	((== val "jcuken")
	 (exec-file "$TEXMACS_PATH/progs/keyboard/russian" "jcuken.scm"))
	((== val "yawerty")
	 (exec-file "$TEXMACS_PATH/progs/keyboard/russian" "yawerty.scm"))
	((== val "koi8-r")
	 (exec-file "$TEXMACS_PATH/progs/keyboard/russian" "koi8.scm"))
	((== val "cp1251")
	 (exec-file "$TEXMACS_PATH/progs/keyboard/russian" "cp1251.scm"))))

(define (notify-printing-command var val)
  (set-printing-command val))

(define (notify-paper-type var val)
  (set-printer-paper-type (locase-first val)))

(define (notify-printer-dpi var val)
  (set-printer-dpi val))

(define (notify-header var val)
  (show-header (== val "on")))

(define (notify-icon-bar var val)
  (cond ((== var "main icon bar")
	 (show-icon-bar 0 (== val "on")))
	((== var "context dependent icons")
	 (show-icon-bar 1 (== val "on")))
	((== var "user provided icons")
	 (show-icon-bar 2 (== val "on")))))

(define (notify-status-bar var val)
  (show-footer (== val "on")))

(define (notify-shrinking-factor var val)
  (set-default-shrinking-factor (string->number val))
  (set-shrinking-factor (string->number val)))

(define (notify-security var val)
  (cond ((== val "accept no scripts") (set-script-status 0))
	((== val "prompt on scripts") (set-script-status 1))
	((== val "accept all scripts") (set-script-status 2))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initialize preferences and consulting preferences
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define preferences-table (make-ahash-table))
(define preferences-call-back (make-ahash-table))
(define saved-preferences '())
(define user-preferences '())
(menu-bind preferences-menu)

(define (notify-preference var)
  (eval (list (get-call-back var) var (get-preference var))))

(define (get-preference-variant x) (list (car x) (get-preference (car x))))
(define (car-cadr l) (list (car l) (cadr l)))
(define (car-caddr l) (list (car l) (caddr l)))

(define (save-preferences)
  (set! user-preferences
	(map-in-order get-preference-variant default-preferences))
  (if (not (== user-preferences saved-preferences))
      (begin
	(save-expr "$TEXMACS_HOME_PATH/system/preferences.scm"
		   user-preferences)
	(set! saved-preferences user-preferences))))

(define (set-preference which what)
  (ahash-set! preferences-table which what)
  (eval (list (get-call-back which) which what))
  (save-preferences))

(define (get-preference which)
  (ahash-ref preferences-table which))

(define (test-preference? which what)
  (== what (get-preference which)))

(set-check-mark! set-preference "*" test-preference?)

(define (toggle-preference which)
  (let ((what (get-preference which)))
    (set-preference which (cond ((== what "on") "off")
				((== what "off") "on")
				(else what)))))

(define (preference-on? which)
  (test-preference? which "on"))

(set-check-mark! toggle-preference "v" preference-on?)

(define (get-call-back what)
  (let ((r (ahash-ref preferences-call-back what)))
    (if r r (lambda (var val) (noop)))))

(define (preference-apply l)
  ;(display* "TeXmacs] Applying '" (car l) "' preference\n")
  (eval (list (ahash-ref preferences-call-back (car l)) (car l) (cadr l))))

(define (preferences-extract l which flag)
  (cond ((null? l) l)
	((xor flag (in? (caar l) which))
	 (cons (car l) (preferences-extract (cdr l) which flag)))
	(else (preferences-extract (cdr l) which flag))))

(define preferences-initialization-flag #f)

(define (pre-initialize-preferences)
  (if (file-exists? "$TEXMACS_HOME_PATH/system" "preferences.scm")
      (set! saved-preferences
	    (load-expr "$TEXMACS_HOME_PATH/system/preferences.scm")))
  (fill-dictionary preferences-table (map-in-order car-cadr default-preferences))
  (fill-dictionary preferences-table saved-preferences)
  (fill-dictionary preferences-call-back
		   (map-in-order car-caddr default-preferences))
  (save-preferences)
  (map-in-order preference-apply
	     (preferences-extract user-preferences preinit-preferences #f)))

(define (initialize-preferences)
  (map-in-order preference-apply
	     (preferences-extract user-preferences preinit-preferences #t))
  (set! preferences-initialization-flag #t))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Computation of the preference menu
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (id-or-car x)
  (if (string? x) x (car x)))

(define (id-or-cadr x)
  (if (string? x) x (cadr x)))

(define (compute-preferences-entry s)
  `(interactive '(,(string-append (upcase-first s) ":"))
		'(lambda (val) (set-preference ,s val))))

(define (compute-preferences-enum s l)
  (cond ((null? l) l)
	((== (car l) '*)
	 (list '--- (list "Other" '... (compute-preferences-entry s))))
	(else (cons (list (id-or-car (car l))
			  `(set-preference ,s ,(id-or-cadr (car l))))
		    (compute-preferences-enum s (cdr l))))))

(define (compute-preferences-menu-sub l)
  (cond ((or (not (list? l)) (null? l)) l)
	((== (car l) 'string)
	 (let* ((x (cadr l))
		(s (id-or-car x))
		(v (id-or-cadr x)))
	   (list s '... (compute-preferences-entry v))))
	((== (car l) 'enum)
	 (let* ((x (cadr l))
		(s (id-or-car x))
		(v (id-or-cadr x)))
	   (cons* '-> s (compute-preferences-enum v (cddr l)))))
	((== (car l) 'toggle)
	 (let* ((x (cadr l))
		(s (id-or-car x))
		(v (id-or-cadr x)))
	   (list s (list 'toggle-preference v))))
	(else (map-in-order compute-preferences-menu-sub l))))

(define (compute-preferences-menu l)
  (eval `(menu-dynamic ,@(compute-preferences-menu-sub l))))
