/*
    chatview.cpp - Chat View

    Copyright (c) 2002      by Olivier Goffart       <ogoffart@tiscalinet.be>
    Copyright (c) 2002-2003 by Martijn Klingens      <klingens@kde.org>

    Kopete    (c) 2002-2003 by the Kopete developers <kopete-devel@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include <qfile.h>
#include <qlayout.h>
#include <qtimer.h>
#include <qtooltip.h>
#include <qclipboard.h>
#include <qcolor.h>

#include <dom/dom_doc.h>
#include <dom/dom_element.h>
#include <dom/html_document.h>
#include <dom/html_base.h>
#include <khtmlview.h>
#include <khtml_part.h>

#include <kiconloader.h>
#include <kapplication.h>
#include <krun.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kdialog.h>
#include <klocale.h>
#include <kpushbutton.h>
#include <kaction.h>
#include <kstdaccel.h>
#include <kstandarddirs.h>
#include <klistview.h>
#include <kdockwidget.h>
#include <kfontdialog.h>
#include <kcolordialog.h>
#include <ksqueezedtextlabel.h>
#include <kglobalsettings.h>
#include <krootpixmap.h>
#include <ktempfile.h>
#include <kshortcut.h>
#include <kiconeffect.h>

#include "chatview.h"
#include "kopetemessage.h"
#include "kopeteemoticons.h"
#include "kopeteprefs.h"
#include "kopetemetacontact.h"
#include "kopetemessagemanager.h"

#include "kopetechatwindow.h"

#include <kdeversion.h>
#if KDE_VERSION >= 306
  #include <ktextedit.h>
#else
  #include <qtextedit.h>
  #define KTextEdit QTextEdit
#endif

namespace Kopete {

ChatView::ChatView( KopeteMessageManager *mgr, KopeteChatWindow *parentWindow, QPixmap &icon, QWidget *parent, const char *name )
	 : KDockMainWindow( parent, name, 0L )
{
	//Create the view dock widget (KHTML Part), and set it to no docking (lock it in place)
	viewDock = createDockWidget( QString::fromLatin1( "viewDock" ), QPixmap(), 0L, QString::fromLatin1( "viewDock" ), QString::fromLatin1( " " ) );
	chatView = new KHTMLPart( this, "view" );

	//Security settings, we don't need this stuff
	chatView->setJScriptEnabled( false ) ;
	chatView->setJavaEnabled( false );
	chatView->setPluginsEnabled( false );
	chatView->setMetaRefreshEnabled( false );

	htmlWidget = chatView->view();
	viewDock->setWidget(htmlWidget);
	viewDock->setDockSite(KDockWidget::DockBottom);
	viewDock->setEnableDocking(KDockWidget::DockNone);

	//Create the botttom dock widget, with the edit area, statusbar and send button
	editDock = createDockWidget( QString::fromLatin1( "editDock" ), QPixmap(), 0L, QString::fromLatin1( "editDock" ), QString::fromLatin1( " " ) );
	QWidget *containerWidget = new QWidget(this);
	QVBoxLayout *v = new QVBoxLayout( containerWidget );
	m_edit = new KTextEdit( containerWidget, "m_edit" );

	//Set params on the edit widget
	m_edit->setMinimumSize( QSize( 75, 20 ) );
	m_edit->setWordWrap( QTextEdit::WidgetWidth );
	m_edit->setTextFormat( Qt::PlainText );
	m_edit->setWrapPolicy( QTextEdit::AtWhiteSpace );
	m_edit->installEventFilter( this );
	v->addWidget( m_edit );

	//Layout for status label and send button
	QHBoxLayout *h = new QHBoxLayout( 0, 4, 4 );

	//Status label
	m_status = new KSqueezedTextLabel( i18n("Ready."), containerWidget );
	QToolTip::add( m_status, m_status->text() );
	h->addWidget( m_status, 0, Qt::AlignLeft | Qt::AlignVCenter );

	//Send Button
	m_button_send = new KPushButton( i18n("Send"), containerWidget );
	m_button_send->setEnabled( false );
	h->addWidget(m_button_send, 0, Qt::AlignRight | Qt::AlignVCenter);
	v->addLayout( h );

	//Make the edit area dockable for now
	editDock->setWidget( containerWidget );
	editDock->setDockSite( KDockWidget::DockNone );
	editDock->setEnableDocking(KDockWidget::DockBottom);

	//Set the view as the main widget
	setMainDockWidget( viewDock );
	setView(viewDock);

	// some signals and slots connections
	connect( m_button_send, SIGNAL( pressed() ), this, SLOT( sendMessage() ) );
	connect( m_edit, SIGNAL( textChanged()), this, SLOT( slotTextChanged() ) );
	connect( KopetePrefs::prefs(), SIGNAL(saved()), this, SLOT( slotConfigChanged() ) );

	// Timers for typing notifications
	m_typingRepeatTimer = new QTimer;
	m_typingStopTimer   = new QTimer;

	m_remoteTypingMap.setAutoDelete( true );

	m_blinkTimer = new QTimer;

	connect( m_typingRepeatTimer, SIGNAL( timeout() ), SLOT( slotRepeatTimer() ) );
	connect( m_typingStopTimer,   SIGNAL( timeout() ), SLOT( slotStopTimer() ) );

	connect( m_blinkTimer,   SIGNAL( timeout() ), SLOT( slotBlinkTimer() ) );

	// Some settings for the HTML part
	htmlWidget->setMarginWidth(4);
	htmlWidget->setMarginHeight(4);
	connect ( chatView->browserExtension(), SIGNAL( openURLRequestDelayed( const KURL &, const KParts::URLArgs & ) ),
		SLOT( slotOpenURLRequest( const KURL &, const KParts::URLArgs & ) ) );
	htmlWidget->setFocusPolicy( NoFocus );
	setFocusProxy( m_edit );
	m_edit->setFocus();

	// Finalize
	historyPos = -1;
	m_mainWindow = parentWindow;
	membersDock = 0L;
	m_manager = mgr;
	backgroundFile = QString::null;
	root = 0L;
	slotConfigChanged();
	isActive = false;
	blinkState = false;
	m_tabBar = 0L;
	m_icon = icon;
	m_iconLight = KIconEffect().apply( m_icon, KIconEffect::ToGamma, 0.5, QColor( QString::fromLatin1("white") ), true );

	setCaption( m_manager->chatName(), false );

	//Restore docking positions
	readOptions();

	//Show chat members
	createMembersList();
}

ChatView::~ChatView()
{
	kdDebug( 14010 ) << k_funcinfo << "Destroying ChatView" << endl;

	saveOptions();

	delete m_typingRepeatTimer;
	delete m_typingStopTimer;
	delete m_blinkTimer;
}

void ChatView::setMainWindow( KopeteChatWindow* parent )
{
	m_mainWindow = parent;
	if( root )
	{
		disconnect(root, SIGNAL(backgroundUpdated(const QPixmap &)), this, SLOT(slotUpdateBackground(const QPixmap &)));
		delete root;
		root = 0L;
		slotConfigChanged();
	}
}

void ChatView::createMembersList(void)
{
	if( !membersDock )
	{
		kdDebug(14010) << k_funcinfo << "Creating members list" << endl;

		//Create the chat members list
		membersDock = createDockWidget( QString::fromLatin1( "membersDock" ), QPixmap(), 0L,
			QString::fromLatin1( "membersDock" ), QString::fromLatin1( " " ) );
		membersList = new KListView(this);
		membersList->addColumn( i18n("Chat Members"), -1 );

		//Add the contacts that are in the message manager
		KopeteContact *contact;
		KopeteContactPtrList chatMembers = m_manager->members();

		for ( contact = chatMembers.first(); contact; contact = chatMembers.next() )
			contactAdded( contact );

		membersDock->setWidget(membersList);

		//Dock the chatmembers list. If this is a group chat, show it always initially
		if( (m_manager->members()).count() > 1 && membersDockPosition == KDockWidget::DockNone)
			placeMembersList( KDockWidget::DockRight );
		else
			placeMembersList( membersDockPosition );

		//Connect the popup menu
		connect( membersList, SIGNAL( contextMenu( KListView*, QListViewItem *, const QPoint &) ),
			SLOT( slotContactsContextMenu(KListView*, QListViewItem *, const QPoint & ) ) );
	}
}

void ChatView::placeMembersList( KDockWidget::DockPosition dp )
{
	membersDockPosition = dp;
	int dockWidth;
	if( dp == KDockWidget::DockNone )
	{
		//Dock it to the desktop then hide it
		membersDock->undock();
  		membersDock->hide();
	}
	else
	{
		if( dp == KDockWidget::DockRight )
			dockWidth = 70;
		else
			dockWidth = 30;
		//Make sure it is shown then place it wherever
		membersDock->setEnableDocking( KDockWidget::DockLeft | KDockWidget::DockRight );
		membersDock->manualDock( viewDock, dp, dockWidth );
		membersDock->show();
		membersDock->setEnableDocking( KDockWidget::DockNone );
	}

	refreshView();
}

void ChatView::slotContactsContextMenu( KListView*, QListViewItem *item, const QPoint &point )
{
	if( item )
	{
		KopeteContactLVI *contactLVI = dynamic_cast<KopeteContactLVI*>( item );
		((KopeteContact*)contactLVI->contact())->showContextMenu( point );
	}
}

void ChatView::anyTyping( const KopeteContact *c, bool isTyping )
{
	// Ensure this contact is in the typing map
	// Strictly speaking the below code does that, but contactAdded() does
	// some additional bookkeeping, hence this call
	if( !typingMap.contains( c ) )
		contactAdded( c );

	// Set his typing status
	typingMap[ c ] = isTyping;

	// Make sure we (re-)add the timer at the end, because the slot will
	// remove the first timer
	// And yes, the const_cast is a bit ugly, but it's only used as key
	// value in this dictionary (no indirections) so it's basically
	// harmless. Unfortunately there's no QConstPtrDictionary in Qt...
	void *key = const_cast<KopeteContact *>( c );
	m_remoteTypingMap.remove( key );
	if( isTyping )
	{
		m_remoteTypingMap.insert( key, new QTimer );
		connect( m_remoteTypingMap[ key ], SIGNAL( timeout() ), SLOT( slotRemoteTypingTimeout() ) );
		m_remoteTypingMap[ key ]->start( 6000, true );
	}

	// Loop through the map, constructing a string of people typing
	QStringList typingList;
	QString statusTyping;
	QMap<const KopeteContact*, bool>::Iterator it;
	for( it = typingMap.begin(); it != typingMap.end(); ++it )
	{
		// FIXME: is it really possible to have null pointers here? The check
		// seems unneeded to me - Martijn
		if( !it.data() )
			continue;

		typingList.append( it.key()->metaContact() ? it.key()->metaContact()->displayName() : it.key()->displayName() );
	}
	statusTyping = typingList.join( QString::fromLatin1( ", " ) );

	// Update the status area
	if( !typingList.isEmpty() )
		setStatus ( i18n( "%1 is typing a message...", "%1 are typing a message...", typingList.count() ).arg( statusTyping ) );
	else
		setStatus ( i18n( "Ready." ) );
}

void ChatView::contactChanged()
{
	//This fires whenever a contact or MC changes displayName, so only
	//update the caption if it changed to avioud unneeded updates that
	//could cause flickering
	QString chatName = m_manager->chatName();
	if( chatName != m_captionText )
		setCaption( chatName, true );
}

void ChatView::contactAdded(const KopeteContact *c)
{
	if( !memberContactMap.contains(c) && (c != m_manager->user()) )
	{
		QString contactName;
		if( c->metaContact() )
		{
			contactName = c->metaContact()->displayName();
			connect( c->metaContact(), SIGNAL(displayNameChanged(KopeteMetaContact*, const QString &)), this, SLOT(contactChanged()));
		}
		else
		{
			contactName = c->displayName();
			connect( c, SIGNAL(displayNameChanged()), this, SLOT(contactChanged()));
		}

		connect( c, SIGNAL(statusChanged(KopeteContact *, KopeteContact::ContactStatus)), this, SLOT(contactStatusChanged(KopeteContact *, KopeteContact::ContactStatus)) );
		typingMap.insert(c,false);

		if( memberContactMap.count() > 0 )
		{
			KopeteMessage m = KopeteMessage(c, m_manager->members(), i18n("%1 has joined the chat.").arg( contactName ), KopeteMessage::Internal );
			addChatMessage( m );
		}
		memberContactMap.insert(c, new KopeteContactLVI( c, membersList ) );
	}
}

void ChatView::contactRemoved(const KopeteContact *c)
{
	if( memberContactMap.contains(c) && (c != m_manager->user()) )
	{
		disconnect( c, SIGNAL(statusChanged(KopeteContact *, KopeteContact::ContactStatus)), this, SLOT(contactStatusChanged(KopeteContact *, KopeteContact::ContactStatus)) );
		typingMap.remove(c);
		m_remoteTypingMap.remove( const_cast<KopeteContact *>( c ) );

		QString contactName;
		if( c->metaContact() )
		{
			contactName = c->metaContact()->displayName();
			disconnect( c->metaContact(), SIGNAL(displayNameChanged(KopeteMetaContact*, const QString &)), this, SLOT(contactChanged()));
		}
		else
		{
			contactName = c->displayName();
			disconnect( c, SIGNAL(displayNameChanged()), this, SLOT(contactChanged()));
		}

		KopeteMessage m = KopeteMessage(c, m_manager->members(), i18n("%1 has left the chat.").arg( contactName ), KopeteMessage::Internal);
		addChatMessage( m );

		membersList->takeItem( memberContactMap[c] );
		memberContactMap.remove(c);
	}
}

void ChatView::setCaption( const QString &text, bool modified )
{
	QString newCaption = text;

	//Save this caption
	m_captionText = text;

	//Turncate if needed
	if( newCaption.length() > 20 )
		newCaption = newCaption.left( 20 ).append( QString::fromLatin1( "..." ) );

	//Call the original set caption
	KDockMainWindow::setCaption( newCaption, false );

	if( m_tabBar )
	{
		m_tabBar->setTabToolTip( this, m_captionText );

		//Blink icon if modified and not active
		if( !isActive && modified )
		{
			/*if( !m_blinkTimer->isActive() )
				m_blinkTimer->start( 750, false );*/
			m_tabBar->setTabLabel( this, newCaption + QString::fromLatin1("*") );
		}
		else
		{
			/*m_blinkTimer->stop();
			blinkState = false;
			m_tabBar->setTabIconSet( this, m_icon );*/
			m_tabBar->setTabLabel( this, newCaption );
		}
	}

	//Tell the parent we changed our caption
	emit( captionChanged( isActive ) );
}

void ChatView::slotBlinkTimer()
{
	if( m_tabBar )
	{
		if( blinkState )
			m_tabBar->setTabIconSet( this, m_icon );
		else
			m_tabBar->setTabIconSet( this, m_iconLight );
		blinkState = !blinkState;
	}
}

void ChatView::messageReceived(const KopeteMessage &message)
{
	anyTyping( message.from(), false );
	//Need to copy this because it comes in as a const
	KopeteMessage m = message;
	addChatMessage(m);
}

void ChatView::contactStatusChanged(KopeteContact *contact, KopeteContact::ContactStatus )
{
	//kdDebug(14010) << "slotContactStatusChanged" << endl;
	// %2 before %1 because displayName can contains '%' . And i don't think the status can
	if(KopetePrefs::prefs()->showEvents())
	{
		KopeteMessage m;
		if (contact->metaContact())
			m = KopeteMessage(contact, m_manager->members(), i18n("%2 changed status to %1.").arg(contact->statusText()).arg(contact->metaContact()->displayName()), KopeteMessage::Internal);
		else
			m = KopeteMessage(contact, m_manager->members(), i18n("%2 changed status to %1.").arg(contact->statusText()).arg(contact->displayName()), KopeteMessage::Internal);
		addChatMessage( m );
	}
	contactChanged();
}

void ChatView::slotOpenURLRequest(const KURL &url, const KParts::URLArgs &/*args*/)
{
	kdDebug(14010) << k_funcinfo << "url=" << url.url() << endl;
	new KRun(url, 0, false); // false = non-local files
}

void ChatView::sendMessage()
{
	m_button_send->setEnabled(false);
	emit SendMessage( currentMessage() );
	historyList.prepend(m_edit->text());
	historyPos = -1;
	m_edit->setText( QString::null );
	slotStopTimer();
}

void ChatView::slotTextChanged()
{
	bool canSend = !m_edit->text().isEmpty();
	m_button_send->setEnabled( canSend );
	m_mainWindow->chatSend->setEnabled( canSend );

	// And they were previously typing
	if( !m_typingRepeatTimer->isActive() )
	{
		m_typingRepeatTimer->start( 4000, false );
		slotRepeatTimer();
	}

	// Reset the stop timer again, regardless of status
	m_typingStopTimer->start( 4000, true );
}

bool ChatView::eventFilter(QObject *w, QEvent *e)
{
	if (e->type() == QEvent::KeyPress)
	{
		QKeyEvent *event = (QKeyEvent*) e;
		KKey kkeyEvent = KKey( event );
		if( m_mainWindow->chatSend->isEnabled() && m_mainWindow->chatSend->shortcut().contains( kkeyEvent ) )
		{
			sendMessage();
			return true;
		}
		else if( m_mainWindow->historyUp->shortcut().contains( kkeyEvent ) )
		{
			historyUp();
			return true;
		}
		else if( m_mainWindow->historyDown->shortcut().contains( kkeyEvent ) )
		{
			historyDown();
			return true;
		}
		else if( KStdAccel::copy().contains( kkeyEvent ) )
		{
			//Our keyboard shortcut was pressed. Copy the text
			copy();
			return true;
		}
	}

	return QWidget::eventFilter(w, e);
}

void ChatView::historyUp()
{
	historyPos++;
	if(historyPos < (int)historyList.count())
		m_edit->setText(historyList[historyPos]);
	else
		historyPos = historyList.count()-1;
}

void ChatView::historyDown()
{
	if(historyPos > 0)
	{
		historyPos--;
		m_edit->setText(historyList[historyPos]);
	}
}

void ChatView::saveOptions()
{
	kdDebug(14010) << k_funcinfo << endl;

	KConfig *config = KGlobal::config();

	writeDockConfig ( config, QString::fromLatin1("ChatViewDock") );
	config->setGroup( QString::fromLatin1("ChatViewDock") );
	config->writeEntry( QString::fromLatin1("membersDockPosition"), membersDockPosition );

	config->setGroup( QString::fromLatin1("ChatViewSettings") );

	config->writeEntry ( QString::fromLatin1("Font"), mFont );
	config->writeEntry ( QString::fromLatin1("TextColor"), mFgColor );
	config->writeEntry ( QString::fromLatin1("BackgroundColor"), mBgColor );

	//config->writeEntry ( "SplitterWidth", editDock->parent()->seperatorPos() );

	config->sync();
}

void ChatView::readOptions()
{
	kdDebug(14010) << k_funcinfo << endl;

	KConfig *config = KGlobal::config();

	/** THIS IS BROKEN !!! */
	//dockManager->readConfig ( config, "ChatViewDock" );

	//Work-around to restore dock widget positions
	config->setGroup( QString::fromLatin1("ChatViewDock") );
	int splitterPos = config->readNumEntry( QString::fromLatin1("viewDock,editDock:sepPos"), 70);
	editDock->manualDock( viewDock, KDockWidget::DockBottom, splitterPos );
	viewDock->setDockSite(KDockWidget::DockLeft | KDockWidget::DockRight );
	editDock->setEnableDocking(KDockWidget::DockNone);
	membersDockPosition = static_cast<KDockWidget::DockPosition>(
		config->readNumEntry( QString::fromLatin1("membersDockPosition"), KDockWidget::DockNone ) );

	config->setGroup( QString::fromLatin1("ChatViewSettings") );

	QFont tmpFont = KGlobalSettings::generalFont();
	setFont( config->readFontEntry( QString::fromLatin1("Font"), &tmpFont) );

	QColor tmpColor = KGlobalSettings::textColor();
	setFgColor( config->readColorEntry ( QString::fromLatin1("TextColor"), &tmpColor ) );

	tmpColor = KGlobalSettings::baseColor();
	setBgColor( config->readColorEntry ( QString::fromLatin1("BackgroundColor"), &tmpColor) );
	//editDock->parent()->setSeperatorPos( config->readNumEntry ( "SplitterWidth", 70 ) );
}

void ChatView::setStatus(const QString &text)
{
	m_status->setText(text);
	if ( text != QToolTip::textFor( m_status ) )
		QToolTip::add( m_status, text );
}

void ChatView::addText(const QString &text)
{
	m_edit->insert(text);
}

void ChatView::addChatMessage( KopeteMessage &m )
{
	const QString model = KopetePrefs::prefs()->kindMessagesHtml();
	if( transparencyEnabled )
		m.setBgOverride( bgOverride );
	QString resultHTML = m.transformMessage( model );

	viewContents.append( resultHTML );

	refreshView();
	
	/*This crashes in KHTML < HEAD for some reason
	DOM::HTMLElement newNode = chatView->document().createElement( QString::fromLatin1("div") );
	newNode.setInnerHTML( resultHTML );
	chatView->htmlDocument().body().appendChild( newNode );
	QTimer::singleShot( 1, this, SLOT( slotScrollView() ) );
	*/

	setCaption( m_captionText, true );
}

void ChatView::setActive( bool value )
{
	isActive = value;
	if( isActive )
		setCaption( m_captionText, false );

}

void ChatView::setTabBar( QTabWidget *tabBar )
{
	m_tabBar = tabBar;
}

void ChatView::refreshView()
{
	chatView->begin();
	if( !backgroundFile.isNull() )
	{
		chatView->write( QString::fromLatin1( "<html><body background=\"" ) + backgroundFile +
			QString::fromLatin1( "\" bgproperties=\"fixed\">" ) + viewContents +
			QString::fromLatin1( "</body></html>" ) );
	}
	else
	{
		chatView->write( viewContents );
	}
	chatView->end();

	QTimer::singleShot( 1, this, SLOT( slotScrollView() ) );
}

void ChatView::slotScrollView()
{
	chatView->view()->scrollBy( 0, chatView->view()->contentsHeight() );
}

void ChatView::setCurrentMessage( const KopeteMessage &message )
{
	m_edit->setText( message.plainBody() );
	setFont( message.font() );
	setBgColor( message.bg() );
	setFgColor( message.fg() );
}

KopeteMessage ChatView::currentMessage()
{
	KopeteMessage currentMsg = KopeteMessage( m_manager->user(), m_manager->members(), m_edit->text(), KopeteMessage::Outbound, KopeteMessage::PlainText );

	currentMsg.setFont( mFont );
	currentMsg.setBg( mBgColor );
	currentMsg.setFg( mFgColor );

	return currentMsg;
}

void ChatView::cut()
{
	m_edit->cut();
}

void ChatView::copy()
{
	kdDebug(14010) << k_funcinfo << endl;

	if ( chatView->hasSelection() )
		QApplication::clipboard()->setText( chatView->selectedText() );
	else
		m_edit->copy();
}

void ChatView::paste()
{
	m_edit->paste();
}

void ChatView::print()
{
	chatView->view()->print();
}

void ChatView::selectAll()
{
	chatView->selectAll();
}

void ChatView::setFont()
{
	KFontDialog::getFont(mFont, false, this);
	m_edit->setFont(mFont);
}

void ChatView::setFont( const QFont &newFont )
{
	mFont = newFont;
	m_edit->setFont(mFont);
}

void ChatView::setFgColor( const QColor &newColor )
{
	if( newColor == QColor() )
		KColorDialog::getColor( mFgColor, this );
	else
		mFgColor = newColor;

	QPalette pal = m_edit->palette();
	pal.setColor(QPalette::Active, QColorGroup::Text, mFgColor );
	pal.setColor(QPalette::Inactive, QColorGroup::Text, mFgColor );

	// unsetPalette() so that color changes in kcontrol are honoured
	// if we ever have a subclass of KTextEdit, reimplement setPalette()
	// and check it there.
	if ( pal == QApplication::palette( m_edit ) )
		m_edit->unsetPalette();
	else
		m_edit->setPalette(pal);
}

void ChatView::setBgColor( const QColor &newColor )
{
	if( newColor == QColor() )
		KColorDialog::getColor( mBgColor, this );
	else
		mBgColor = newColor;

	QPalette pal = m_edit->palette();
	pal.setColor(QPalette::Active, QColorGroup::Base, mBgColor );
	pal.setColor(QPalette::Inactive, QColorGroup::Base, mBgColor );
	pal.setColor(QPalette::Disabled, QColorGroup::Base, mBgColor );

	// unsetPalette() so that color changes in kcontrol are honoured
	// if we ever have a subclass of KTextEdit, reimplement setPalette()
	// and check it there.
	if ( pal == QApplication::palette( m_edit ) )
		m_edit->unsetPalette();
	else
		m_edit->setPalette(pal);
}

void ChatView::slotRepeatTimer()
{
	emit Typing( true );
}

void ChatView::slotRemoteTypingTimeout()
{
	kdDebug() << k_funcinfo << endl;

	// Remove the topmost timer from the list. Why does QPtrDict use void* keys and not typed keys? *sigh*
	if( !m_remoteTypingMap.isEmpty() )
		anyTyping( reinterpret_cast<const KopeteContact *>( QPtrDictIterator<QTimer>( m_remoteTypingMap ).currentKey() ), false );
}

void ChatView::slotStopTimer()
{
	kdDebug() << k_funcinfo << endl;

	m_typingRepeatTimer->stop();
	emit Typing( false );
}

void ChatView::slotConfigChanged()
{
//	kdDebug(14010) << k_funcinfo << "called." << endl;

	transparencyEnabled = KopetePrefs::prefs()->transparencyEnabled();
	bgOverride = KopetePrefs::prefs()->bgOverride();

//	kdDebug(14010) << k_funcinfo << "transparencyEnabled=" << transparencyEnabled << ", bgOverride=" << bgOverride << "." << endl;

	if( transparencyEnabled )
	{
		if( root == 0L )
		{
//			kdDebug(14010) << k_funcinfo << "enabling transparency" << endl;
			root = new KRootPixmap( htmlWidget );
			connect(root, SIGNAL(backgroundUpdated(const QPixmap &)), this, SLOT(slotUpdateBackground(const QPixmap &)));
			root->setCustomPainting ( true );
			root->setFadeEffect(KopetePrefs::prefs()->transparencyValue() * 0.01, KopetePrefs::prefs()->transparencyColor() );
			root->start();
		} else {
			root->setFadeEffect(KopetePrefs::prefs()->transparencyValue() * 0.01, KopetePrefs::prefs()->transparencyColor() );
		}
		root->repaint( true );
	}
	else
	{
		if ( root != 0L )
		{
//			kdDebug(14010) << k_funcinfo << "disabling transparency" << endl;
			disconnect(root, SIGNAL(backgroundUpdated(const QPixmap &)), this, SLOT(slotUpdateBackground(const QPixmap &)));
			delete root;
			root = 0L;
			backgroundFile = QString();
			refreshView();
		}
	}
}

void ChatView::slotUpdateBackground(const QPixmap &pm)
{
	if( isActive )
	{
		m_mainWindow->updateBackground( pm );
	}

	if( m_mainWindow->backgroundFile )
		backgroundFile = m_mainWindow->backgroundFile->name();
	refreshView();
}

//-----------------------
//  KopeteContactLVI

KopeteContactLVI::KopeteContactLVI( const KopeteContact *contact, KListView *parent ) : KListViewItem( parent )
{
	m_contact = contact;
	m_parentView = parent;

	connect( m_contact, SIGNAL( displayNameChanged( const QString & ) ),
		this, SLOT( slotDisplayNameChanged() ) );
	connect( m_contact, SIGNAL( statusChanged( KopeteContact *, KopeteContact::ContactStatus ) ),
		this, SLOT( slotStatusChanged() ) );

	slotStatusChanged();
	slotDisplayNameChanged();
}

void KopeteContactLVI::slotDisplayNameChanged()
{
	if( m_contact->metaContact() )
		setText( 0, m_contact->metaContact()->displayName().prepend( QString::fromLatin1( " " ) ) );
	else
		setText( 0, m_contact->displayName().prepend( QString::fromLatin1( " " )) );

	m_parentView->sort();
}


void KopeteContactLVI::slotStatusChanged()
{
	setPixmap( 0, UserIcon( m_contact->statusIcon() ) );
}

} //end namespace Kopete

#include "chatview.moc"

// vim: set noet ts=4 sts=4 sw=4:
