/*
    chatview.h - Chat View

    Copyright (c) 2002      by Olivier Goffart       <ogoffart@tiscalinet.be>

    Kopete    (c) 2002-2003 by the Kopete developers <kopete-devel@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#ifndef __CHATVIEW_H__
#define __CHATVIEW_H__

#include <qptrdict.h>
#include <qstringlist.h>

#include <kopetecontact.h>
#include <kdockwidget.h>
#include <klistview.h>

class QPixmap;
class QTabWidget;
class QTextEdit;
class QTimer;

class KHTMLPart;
class KHTMLView;
class KPushButton;
class KRootPixmap;
class KSqueezedTextLabel;

class KopeteChatWindow;
class KopeteMessageManager;

typedef QPtrList<KopeteContact> KopeteContactPtrList;
class KURL;

namespace KParts { struct URLArgs; }

class KopeteMessage;

namespace Kopete
{

struct ChatViewPrivate;

/**
 * @author Olivier Goffart
 */
class KopeteContactLVI : public QObject, public KListViewItem
{
	Q_OBJECT

	// FIXME: Why is this class duplicated and isn't the 'normal' version
	// from the contact list used? - Martijn

public:
	KopeteContactLVI( const KopeteContact *contact, KListView *parent );
	const KopeteContact *contact() { return m_contact; }

private:
	// FIXME: const is ok but then aboves contact()
	// has to return a const KopeteContact* as well
	const KopeteContact *m_contact;
	KListView *m_parentView;

private slots:
	void slotDisplayNameChanged();
	void slotStatusChanged();
};

/**
 * @author Olivier Goffart
 */
class ChatView : public KDockMainWindow
{
	Q_OBJECT

public:
	ChatView( KopeteMessageManager *manager, KopeteChatWindow *parentWindow, QPixmap &icon,
		QWidget *parent = 0, const char *name = 0 );
	~ChatView();

	/**
	 * Adds text into the edit area. Used when you select an emotiocon
	 * @param text The text to be inserted
	 */
	void addText( const QString &text );

	/**
	 * Saves window settings such as splitter positions
	 */
	void saveOptions();

	/**
	 * Sets the text to be displayed on the status label
	 * @param text The text to be displayed
	 */
	void setStatus( const QString &text );

	/**
	 * Tells this view if it is the active view
	 */
	void setActive( bool value );

	/**
	 * Sets the text to be displayed on tab label and window caption
	 */
	virtual void setCaption( const QString &text, bool modified );

	/**
	 * Changes the pointer to the chat widnow. Used to re-parent the view
	 * @param parent The new chat window
	 */
	void setMainWindow( KopeteChatWindow* parent );

	/**
	 * Sets the current message in the chat window
	 * @param parent The new chat window
	 */
	void setCurrentMessage( const KopeteMessage &newMessage );

	void setTabBar( QTabWidget *tabBar );

	/**
	 * Sets the placement of the members contact list.
	 * DockLeft, DockRight, or DockNone.
	 * @param dp The dock position of the list
	 * @param width The % width the left hand widget should take up
	 */
	void placeMembersList( KDockWidget::DockPosition dp = KDockWidget::DockRight );

	/**
	 * Returns the message currently in the edit area
	 * @return The message
	 */
	KopeteMessage currentMessage();

	/**
	 * Returns the KMM associated with this chat view
	 * @return The KMM
	 */
	KopeteMessageManager *manager() { return m_manager; }

	/**
	 * Returns the chat window this view is in
	 * @return The chat window
	 */
	KopeteChatWindow *mainWindow() { return m_mainWindow; }

	/**
	 * Returns the current position of the chat member slist
	 * @return The position of the chat members list
	 */
	const KDockWidget::DockPosition membersListPosition()  { return membersDockPosition; }

	/**
	 * Returns the HTML contents of the KHTML widget
	 * @return The contents of the view
	 */
	const QString viewsText() { return viewContents; }

	const bool docked() { return ( m_tabBar != 0L ); }

	const QString &caption() { return m_captionText; }

	void print();

	void historyUp();

	void historyDown();

public slots:
	/**
	 * Initiates a cut action on the edit area of the chat view
	 */
	void cut();

	/**
	 * Initiates a copy action
	 * If there is text selected in the HTML view, that text is copied
	 * Otherwise, the entire edit area is copied.
	 */
	void copy();

	/**
	 * Initiates a paste action into the edit area of the chat view
	 */
	void paste();

	/**
	 * Selects all text in the chat view
	 */
	void selectAll();

	/**
	 * Sets the background color of the entry area, and outgoing messages
	 * @param newColor The new background color. If this is QColor(), then
	 * a color chooser dialog is opened
	 */
	void setBgColor( const QColor &newColor = QColor() );

	/**
	 * Sets the foreground color of the entry area, and outgoing messages
	 * @param newColor The new foreground color. If this is QColor(), then
	 * a color chooser dialog is opened
	 */
	void setFgColor( const QColor &newColor = QColor() );

	/**
	 * Sets the font of the edit area and outgoign messages. Opens a font chooser dialog.
	 */
	void setFont();

	/**
	 * Sets the font of the edit area and outgoing messages to the specified value.
	 * @param newFont The new font to use.
	 */
	void setFont( const QFont &newFont );

	/**
	 * Sends the text currently entered into the edit area
	 */
	void sendMessage();

	/**
	 * Appends a message to the chat view display area
	 * @param message The message to be appended
	 */
	void addChatMessage( KopeteMessage &message );

	/**
	 * Called when a contact is added to the KMM instance (A new person joins the chat).
	 * Adds this contact to the typingMap and the contact list view
	 * @param c The contact that joined the chat
	 */
	void contactAdded( const KopeteContact *c );

	/**
	 * Called when a contact is removed from the KMM instance (A person left the chat).
	 * Removes this contact from typingMap and the contact list view
	 * @param c The contact left the chat
	 */
	void contactRemoved( const KopeteContact *c );

	/**
	 * Called when a contact changes status or displayName, updates the contact list view and
	 * window caption
	 */
	void contactChanged();

	/**
	 * Called when a contact changes status, updates the contact list view and
	 * @param contact The contact who changed status
	 * @param status The new status of the contact
	 */
	void contactStatusChanged( KopeteContact *contact, KopeteContact::ContactStatus status );

	/**
	 * Called when a message is recieved from someone
	 * @param message The message recieved
	 */
	void messageReceived( const KopeteMessage &message );

	/**
	 * Called when a typing event is recieved from a contact
	 * Updates the typing map and outputs the typing message into the status area
	 * @param contact The contact who is / isn't typing
	 * @param typing If the contact is typing now
	 */
	void anyTyping( const KopeteContact *contact, bool typing );

signals:
	/**
	 * Emits when a message is sent
	 * @param message The message sent
	 */
	void SendMessage( const KopeteMessage &message );

	/**
	 * Emits every 4 seconds while the user is typing
	 * @param bool if the user is typing right now
	 */
	void Typing( bool );

	/**
	 * Emits when the chat view is shown
	 */
	void Shown();

	void captionChanged( bool active );

protected:
	/**
	 * This event filter captures the keypress events in the edit area to check for
	 * CTRL+Enter, SHIFT+Enter, etc.
	 */
	virtual bool eventFilter( QObject *watched, QEvent *e );

private slots:
	void slotOpenURLRequest( const KURL &url, const KParts::URLArgs &args );
	void slotContactsContextMenu( KListView*, QListViewItem *item, const QPoint &point );
	void slotRepeatTimer();
	void slotRemoteTypingTimeout();
	void slotBlinkTimer();
	void slotScrollView();

	/**
	 * Called when text is changed in the edit area
	 */
	void slotTextChanged();

	void slotStopTimer();

	/**
	 * Called when KopetePrefs are saved
	 */
	void slotConfigChanged();

	/**
	 * Sets the background of the widget
	 * @param pm The new background image
	 */
	void slotUpdateBackground( const QPixmap &pm );

private:
	QPtrDict<QTimer> m_remoteTypingMap;
	KSqueezedTextLabel *m_status;
	KHTMLPart *chatView;
	KHTMLView *htmlWidget;
	bool isTyping;
	QStringList historyList;
	int historyPos;
	QString viewContents;
	QMap<const KopeteContact *, bool> typingMap;
	QMap<const KopeteContact *, KopeteContactLVI *> memberContactMap;
	QTextEdit* m_edit;
	KPushButton* m_button_send;
	QFont mFont;
	QColor mFgColor;
	QColor mBgColor;
	KListView *membersList;
	bool transparencyEnabled;
	bool bgOverride;
	bool isActive;
	bool blinkState;
	QString backgroundFile;
	QPixmap m_icon;
	QPixmap m_iconLight;

	KRootPixmap *root;
	KDockWidget *viewDock;
	KDockWidget *membersDock;
	KDockWidget *editDock;
	QTabWidget *m_tabBar;
	KDockWidget::DockPosition membersDockPosition;

	KopeteChatWindow *m_mainWindow;
	KopeteMessageManager *m_manager;

	QTimer *m_typingRepeatTimer;
	QTimer *m_typingStopTimer;
	QTimer *m_blinkTimer;

	/**
	 * Holds the string to be displayed on the tabs and window title
	 */
	QString m_captionText;

	/**
	 * Creates the members list widget
	 */
	void createMembersList();

	/**
	 * Read in saved options, such as splitter positions
	 */
	void readOptions();

	/**
	 * Refreshes the cuat view. Used to update the background.
	 */
	void refreshView();
};

} // namespace Kopete

#endif

// vim: set noet ts=4 sts=4 sw=4:

