/*
    icqprotocol.h  -  ICQ Protocol Plugin

    Copyright (c) 2002 by Nick Betcher <nbetcher@kde.org>

    Kopete    (c) 2002 by the Kopete developers  <kopete-devel@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/

#include <fstream>

#include <qapplication.h>
#include <qcursor.h>
#include <qtimer.h>

#include <kconfig.h>
#include <kdebug.h>
#include <kgenericfactory.h>
#include <kglobal.h>
#include <kiconloader.h>
#include <kmessagebox.h>
#include <ksimpleconfig.h>
#include <kstandarddirs.h>
#include <kpopupmenu.h>

#include "icqprotocol.h"
#include "icqaddcontactpage.h"
#include "icqcontact.h"
#include "icqchangestatus.h"
#include "icqadd.h"
#include "icquserinfo.h"
#include "icqpreferences.h"
#include "icqfiletransfer.h"

#include "kopeteaway.h"
#include "kopetestdaction.h"
#include "kopetemetacontact.h"
#include "kopetecontactlist.h"
#include "kopetetransfermanager.h"

#include "client.h"
#include "libicq/log.h"

#define ICON_OFFLINE "icq_offline"
#define ICON_DND "icq_dnd"
#define ICON_OCC "icq_occupied"
#define ICON_NA  "icq_na"
#define ICON_AWAY "icq_away"
#define ICON_ONLINE "icq_online"
#define ICON_INVISIBLE "icq_invisible"
#define ICON_CONNECTING "icq_connecting"


K_EXPORT_COMPONENT_FACTORY( kopete_icq, KGenericFactory<ICQProtocol> );

static void logMonitor(unsigned short l, const char *s)
{
	if ((l & (L_ERROR|L_WARN/*|L_DEBUG|L_PACKET*/)) == 0) return;

	switch (l)
	{
		case L_ERROR:
			kdError(14110) << s << endl;
			break;
		case L_WARN:
			kdWarning(14110) << s << endl;
			break;
		case L_DEBUG:
			kdDebug(14110) << s << endl;
			break;
		case L_PACKET:
			kdDebug(14111) << s << endl;
			break;
		default:
			kdDebug(14110) << s << endl;
			break;
	}
}

ICQProtocol::ICQProtocol(QObject *parent, const char *name, const QStringList & /* args */): KopeteProtocol( parent, name )
{
	gotOwnInfo = false;

	setStatusIcon( ICON_OFFLINE );

	mPrefs = new ICQPreferences("icq_protocol", this);
	QObject::connect( mPrefs, SIGNAL(saved()), this, SLOT(slotUpdatedSettings()));

	// Instantiate the ICQChangeStatus dialog
	mChangeStatus = new ICQChangeStatus(this);
	
	// Initialize the actions
	initActions();

//	kdDebug(14110) << k_funcinfo << "Creating Client Object..." << endl;
	mEngine = new KICQClient( this );
//	kdDebug(14110) << "Created." << endl;

	setLogProc(&logMonitor);

	// ====================
	// TODO: Just do this in slotUpdatedSettings

	mEngine->owner->Uin = mPrefs->uin().toULong();
	mEngine->storePassword( mPrefs->passwd() );
	SocketFactory *factory = mEngine->factory();
	mEngine->ServerHost = mPrefs->serverAdress().latin1();
	mEngine->ServerPort = mPrefs->serverPort();
	factory->MinTCPPort = mPrefs->listenFromPort();
	factory->MaxTCPPort = mPrefs->listenToPort();

	if ( mPrefs->useProxy() )
	{
		factory->ProxyType = (PROXY_TYPE)(mPrefs->proxyType());
		factory->ProxyHost = mPrefs->proxyServerAdress().latin1();
		factory->ProxyPort = mPrefs->proxyPort();
		factory->ProxyAuth = mPrefs->proxyAuth();
		factory->ProxyUser = mPrefs->proxyAuthUser().latin1();
		factory->ProxyPasswd = mPrefs->proxyAuthPassword().latin1();
	}
	else
	{
		factory->ProxyType = PROXY_NONE;
	}

	if ( mPrefs->useCustomPorts() && mPrefs->customMinPort() >= 1 && mPrefs->customMaxPort() >= 2)
	{
		factory->MinTCPPort = mPrefs->customMinPort();
		factory->MaxTCPPort = mPrefs->customMaxPort();
	}

	// security :)
	mEngine->HideIp = mPrefs->hideIP();
	mEngine->setSecurityInfo( mPrefs->auth(), mPrefs->webAware() );
	QObject::connect( mEngine, SIGNAL(event(ICQEvent *)), this, SLOT(slotEvent(ICQEvent *)));

	// ====================

//	kdDebug(14110) << k_funcinfo << "Creating ICQContact of ourselves..." << endl;

	myContact = new ICQContact( mPrefs->uin(), mPrefs->nickName(), QStringList(i18n("Unknown")), this, 0L );

	myContact->mUser = static_cast<ICQUser*>(mEngine->owner);
	myContact->mUser->Uin = mPrefs->uin().toULong();
	myContact->mUser->Nick = mPrefs->nickName().utf8();
	myContact->setDisplayName( mPrefs->nickName() );

//	kdDebug(14110) << "[ICQProtocol] setInfo .. with mUser = " << (myContact->mUser) << endl;
//	mEngine->setInfo( (myContact->mUser) );

	KConfig *cfg = KGlobal::config();
	cfg->setGroup("ICQ");
	QString listVersion=cfg->readEntry( "ContactList Version", "0.4.x" ) ;
	if ( listVersion=="0.4.x" )
	{
		kdDebug(14110) << "[ICQProtocol] import contact from kopete 0.4.x" << endl;
		importOldContactList();
		cfg->setGroup("ICQ");
		cfg->writeEntry ( "ContactList Version", "0.5" );
	}

	QObject::connect( KopeteTransferManager::transferManager(),
		SIGNAL( accepted( KopeteTransfer *, const QString & ) ),
		this, SLOT( slotFileAccepted( KopeteTransfer *, const QString & ) ) );
	QObject::connect( KopeteTransferManager::transferManager(),
		SIGNAL( refused( const KopeteFileTransferInfo & ) ) ,
		this, SLOT( slotFileDeclined( const KopeteFileTransferInfo & ) ) );

	if ( mPrefs->autoConnect() )
	{
		kdDebug(14110) << "[ICQProtocol] autoconnect..." << endl;
		QTimer::singleShot( 0, this, SLOT( connect() ) );
	}

	addAddressBookField( "messaging/icq", KopetePlugin::MakeIndexField );

//	kdDebug() << k_funcinfo << "END ICQProtocol constructor." << endl;
}

ICQProtocol::~ICQProtocol()
{
	delete mChangeStatus;
	setLogProc(NULL);
}

/***************************************************************************
 *                                                                         *
 *   Re-implementation of Plugin class methods                             *
 *                                                                         *
 ***************************************************************************/

void ICQProtocol::deserializeContact( KopeteMetaContact *metaContact, const QMap<QString, QString> &serializedData,
	const QMap<QString, QString> & /* addressBookData */ )
{
	QString contactId   = serializedData[ "contactId" ];
	QString displayName = serializedData[ "displayName" ];

	new ICQContact( contactId, displayName, metaContact->groups().toStringList(), this, metaContact );
}

bool ICQProtocol::addContactToMetaContact( const QString &contactId, const QString &displayName,
		KopeteMetaContact *parentContact )
{
	//TODO: When Server-side groups are implemented, do not add group to server if
	//temporary contact!
//	kdDebug(14110) << k_funcinfo<< "contactId=" << contactId << ", displayName=" << displayName << endl;

	ICQContact *newContact = static_cast<ICQContact*>( contacts()[ contactId ] );
	if( newContact )
	{
//		kdDebug() << k_funcinfo << "adding already existing ICQContact to MetaContact" << endl;
		parentContact->addContact( newContact );
	}
	else
	{
//		kdDebug() << k_funcinfo << "creating new ICQContact in MetaContact" << endl;
		newContact = new ICQContact( contactId, displayName, parentContact->groups().toStringList(), this, parentContact );
	}

	return (newContact != 0L);
}

ICQGroup *ICQProtocol::findGroup(const QString &groupName, bool create)
{
//	kdDebug(14110) << k_funcinfo << "group=" << groupName << ", create=" << create << endl;

	vector<ICQGroup*>::iterator it;
	for (it = mEngine->contacts.groups.begin(); it != mEngine->contacts.groups.end(); it++)
	{
		if ( !(*it) )
			continue;

		if ( QString::fromLocal8Bit((*it)->Name.c_str()).lower() == groupName.lower() )
			return (*it);
	}

	if ( create && groupName.length() )
	{
		kdDebug(14110) << k_funcinfo << "creating group '" << groupName << "'." << endl;
		ICQGroup *tmpG = new ICQGroup();
		ICQGroup *group = mEngine->getGroup(mEngine->contacts.getGroupId(tmpG), true);
		group->Name = groupName.utf8();
		delete tmpG;
		return group;
	}
	else
	{
		return 0L;
	}
}

void ICQProtocol::initActions()
{
//	kdDebug(14110) << k_funcinfo << " called." << endl;
	
	actionGoOnline		= new KAction ( i18n("Online"), ICON_ONLINE, 0, this, SLOT(slotGoOnline()), this, "actionICQOn" );
	actionGoOffline	= new KAction ( i18n("Offline"), ICON_OFFLINE, 0, this, SLOT(slotGoOffline()), this, "actionICQOff" );
	actionGoAway		= new KAction ( i18n("Away"), ICON_AWAY, 0, this, SLOT(slotGoAway()), this, "actionICQAway" );
	actionGoNA			= new KAction ( i18n("Not Available"), ICON_NA, 0, this, SLOT(slotGoNA()), this, "actionICQNA" );
	actionGoDND			= new KAction ( i18n("Do Not Disturb"), ICON_DND, 0, this, SLOT(slotGoDND()), this, "actionICQDND" );
	actionGoOccupied	= new KAction ( i18n("Occupied"), ICON_OCC, 0, this, SLOT(slotGoOccupied()), this, "actionICQOcc" );
	actionEditOwnInfo	= KopeteStdAction::contactInfo( this, SLOT(slotEditOwnInfo()), this, "actionInfo" );
	// TODO: unused
	actionGoInvisible	= new KToggleAction ( i18n("Invisible"), ICON_INVISIBLE, 0, this, SLOT(slotGoInvisible()), this, "actionICQInv" );

	actionStatusMenu = new KActionMenu ( "ICQ", this );
	actionStatusMenu->insert( actionGoOnline );
	actionStatusMenu->insert( actionGoOffline );
	actionStatusMenu->insert( actionGoAway );
	actionStatusMenu->insert( actionGoNA );
	actionStatusMenu->insert( actionGoDND );
	actionStatusMenu->insert( actionGoOccupied );
//	actionStatusMenu->insert( actionGoInvisible );
	actionStatusMenu->insert( actionEditOwnInfo );

	protocolActionMenu = new KActionMenu( "ICQ", this );
	protocolActionMenu->popupMenu()->insertTitle(
		SmallIcon( ICON_ONLINE ),
		i18n("%1 (%2)").arg( mPrefs->nickName() ).arg( mPrefs->uin() ) );
	protocolActionMenu->insert(actionGoOnline);
	protocolActionMenu->insert(actionGoOffline);
	protocolActionMenu->insert(actionGoAway);
	protocolActionMenu->insert(actionGoNA);
	protocolActionMenu->insert(actionGoDND);
	protocolActionMenu->insert(actionGoOccupied);
//	protocolActionMenu->insert(actionGoInvisible);
	protocolActionMenu->popupMenu()->insertSeparator();
	protocolActionMenu->insert(actionEditOwnInfo);
}

KActionMenu *ICQProtocol::protocolActions(void)
{
	return protocolActionMenu;
}

void ICQProtocol::slotEditOwnInfo(void)
{
	ICQUserInfo *icquserinfo = new ICQUserInfo( myContact, mPrefs->nickName(), this, true);
	icquserinfo->exec(); // This is a modal dialog
	delete icquserinfo;
}

void ICQProtocol::slotEvent(ICQEvent *e)
{
	// The real magic
	if (!e)
		return;

	switch ( e->type() )
	{
		case EVENT_STATUS_CHANGED:
		{
			ICQUser *u = mEngine->getUser(e->Uin(),false);

			if (!u)
			{
				kdDebug(14110) << k_funcinfo << "EVENT_STATUS_CHANGED for UNKNOWN user " << e->Uin() << endl;
				break;
			}

			if ( u->Uin == mEngine->owner->Uin ) // OWN status changed
			{
				statusChange();
				/*
				if ( !gotOwnInfo ) // we did not get our own info yet, do that now
				{
					kdDebug(14110) << "[ICQProtocol] REQUEST OWN INFO" << endl;
					mEngine->addInfoRequest(mEngine->Uin, true);
				}

				// FIXME: WHY THIS?
				// We want to send the status changes for our UIN to all contacts
				for (QMap<unsigned long, ICQContact *>::Iterator it = contactsMap.begin(); it != contactsMap.end(); ++it)
				{
					if (it.data())
						it.data()->statusHasChanged(u);
				}
				*/
			}
			else
			{
				ICQContact *contact = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
				if ( contact && (u->uStatus!=u->prevStatus) )
					contact->statusHasChanged(u);
			}
			break;
		}

		case EVENT_MESSAGE_RECEIVED:
		{
			ICQMessage *message = e->message();
			if ( !message )
				return;

//			kdDebug(14110) << k_funcinfo << "Received Message..." << endl;

			if (e->state == ICQEvent::Fail)
			{
				kdDebug(14110) << "[ICQProtocol] Receiving a message failed...?" << endl;
				break;
			}
			else
			{
				switch(message->Type())
				{
					case ICQ_MSGxURL:
					case ICQ_MSGxMSG:
					case ICQ_MSGxWEBxPANEL:
					case ICQ_MSGxEMAILxPAGER:
					{
						unsigned long uin = message->getUin();

						ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( uin ) ] );
						if (c)
						{
//							kdDebug(14110) << "[ICQProtocol] Normal/URL/Web/Email Message from " << c->displayName() << " (" << uin << ") [" << c->mUser->GrpId << "]" << endl;
							c->newMessage(message);
						}
						else
						{
							if (!mPrefs->rejNotOnList())
							{
								kdDebug(14110) << "[ICQProtocol] Normal/URL/Web/Email Message from UNKNOWN (" << uin << ")" << endl;
								addContact( QString::number(uin), QString::number(uin), 0L, QString::null, true); //add a temporary contact
								c = static_cast<ICQContact*>( contacts()[ QString::number( uin ) ] ); // try again
								if (c)
									c->newMessage(message);
							}
							else
							{
								kdDebug(14110) << "[ICQProtocol] IGNORED Normal/URL/Web/Email Message from UNKNOWN (" << uin << ")" << endl;
							}
						}
						break;
					}

					case ICQ_MSGxFILE:
					{
						ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
						if ( !c )
							break;
/*
						ICQUser *u = mEngine->getUser(message->getUin(), false);
						if (!u)
							return;
*/
						ICQFile *f = static_cast<ICQFile *>(e->message());
						if (!f)
							return;
/*
						QString nameToUse;
						if (!QString(u->Alias.c_str()).isEmpty())
							nameToUse = QString::fromLocal8Bit(u->Alias.c_str());
						else if (!QString::fromLocal8Bit(u->Nick.c_str()).isEmpty())

							nameToUse = QString::fromLocal8Bit(u->Nick.c_str());
						else
							nameToUse = QString(message->getUin());
*/
						unsigned long size = f->Size;

						KopeteTransferManager::transferManager()->askIncomingTransfer(c, c->displayName(), size, QString::fromLocal8Bit(f->Description.c_str()) , f);
						break;
					}

					case ICQ_MSGxADDEDxTOxLIST:
					{
						kdDebug(14110) << "[ICQProtocol] ICQ_MSGxADDEDxTOxLIST; Got added UNKNOWN (" << e->Uin() << ")" << endl;
						break;
					}

					default:
					{
						ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
						if (c)
						{
							kdDebug(14110) << "[ICQProtocol] Received Unhandled ICQ_MSGx... TYPE " << message->Type() <<
								" from " << c->displayName() << "(" << e->Uin() << ") [" << c->user()->GrpId << "]" << endl;
						}
						else
						{
							kdDebug(14110) << "[ICQProtocol] Received Unhandled ICQ_MSGx... TYPE " << message->Type() <<
								" from UNKNOWN (" << e->Uin() << ")" << endl;
						}
						break;
					}
				}
				break;
			}
			break;
		} // case EVENT_MESSAGE_RECEIVED

		case EVENT_ACKED:
		{
			ICQMessage *message = e->message();
			if (!message)
				return;

			kdDebug(14110) << "[ICQProtocol] Received Acknowledgment..." << endl;

			switch(message->Type())
			{
				case ICQ_MSGxFILE:
				{
					ICQFile *f = static_cast<ICQFile *>(message);
					if (!f)
						return;

					if (!myself())
						return;

					/*ICQUser *u = mEngine->getUser(mEngine->owner->Uin, false);
					if (!u)
						return;
					QString nameToUse;
					if (QString::fromLocal8Bit(u->Alias.c_str()).isEmpty())
					{
						if (!QString::fromLocal8Bit(u->Nick.c_str()).isEmpty())
							nameToUse = QString::fromLocal8Bit(u->Nick.c_str());
					} else {
						nameToUse = QString::fromLocal8Bit(u->Alias.c_str());
					}
					unsigned long size = f->Size;
					ICQContact *c = contacts()[ QString::number( e->Uin() ) ];
					if ( !c )
						break;

					KopeteTransfer *kt = KopeteTransferManager::transferManager()->addTransfer(c, QString(f->Name.c_str()), size, nameToUse , KopeteFileTransferInfo::Incomming);
					if (!kt)
						break;*/
					KopeteTransfer *kt = m_tmpFileTransfer;
					m_tmpFileTransfer=0L;
					
					kdDebug(14110) << "[ICQProtocol] ICQ_MSGxFILE" << endl;
					QObject::connect( mEngine, SIGNAL(fileExist(ICQFile*, const QString&, bool)), this, SLOT(slotFileExist(ICQFile*, const QString&, bool)));
					//FIXME: What was that signal actually for?
					//QObject::connect( mEngine, SIGNAL(fileProcess(ICQFile*)), this, SLOT(processed(ICQFile*)));
					new ICQFileTransfer( f, kt, this, this, "ICQFileTransfer" );
					break;
				}

				default:
				{
					ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
					if (c)
					{
						kdDebug(14110) << "[ICQProtocol] Received Unhandled ACK for Message TYPE " << message->Type() <<
							" from " << c->displayName() << "(" << e->Uin() << ") [" << c->user()->GrpId << "]" << endl;
					}
					else
					{
						kdDebug(14110) << "[ICQProtocol] Received Unhandled ACK for Message TYPE " << message->Type() <<
							" from UNKNOWN (" << e->Uin() << ")" << endl;
					}
					break;
				}
			} // switch(message->Type())
			break;
		} // case EVENT_ACKED

		case EVENT_INFO_CHANGED:
		{
			if ( e->Uin() < 10000 )
			{
				kdDebug(14110) << "[ICQProtocol] EVENT_INFO_CHANGED for LOW Uin " << e->Uin() << "." << endl;
				break;
			}

			if ( (e->Uin() == mEngine->owner->Uin) && (e->subType() == EVENT_SUBTYPE_FULLINFO) )
			{
				kdDebug(14110) << "[ICQProtocol] EVENT_INFO_CHANGED for OURSELVES" << endl;
/*				ICQUser u;
				u = *mEngine->owner;
				mEngine->setInfo(&u);*/
				myContact->infoUpdated( e->subType() );
				break;
			}

			ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
			if (c)
				c->infoUpdated( e->subType() );
			break;
		}

		case EVENT_MESSAGE_SEND:
		{
			ICQMessage *m = e->message();
			if (!m)
				return;

			if (e->state == ICQEvent::Success)
			{
				ICQContact *c = pendingMessages[m];
				if (!c)
					return;
				c->messageAck( m );
				pendingMessages.remove( m );
			}
			else
			{
				if (e->Uin()==0 || m->getUin()==0)
				{
					kdDebug(14110) << "[ICQProtocol] Sending Message to Server failed," << endl <<
						"              message type=" << m->Type() <<
						", message UIN=" << m->getUin() << ", event UIN=" << e->Uin() << endl;
					break;
				}

				QString tmpName;
				ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
				if (c)
					tmpName = c->displayName();
				else
					tmpName = QString::number(m->getUin());

				KMessageBox::error(qApp->mainWidget(), i18n("<qt>There was an error sending your last message to user %1 via ICQ. Please make sure that you are still connected to the server, otherwise make sure that the third party is not having connection problems.</qt>").arg(c->displayName()), i18n("Error Sending Message"));
				break;
			}
			break;
		}

		case EVENT_USER_DELETED:
		{
			kdDebug(14110) << "[ICQProtocol] EVENT_USER_DELETED uin= " << e->Uin() << endl;
			break;
		}

		case EVENT_USERGROUP_CHANGED:
		{
			kdDebug(14110) << "[ICQProtocol] EVENT_USERGROUP_CHANGED; uin= " << e->Uin() << endl;
			break;
		}

		case EVENT_GROUP_CHANGED:
		{
			kdDebug(14110) << "[ICQProtocol] EVENT_GROUP_CHANGED; uin= " << e->Uin() << "." << endl;
			// YES, this is the event where we have to check for new serverside contacts :)
			if ( e->Uin() == 0 )
				receivedContactlist();
			break;
		}

		case EVENT_GROUP_CREATED:
		{
			kdDebug(14110) << "[ICQProtocol] EVENT_GROUP_CREATED, doing nothing" << endl;
			break;
		}

		case EVENT_SEARCH:
		{
			emit searchEvent(e);
			break;
		}

		case EVENT_LOGIN_ERROR:
		{
			KMessageBox::error(qApp->mainWidget(), i18n("<qt>ICQ Plugin: There was an unknown error connecting to the ICQ server. Please try again later.</qt>"), i18n("Error connecting"));
			//disconnect();
			setStatusIcon(ICON_OFFLINE);
			QTimer::singleShot(10, this, SLOT(slotGoOffline()));
			break;
		}

		case EVENT_RATE_LIMIT:
		{
			KMessageBox::error(qApp->mainWidget(), i18n("<qt>ICQ Plugin: You are connecting, disconnecting and reconnecting too quickly. The server is throttling your connection speed. You must try again in 10 minutes. Connecting any sooner will result in the wait time increasing exponentially.</qt>"), i18n("Connection Throttled"));
			//disconnect();
			setStatusIcon(ICON_OFFLINE);
			QTimer::singleShot(10, this, SLOT(slotGoOffline()));
			break;
		}

		case EVENT_BAD_PASSWORD:
		{
			KMessageBox::error(qApp->mainWidget(), i18n("<qt>ICQ Plugin: You have specified an invalid password or UIN in the ICQ plugin preferences page. Please change it, and try again.</qt>"), i18n("Invalid Login"));
			//disconnect();
			setStatusIcon(ICON_OFFLINE);
			QTimer::singleShot(10, this, SLOT(slotGoOffline()));
			break;
		}

		case EVENT_ANOTHER_LOCATION:
		{
			KMessageBox::error(qApp->mainWidget(), i18n("<qt>ICQ Plugin: Someone is using your UIN from another location. You have been disconnected.</qt>"), i18n("Disconnected..."));
			//disconnect();
			setStatusIcon(ICON_OFFLINE);
			QTimer::singleShot(10, this, SLOT(slotGoOffline()));
			break;
		}

		default:
		{
			ICQContact *c = static_cast<ICQContact*>( contacts()[ QString::number( e->Uin() ) ] );
			if (c)
			{
				kdDebug(14110) << "[ICQProtocol] Received Unhandled EVENT with TYPE " << e->type() <<
					" from " << c->displayName() << "(" << e->Uin() << ") [" << c->user()->GrpId << "]" << endl;
			}
			else
			{
				kdDebug(14110) << "[ICQProtocol] Received Unhandled EVENT with TYPE " << e->type() <<
					" from UNKNOWN (" << e->Uin() << ")" << endl;
			}
			
			ICQMessage *m = e->message();
			if(m)
			{
				kdDebug(14110) << "[ICQProtocol] Message() DUMP for unhandled EVENT:" << endl;
				kdDebug(14110) << "============================================================" << endl;
				kdDebug(14110) << "Type()        = " << m->Type() << endl;
				kdDebug(14110) << "Time          = " << m->Time << endl;
				kdDebug(14110) << "Received      = " << m->Received << endl;
				kdDebug(14110) << "direct        = " << m->Direct << endl;
				kdDebug(14110) << "Charset       = " << m->Charset.c_str() << endl;
				kdDebug(14110) << "Id            = " << m->Id << endl;
				kdDebug(14110) << "state         = " << m->state << endl;
				kdDebug(14110) << "id1,id2       = " << m->id1 << "," << m->id2 << endl;
				kdDebug(14110) << "timestamp1    = " << m->timestamp1 << endl;
				kdDebug(14110) << "timestamp2    = " << m->timestamp2 << endl;
				kdDebug(14110) << "cookie1       = " << m->cookie1 << endl;
				kdDebug(14110) << "cookie2       = " << m->cookie2 << endl;
				kdDebug(14110) << "DeclineReason = " << m->DeclineReason.c_str() << endl;
				kdDebug(14110) << "isExt         = " << m->isExt << endl;
				kdDebug(14110) << "bDelete       = " << m->bDelete << endl;
				kdDebug(14110) << "============================================================" << endl;
			}
			break;
		}
	}
}


void ICQProtocol::slotFileExist(ICQFile*f, const QString &file, bool canResume)
{
	kdDebug(14110) << "[ICQProtocol] slotFileExist() called; file=" << file << ", canResume=" << canResume << "." << endl;
	if (canResume)
		f->resume(FT_RESUME);
	else
		f->resume(FT_REPLACE);
}


void ICQProtocol::slotFileAccepted(KopeteTransfer *trans , const QString &filename)
{
	if( trans->info().contact()->protocol()->pluginId() != pluginId())
		return;
	ICQFile *f=dynamic_cast<ICQFile *>((ICQFile *)(trans->info().internalId()));
	if(!f)
		return;

	m_tmpFileTransfer=trans;

	kdDebug(14110) << "[ICQProtocol] slotFileAccepted() called" << endl;
	/*QCString tmp = path.local8Bit().append("/");
	tmp.append(f->shortName().c_str());*/
	f->localName = filename.local8Bit();
	f->autoAccept = true;
	mEngine->acceptMessage(f);
}

void ICQProtocol::slotFileDeclined(const KopeteFileTransferInfo &info)
{
	if( info.contact()->protocol()->pluginId()!= pluginId())
		return;
	ICQFile *f=dynamic_cast<ICQFile *>((ICQFile *)(info.internalId()));
	if(!f)
		return;

	kdDebug(14110) << "[ICQProtocol] slotFileDeclined() called" << endl;
	mEngine->declineMessage(f, QString(i18n("User openly declined the file.")));
}



void ICQProtocol::receivedContactlist( void )
{
// mETz: I cannot leave this in for the release or else I will get bugreports for every move/copy of a
// serverside contact in kopete. I can either ignore the list like I do now or else I have to COMPLETELY support
// every action possible to a contact. This is TOO much work for 0.5

	kdDebug(14110) << "[ICQProtocol] START receivedContactlist() =============================================" << endl;

	ICQGroup *group = 0L;
	ICQContact *c = 0L;
	KopeteMetaContact *newContact = 0L;
	list<ICQUser*>::iterator it;
	ICQContactList &engineContacts = mEngine->contacts;

	vector<ICQGroup*>::iterator git;
	kdDebug(14110) << "Serverside Groups ===========================================" << endl;
	for (git = engineContacts.groups.begin(); git != engineContacts.groups.end(); git++)
	{
		kdDebug(14110) << "'" << ((*git)->Name).c_str() << "' (" << (*git)->Id << ")" << endl;
	}
	kdDebug(14110) << "=============================================================" << endl;

	for (it = engineContacts.users.begin(); it != engineContacts.users.end(); it++)
	{
		c = static_cast<ICQContact*>( contacts()[ QString::number( ( *it )->Uin ) ] );
		if ( c )
		{
//			kdDebug(14110) << "[ICQProtocol] Serverside contact " << (*it)->Nick << "(" << (*it)->Uin << ") [" << (*it)->GrpId << "] ALREADY IN KOPETE" << endl;
			continue;
		}

		// Found Contact in Engine that is not unknown by the Kopete-Plugin -> Serverside Contact!
		kdDebug(14110) << "[ICQProtocol] Serverside contact " << (*it)->Alias.c_str() << "(" << (*it)->Uin << ") [" << (*it)->GrpId << "]" << endl;

		group = mEngine->getGroup( (*it)->GrpId, false );
		if ( group )
		{
			kdDebug(14110) << "[ICQProtocol] Group for contact is: " << group->Name.c_str() << "(" << (*it)->GrpId << ")" << endl;
			if ( !addContact( QString::number((*it)->Uin), QString::fromLocal8Bit((*it)->Alias.c_str()), 0L, QString::fromLocal8Bit(group->Name.c_str()), false ) )
				kdDebug(14110) << "[ICQProtocol] ERROR, could not add a contact" << endl;
		}
		else
		{
			kdDebug(14110) << "[ICQProtocol] Group for serverside contact NOT found, what shall we do now?." << endl;
		}

//		kdDebug(14110) << "[ICQProtocol] receivedContactlist() resetting Pointers" << endl;
		c = 0L;
		group = 0L;
		newContact = 0L;
	} // for()

	kdDebug(14110) << "[ICQProtocol] END receivedContactlist() ===============================================" << endl;
}

void ICQProtocol::statusChange()
{
	if ( mEngine->isConnecting() )
	{
		kdDebug(14110) << "[ICQProtocol] statusChange() called in a connect attempt, setting ICON_CONNECTING." << endl;
		setStatusIcon(ICON_CONNECTING);
		return;
	}

	if ( mEngine->owner->uStatus == mEngine->owner->prevStatus )
	{
		kdDebug(14110) << "[ICQProtocol] status did not change, aborting statusChange()" << endl;
		return;
	}
	kdDebug(14110) << "[ICQProtocol] OUR status changed to " << mEngine->owner->uStatus << endl;
	mEngine->owner->prevStatus = mEngine->owner->uStatus;

	int status = mEngine->owner->uStatus;
	if ((status & 0xFFFF) == ICQ_STATUS_OFFLINE)
	{
		setStatusIcon(ICON_OFFLINE);
		return;
	}
	
	if (status & ICQ_STATUS_DND)
		setStatusIcon(ICON_DND);
	else if (status & ICQ_STATUS_OCCUPIED)
		setStatusIcon(ICON_OCC);
	else if (status & ICQ_STATUS_NA)
		setStatusIcon(ICON_NA);
	else if (status & ICQ_STATUS_AWAY)
		setStatusIcon(ICON_AWAY);
	else
		setStatusIcon(ICON_ONLINE);

	if ( mEngine->owner->InvisibleId )
		kdDebug(14110) << "[ICQProtocol] status + INVISIBLE, FIXME find representation for that" << endl;
}

const QString ICQProtocol::protocolIcon()
{
	return ICON_ONLINE;
}

void ICQProtocol::slotGoOnline()
{
	if(mPrefs->uin().toULong()==0)
	{
		int r=KMessageBox::warningContinueCancel(qApp->mainWidget(),
				i18n("<qt>You have not yet specified your ICQ UIN and password. "
					"You can specify your ICQ settings in the Kopete "
					"configuration dialog<br>"
					"Do you want to configure ICQ now?</qt>" ),
				i18n( "ICQ plugin not configured yet" ),
				KGuiItem( i18n( "C&onfigure..." ), "configure" ) );
		if(r!=KMessageBox::Cancel)
		{
			mPrefs->activate();
		}
		return;
	}

	kdDebug(14110) << "[ICQProtocol] slotGoOnline()" << endl;
	if ( mEngine->isLogged() )
	{
		mEngine->setStatus(ICQ_STATUS_ONLINE);
	}
	else
	{
		/* Absolutely do NOT call Connect(). Connect() will use your "auto-set-status-after-connect mode" (set in the preferences)
		    when it should go into "Online" status.
		*/
		mEngine->setStatus(ICQ_STATUS_ONLINE);
	}
}

void ICQProtocol::slotGoOffline()
{
	kdDebug(14110) << "[ICQProtocol] slotGoOffline()" << endl;
	disconnect();
}


void ICQProtocol::slotSetAway(QString type)
{
	// Pass the messsage on to our away dialog
	mChangeStatus->show(type);
}


void ICQProtocol::slotGoAway()
{
	kdDebug(14110) << "[ICQProtocol] slotGoAway()" << endl;
	if(mPrefs->uin().toULong()==0)
	{
		int r=KMessageBox::warningContinueCancel(qApp->mainWidget(),
				i18n("<qt>You have not yet specified your ICQ UIN and password. "
					"You can specify your ICQ settings in the Kopete "
					"configuration dialog<br>"
					"Do you want to configure ICQ now?</qt>" ),
				i18n( "ICQ plugin not configured yet" ),
				KGuiItem( i18n( "C&onfigure..." ), "configure" ) );
		if(r!=KMessageBox::Cancel)
		{
			mPrefs->activate();
		}
		return;
	}

	// Call our global away method
	slotSetAway("ICQ_STATUS_AWAY");
	
}

void ICQProtocol::slotGoNA()
{
	kdDebug(14110) << "ICQProtocol::slotGoNA()" << endl;
	if(mPrefs->uin().toULong()==0)
	{
		int r=KMessageBox::warningContinueCancel(qApp->mainWidget(),
				i18n("<qt>You have not yet specified your ICQ UIN and password. "
					"You can specify your ICQ settings in the Kopete "
					"configuration dialog<br>"
					"Do you want to configure ICQ now?</qt>" ),
				i18n( "ICQ plugin not configured yet" ),
				KGuiItem( i18n( "C&onfigure..." ), "configure" ) );
		if(r!=KMessageBox::Cancel)
		{
			mPrefs->activate();
		}
		return;
	}

	// Call our global away
	slotSetAway("ICQ_STATUS_NA");

}

void ICQProtocol::slotGoInvisible()
{
/*
	mEngine->setInvisible( !mEngine->inInvisible() ); // this is a toggle!
//	mEngine->setStatus(ICQ_STATUS_ONLINE);
*/
}

void ICQProtocol::slotGoOccupied()
{
	kdDebug(14110) << "ICQProtocol::slotGoOccupied()" << endl;
	if(mPrefs->uin().toULong()==0)
	{
		int r=KMessageBox::warningContinueCancel(qApp->mainWidget(),
				i18n("<qt>You have not yet specified your ICQ UIN and password. "
					"You can specify your ICQ settings in the Kopete "
					"configuration dialog<br>"
					"Do you want to configure ICQ now?</qt>" ),
				i18n( "ICQ plugin not configured yet" ),
				KGuiItem( i18n( "C&onfigure..." ), "configure" ) );
		if(r!=KMessageBox::Cancel)
		{
			mPrefs->activate();
		}
		return;
	}

	// Call our global away
	slotSetAway("ICQ_STATUS_OCCUPIED");

}

void ICQProtocol::slotGoDND()
{
	kdDebug(14110) << "ICQProtocol::slotGoDND()" << endl;
	if(mPrefs->uin().toULong()==0)
	{
		int r=KMessageBox::warningContinueCancel(qApp->mainWidget(),
				i18n("<qt>You have not yet specified your ICQ UIN and password. "
					"You can specify your ICQ settings in the Kopete "
					"configuration dialog<br>"
					"Do you want to configure ICQ now?</qt>" ),
				i18n( "ICQ plugin not configured yet" ),
				KGuiItem( i18n( "C&onfigure..." ), "configure" ) );
		if(r!=KMessageBox::Cancel)
		{
			mPrefs->activate();
		}
		return;
	}

	// Call our global away
	slotSetAway("ICQ_STATUS_DND");

}

void ICQProtocol::setAway(void)
{
	mEngine->owner->AutoResponseAway = KopeteAway::message().utf8();
	mEngine->setStatus ( ICQ_STATUS_AWAY );
}

void ICQProtocol::setAvailable(void)
{
	mEngine->setStatus ( ICQ_STATUS_ONLINE );
}

bool ICQProtocol::isAway(void) const
{
//	kdDebug(14110) << "[ICQProtocol] isAway() called" << endl;
	const unsigned short status = mEngine->owner->uStatus;
	if ( status == ICQ_STATUS_OFFLINE || (status & 0xFF) == ICQ_STATUS_ONLINE )
	{
		return false;
	}
	return true;
}

void ICQProtocol::slotUpdatedSettings()
{
//	kdDebug(14110) << k_funcinfo << "called." << endl;

	ICQUser *mySelf = myContact->mUser;
	mySelf->Uin = mPrefs->uin().toULong();
	mySelf->Nick = mPrefs->nickName().utf8();
	myContact->setDisplayName ( mPrefs->nickName() );

//	if (mEngine->isLogged())
//		mEngine->setInfo(mySelf);

	mEngine->owner->Uin = mPrefs->uin().toULong();

	mEngine->ServerHost = mPrefs->serverAdress().latin1();
	mEngine->ServerPort = mPrefs->serverPort();

//	kdDebug(14110) << "[ICQProtocol] Set Username and UIN" << endl;

	SocketFactory *factory = mEngine->factory();

	factory->MinTCPPort = mPrefs->listenFromPort();
	factory->MaxTCPPort = mPrefs->listenToPort();

	if ( mPrefs->useProxy() )
	{
		factory->ProxyType = static_cast<PROXY_TYPE>(mPrefs->proxyType());
		factory->ProxyHost = mPrefs->proxyServerAdress().latin1();
		factory->ProxyPort = mPrefs->proxyPort();
		factory->ProxyAuth = mPrefs->proxyAuth();
		factory->ProxyUser = mPrefs->proxyAuthUser().latin1();
		factory->ProxyPasswd = mPrefs->proxyAuthPassword().latin1();
	}
	else
	{
		factory->ProxyType = PROXY_NONE;
	}

	if ( mPrefs->useCustomPorts() && mPrefs->customMinPort() >= 1 && mPrefs->customMaxPort() >= 2)
	{
		factory->MinTCPPort = mPrefs->customMinPort();
		factory->MaxTCPPort = mPrefs->customMaxPort();
	}

//	kdDebug(14110) << "[ICQProtocol] Set Proxy" << endl;

	mEngine->storePassword( mPrefs->passwd() );

	mEngine->WebAware = mPrefs->webAware();
	mEngine->HideIp = mPrefs->hideIP();
	mEngine->Authorize = mPrefs->auth();
//	mEngine->RejectMessage;
	mEngine->RejectURL = mPrefs->rejURLMessages();
	mEngine->RejectWeb = mPrefs->rejWebMessages();
	mEngine->RejectEmail = mPrefs->rejEmailMessages();
//	mEngine->RejectOther;

	if (mEngine->isLogged())
		mEngine->setSecurityInfo(mPrefs->auth(), mPrefs->webAware());

//	kdDebug(14110) << "[ICQProtocol] END slotUpdatedSettings()" << endl;
	emit settingsChanged();
}

void ICQProtocol::init()
{

}

///////////////////////////////////////////////////
//           KopeteProtocol Class reimplementation
///////////////////////////////////////////////////

bool ICQProtocol::isConnected() const
{
//	kdDebug(14110) << "[ICQProtocol] isConnected() returning " << (mEngine->isLogged()) << endl;
	return (mEngine->isLogged());
}

bool ICQProtocol::unload()
{
//	kdDebug(14110) << "[ICQProtocol] ICQProtocol::unload" << endl;

	QObject::disconnect(mEngine,0,0,0); // make sure we do not get called by engine anymore!

	if ( mEngine->isLogged() || mEngine->isConnecting() )
		mEngine->setStatus(ICQ_STATUS_OFFLINE);

	// as mEngine has 'this' as parent the delete should be unneeded
//	delete mEngine;

//	kdDebug(14110) << "[ICQProtocol] END ICQProtocol::unload" << endl;
	return KopeteProtocol::unload();
}

void ICQProtocol::slotGroupAdded(const QString &groupName)
{
	ICQGroup *group = findGroup( groupName, false );
	if ( group )
	{
		kdDebug(14110) << "[ICQProtocol] slotGroupAdded() for KNOWN group: " << groupName << ", aborting." << endl;
	}
	else
	{
		kdDebug(14110) << "[ICQProtocol] slotGroupAdded() for new group: " << groupName << ", creating group." << endl;
		mEngine->createGroup( groupName.local8Bit() );
	}
}

///////////////////////////////////////////////////
//           KopeteProtocol Class reimplementation
///////////////////////////////////////////////////

void ICQProtocol::disconnect()
{
	kdDebug(14110) << "ICQProtocol::Disconnect()" << endl;

	if ( mEngine->isLogged() || mEngine->isConnecting() )
		mEngine->setStatus(ICQ_STATUS_OFFLINE);
}

void ICQProtocol::connect()
{
	kdDebug(14110) << "ICQProtocol::Connect()" << endl;

	if(mPrefs->uin().toULong()==0)
	{
		int r=KMessageBox::warningContinueCancel(qApp->mainWidget(),
				i18n("<qt>You have not yet specified your ICQ UIN and password. "
					"You can specify your ICQ settings in the Kopete "
					"configuration dialog<br>"
					"Do you want to configure ICQ now?</qt>" ),
				i18n( "ICQ plugin not configured yet" ),
				KGuiItem( i18n( "C&onfigure..." ), "configure" ) );
		if(r!=KMessageBox::Cancel)
		{
			mPrefs->activate();
		}
		return;
	}

	Q_UINT16 autoConnectStatus = mPrefs->autoConnectStatus();

	// This looks ugly. Does anyone know a better method? (remenic)
	if (
			autoConnectStatus == ICQ_STATUS_ONLINE ||
			autoConnectStatus == ICQ_STATUS_NA ||
			autoConnectStatus == ICQ_STATUS_AWAY ||
			autoConnectStatus == ICQ_STATUS_OCCUPIED ||
			autoConnectStatus == ICQ_STATUS_DND)
	{
		mEngine->setStatus(autoConnectStatus);
	}
	else
	{
		mEngine->setStatus(ICQ_STATUS_ONLINE);
	}
}

AddContactPage *ICQProtocol::createAddContactWidget(QWidget *parent)
{
	return (new ICQAddContactPage(this,parent));
}

KopeteContact *ICQProtocol::myself() const
{
//	kdDebug(14110) << "[ICQProtocol] myself() called, displayName() is " << myContact->displayName() << endl;
	return myContact;
};

///////////////////////////////////////////////////
//           Internal functions implementation
///////////////////////////////////////////////////

QString ICQProtocol::statusToString ( unsigned int status ) const
{
	if( ( status & 0xFFFF ) == ICQ_STATUS_OFFLINE )
		return i18n ( "Offline" );

	QString result;
	if( status & ICQ_STATUS_DND )
		result = i18n( "Do Not Disturb" );
	else if( status & ICQ_STATUS_OCCUPIED )
		result = i18n( "Occupied" );
	else if( status & ICQ_STATUS_NA )
		result = i18n( "Not Available" );
	else if( status & ICQ_STATUS_AWAY )
		result = i18n( "Away" );
	else
		result = i18n( "Online" );

	return result;
}

void ICQProtocol::importOldContactList()
{
	//this code import contact list from old kxEngine icq plugin (Kopete 0.4.x)
	KSimpleConfig *contactsFile=new KSimpleConfig(locateLocal("data","kopete/icq.contacts"));

	//groups
	contactsFile->setGroup("Default");

	QStrList nameList, numList;
	contactsFile->readListEntry("Group Names",nameList);
	contactsFile->readListEntry("Group IDs",numList);
	//unsigned int groupIDCount=contactsFile->readNumEntry("Group ID Count",0);
	QMap<unsigned int, QString> groupList;

	for (unsigned int t=0; t<nameList.count(); t++)
	{
		QString grname=QString::fromUtf8(nameList.at(t));
		unsigned int groupID=QString(numList.at(t)).toUInt();
		groupList.insert(groupID,grname);
	}

	//contacts
	contactsFile->setGroup("Default");
	unsigned int count=contactsFile->readNumEntry("Count",0);
	QStrList tmp;
	contactsFile->readListEntry("Contacts",tmp);
	//nonICQIDcount=contactsFile->readNumEntry("Non ICQ ID Count",0);
	bool ignoreOnly;

	if (count>0)
	{
		for (unsigned int t=1; t<=count; t++)
		{
			QString str;
			str.sprintf("Contact %s",tmp.at(t-1));
			contactsFile->setGroup(str);
			ignoreOnly=contactsFile->readBoolEntry("Ignore List", FALSE);
			if (!ignoreOnly) // normal contact
			{
				unsigned long uin=contactsFile->readUnsignedNumEntry("UIN");
				QString nick=contactsFile->readEntry("Nick");
				unsigned int groupID=contactsFile->readUnsignedNumEntry("Group", 0);

				kdDebug(14110) << "[ICQProtocol] importOldContactList() : import contact "<< uin <<" nick: "<<nick << " group: " << groupList[groupID] <<"(" <<groupID<<")" <<endl;
				addContact( QString(QString::number(uin)), nick, 0L, groupList[groupID] );

/*				oc->nonICQID=contactsFile->readUnsignedNumEntry("Non ICQ ID",0);
				oc->invisible=contactsFile->readUnsignedNumEntry("Invisible", 0);
				oc->visible=contactsFile->readUnsignedNumEntry("Visible", 0);
				oc->alwaysServer=contactsFile->readBoolEntry("Always via Server", FALSE);
				oc->waitAuth=contactsFile->readBoolEntry("Awaiting Auth", FALSE);
				oc->ignore=contactsFile->readBoolEntry("Ignore", FALSE);
				oc->authorized=contactsFile->readBoolEntry("Authorized", TRUE);
				oc->onlineNotify=contactsFile->readBoolEntry("Online Notify",FALSE);
				oc->gsm=contactsFile->readEntry("Cellular");
				oc->needAuth=contactsFile->readBoolEntry("Need Auth", FALSE);*/
			}
			else // ignore list
			{
				/*kxContactStruct *c=new kxContactStruct;
				c->nick=contactsFile->readEntry("Nick");
				c->uin=contactsFile->readUnsignedNumEntry("UIN");
				c->ignore=contactsFile->readBoolEntry("Ignore", TRUE);
				ignoreList.append(c);*/
			}
		}
	}
}

#include "icqprotocol.moc"

// vim: set noet ts=4 sts=4 sw=4:

