/*
 * qssl.cpp - Qt OpenSSL plugin
 * Copyright (C) 2001, 2002  Justin Karneges
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include"qssl.h"
#include"qssl_p.h"

#include<qstring.h>
#include<qcstring.h>

#include<openssl/ssl.h>
#include<openssl/err.h>


//! \brief plugin hook
QSSL *createQSSL()
{
	return new _QSSL;
}

//! \class QSSL qssl.h
//! \brief QT OpenSSL plug-in

//----------------------------------------------------------------------------
// QSSL
//----------------------------------------------------------------------------
//! \brief Creates QSSL object
_QSSL::_QSSL()
{
	SSL_library_init();
	SSL_load_error_strings();
}

//! \brief Destroys QSSL object
_QSSL::~_QSSL()
{
	ERR_free_strings();
	ERR_remove_state(0);
}

//! \brief gives you access to QSSLFilter
QSSLFilter *_QSSL::createFilter()
{
	return new _QSSLFilter;
}

//----------------------------------------------------------------------------
// QSSLFilter
//----------------------------------------------------------------------------
//! \if _hide_doc_
struct QSSLFilterPrivate
{
	int sslAction;
	QString errMsg;
	QString trustedCertStoreDir;
	QByteArray sendQueue, recvQueue;

	SSL *ssl;
	SSL_METHOD *method;
	SSL_CTX *context;
	BIO *rbio, *wbio;
};
//! \endif

//! \class QSSLFilter qssl.h
//! \brief QSSLFilter - Main SSL wrapper
//!
//! This class should be used if you want basic ssl support.
//!

//! \fn void QSSLFilter::handshaken()
//! \brief Signals letting you know the initialization has finished
//!
//! This is when you put those pad locks at the corner of your application ;D

//! \fn void QSSLFilter::readyRead()
//! \brief Signals non-encrypted data is ready for read

//! \fn void QSSLFilter::outgoingSSLDataReady()
//! \brief Singals encrypted data is ready for read

//! \fn void QSSLFilter::error(bool isWarning, const QString &)
//! \brief Signals any error
//!
//! the isWarning = 1 if this is a warning information. You may choose to
//!   procede with the process if a warning is signaled.
//!
//! \param bool - 1 if warning and 0 if error
//! \param QString - warning/error message
//!
//! \sa continueAfterWarning()

//! \brief Creates QSSLFilter object
_QSSLFilter::_QSSLFilter()
{
	d = new QSSLFilterPrivate;
	d->ssl = 0;
	d->context = 0;
	d->trustedCertStoreDir = "";
}

//! \brief Destroys Object
_QSSLFilter::~_QSSLFilter()
{
	reset();
	delete d;
}

//! \brief Set the SSL certificate directory.
//!
//! \param QString - directory the certificate is located
void _QSSLFilter::setSSLTrustedCertStoreDir(const QString &s)
{
	QString str = s;
	if(str.at(str.length()-1) == '/')
		str.truncate(str.length()-1);

	d->trustedCertStoreDir = str;
}

//! \brief Reset QSSLFilter's vars.
//!
//! Reset all the variable inside QSSLFilter.
//!
//! Use this function if you want to establish a new connection.
void _QSSLFilter::reset()
{
	if(d->ssl) {
		SSL_shutdown(d->ssl);
		SSL_free(d->ssl);
		d->ssl = 0;
	}
	if(d->context) {
		SSL_CTX_free(d->context);
		d->context = 0;
	}

	d->sendQueue.resize(0);
	d->recvQueue.resize(0);
	d->errMsg = "";
	d->sslAction = SSL_IDLE;
}

//! \brief Initializes SSL filtering.
//!
//! Initializes SSL filtering.
//! \warning Make sure you establish a socket connection before you call this
//! function!
//! \return true - if success\n
//! false - if anything goes wrong
bool _QSSLFilter::begin()
{
	reset();

	d->ssl = 0;
	d->method = 0;
	d->context = 0;

	// get our handles
	d->method = TLSv1_client_method();
	if(!d->method) {
		reset();
		return false;
	}
	d->context = SSL_CTX_new(d->method);
	if(!d->context) {
		reset();
		return false;
	}

	if(!d->trustedCertStoreDir.isEmpty()) {
		QString certfile = d->trustedCertStoreDir + "/cacert.pem";
		int ret = SSL_CTX_load_verify_locations(d->context, certfile.latin1(), NULL);
		if(!ret) {
			reset();
			return false;
		}
	}

	d->ssl = SSL_new(d->context);
	if(!d->ssl) {
		reset();
		return false;
	}
	SSL_set_ssl_method(d->ssl, d->method); // can this return error?

	// setup the memory bio
	// these could error out, but i don't see how
	d->rbio = BIO_new(BIO_s_mem());
	d->wbio = BIO_new(BIO_s_mem());

	// these always work
	SSL_set_bio(d->ssl, d->rbio, d->wbio);
	SSL_set_connect_state(d->ssl);

	d->sslAction = SSL_CONNECT;

	sslUpdate();

	return true;
}

//! \brief Send information to decode
//!
//! Send any information you want to decode through this method.
//! \param QByteArray - data to be encrypted
void _QSSLFilter::putIncomingSSLData(const QByteArray &a)
{
	BIO_write(d->rbio, a.data(), a.size());
	sslUpdate();
}

//! \brief Tells you if any data is pending.
bool _QSSLFilter::isOutgoingSSLData()
{
	return (BIO_pending(d->wbio) > 0) ? true: false;
}

//! \brief Returns ssl encrypted data.
//!
//! This is what you want to send out your socket in most cases.
//! \return QByteArray - ssl encrypted data
QByteArray _QSSLFilter::getOutgoingSSLData()
{
	QByteArray a;

	int size = BIO_pending(d->wbio);
	if(size <= 0)
		return a;
	a.resize(size);

	int r = BIO_read(d->wbio, a.data(), size);
	if(r <= 0) {
		a.resize(0);
		return a;
	}
	if(r != size)
		a.resize(r);

	return a;
}

void _QSSLFilter::sslUpdate()
{
	if(d->sslAction == SSL_IDLE)
		return;

	if(d->sslAction == SSL_CONNECT) {
		int ret = SSL_connect(d->ssl);
		if(ret == 0) {
			reset();
			doError();
			return;
		}
		if(ret > 0) {
			if(SSL_do_handshake(d->ssl) < 0) {
				reset();
				doError();
				return;
			}
			d->sslAction = SSL_HANDSHAKING;

			// Added by Kiko 020621: verifies if the certificate chain
			// is valid. If not, raises an error and set an appropriate
			// diagnostic message to be retrieved by the upper layers
			if(!d->trustedCertStoreDir.isEmpty()) {
				if((ret = SSL_get_verify_result(d->ssl)) != X509_V_OK) {
					d->errMsg = sslErrMessage(ret);
					doWarning();
					return;
				}
			}

			finishHandshake();
		}
	}

	if(isOutgoingSSLData()) {
		outgoingSSLDataReady();
	}

	// try to read incoming unencrypted data
	sslReadAll();

	if(isRecvData())
		readyRead();
}

void _QSSLFilter::finishHandshake()
{
	d->sslAction = SSL_ACTIVE;
	processSendQueue();
	handshaken();
}

//! \brief continue after warning
//!
//! Tells the method to procede with the process.
//! \note You should call this function after you establish a socket connection
void _QSSLFilter::continueAfterWarning()
{
	if(d->sslAction == SSL_IDLE || d->sslAction == SSL_CONNECT)
		sslUpdate();
	else if(d->sslAction == SSL_HANDSHAKING)
		finishHandshake();
	else
		sslUpdate();
}

//! \brief send non-encrypted data to object
//!
//! \param QByteArray - non-encrypted data
void _QSSLFilter::send(const QByteArray &a)
{
	int oldsize = d->sendQueue.size();
	d->sendQueue.resize(oldsize + a.size());
	memcpy(d->sendQueue.data() + oldsize, a.data(), a.size());

	if(d->sslAction == SSL_ACTIVE)
		processSendQueue();
}

//! \brief Data waiting for you to take
//!
//! Tells you if there is any encrypted data in the queue for you to take.
//! \return true - if data is queued\n
//!   false - if no data is queued
bool _QSSLFilter::isRecvData()
{
	return (d->recvQueue.size() > 0) ? true: false;
}

void _QSSLFilter::sslReadAll()
{
	QByteArray a;

	while(1) {
		a.resize(4096);
		int x = SSL_read(d->ssl, a.data(), a.size());
		if(x <= 0)
			break;

		if(x != (int)a.size())
			a.resize(x);

		int oldsize = d->recvQueue.size();
		d->recvQueue.resize(oldsize + a.size());
		memcpy(d->recvQueue.data() + oldsize, a.data(), a.size());
	}
}

//! \brief Get SSL decrypted data.
//!
//! \return QByteArray - SSL decrypted data
QByteArray _QSSLFilter::recv()
{
	QByteArray a = d->recvQueue;
	a.detach();
	d->recvQueue.resize(0);
	return a;
}

void _QSSLFilter::processSendQueue()
{
	if(d->sendQueue.size() > 0) {
		SSL_write(d->ssl, d->sendQueue.data(), d->sendQueue.size());
		d->sendQueue.resize(0);
		sslUpdate();
	}
}

void _QSSLFilter::doError()
{
	error(false, d->errMsg);
}

void _QSSLFilter::doWarning()
{
	error(true, d->errMsg);
}

QString _QSSLFilter::sslErrMessage(int errcode)
{
	QString msg;

	switch (errcode) {
		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
			msg=tr("Unable to get issuer certificate");
			break;
		case X509_V_ERR_UNABLE_TO_GET_CRL:
			msg=tr("Unable to get certificate CRL");
			break;
		case X509_V_ERR_UNABLE_TO_DECRYPT_CERT_SIGNATURE:
			msg=tr("Unable to decrypt certificate's signature");
			break;
		case X509_V_ERR_UNABLE_TO_DECRYPT_CRL_SIGNATURE:
			msg=tr("Unable to decrypt CRL's signature");
			break;
		case X509_V_ERR_UNABLE_TO_DECODE_ISSUER_PUBLIC_KEY:
			msg=tr("Unable to decode issuer public key");
			break;
		case X509_V_ERR_CERT_SIGNATURE_FAILURE:
			msg=tr("Invalid certificate signature");
			break;
		case X509_V_ERR_CRL_SIGNATURE_FAILURE:
			msg=tr("Invalid CRL signature");
			break;
		case X509_V_ERR_CERT_NOT_YET_VALID:
			msg=tr("Certificate not yet valid");
			break;
		case X509_V_ERR_CERT_HAS_EXPIRED:
			msg=tr("Certificate has expired");
			break;
		case X509_V_ERR_CRL_NOT_YET_VALID:
			msg=tr("CRL not yet valid");
			break;
		case X509_V_ERR_CRL_HAS_EXPIRED:
			msg=tr("CRL has expired");
			break;
		case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
			msg=tr("Invalid time in certifiate's notBefore field");
			break;
		case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
			msg=tr("Invalid time in certificate's notAfter field");
			break;
		case X509_V_ERR_ERROR_IN_CRL_LAST_UPDATE_FIELD:
			msg=tr("Invalid time in CRL's lastUpdate field");
			break;
		case X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD:
			msg=tr("Invalid time in CRL's nextUpdate field");
			break;
		case X509_V_ERR_OUT_OF_MEM:
			msg=tr("Out of memory while checking the certificate chain");
			break;
		case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
			msg=tr("Certificate is self-signed but isn't found in the list of trusted certificates");
			break;
		case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
			msg=tr("Certificate chain ends in a self-signed cert that isn't found in the list of trusted certificates");
			break;
		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
			msg=tr("Unable to get issuer certificate locally");
			break;
		case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
			msg=tr("Certificate chain contains only one certificate and it's not self-signed");
			break;
		case X509_V_ERR_CERT_CHAIN_TOO_LONG:
			msg=tr("Certificate chain too long");
			break;
		case X509_V_ERR_CERT_REVOKED:
			msg=tr("Certificate is revoked");
			break;
		case X509_V_ERR_INVALID_CA:
			msg=tr("Invalid CA certificate");
			break;
		case X509_V_ERR_PATH_LENGTH_EXCEEDED:
			msg=tr("Maximum certificate chain length exceeded");
			break;
		case X509_V_ERR_INVALID_PURPOSE:
			msg=tr("Invalid certificate purpose");
			break;
		case X509_V_ERR_CERT_UNTRUSTED:
			msg=tr("Certificate not trusted for the required purpose");
			break;
		case X509_V_ERR_CERT_REJECTED:
			msg=tr("Root CA is marked to reject the specified purpose");
			break;
		case X509_V_ERR_SUBJECT_ISSUER_MISMATCH:
			msg=tr("Subject issuer mismatch");
			break;
		case X509_V_ERR_AKID_SKID_MISMATCH:
			msg=tr("Subject Key Identifier doesn't match the Authority Key Identifier");
			break;
		case X509_V_ERR_AKID_ISSUER_SERIAL_MISMATCH:
			msg=tr("Subject Key Identifier serial number doesn't match the Authority's");
			break;
		case X509_V_ERR_KEYUSAGE_NO_CERTSIGN:
			msg=tr("Key Usage doesn't include certificate signing");
			break;
		default:
			msg=tr("Generic certificate validation error");
			break;
	}
	return msg;
}
