/*
	oscarprotocol.cpp  -  Oscar Protocol Plugin

	Copyright (c) 2002 by Tom Linsky <twl6@po.cwru.edu>

	Kopete    (c) 2002 by the Kopete developers  <kopete-devel@kde.org>

	*************************************************************************
	*                                                                       *
	* This program is free software; you can redistribute it and/or modify  *
	* it under the terms of the GNU General Public License as published by  *
	* the Free Software Foundation; either version 2 of the License, or     *
	* (at your option) any later version.                                   *
	*                                                                       *
	*************************************************************************
	*/

#include "oscarprotocol.h"

#include <qapplication.h>
#include <qcursor.h>
#include <qtimer.h>

#include <kconfig.h>
#include <kdebug.h>
#include <kgenericfactory.h>
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kstddirs.h>
#include <kmessagebox.h>

#include "aim.h"
#include "kopeteaway.h"
#include "kopetecontactlist.h"
#include "kopetegroup.h"
#include "kopetemetacontact.h"
#include "oscaraddcontactpage.h"
#include "oscarchangestatus.h"
#include "oscarcontact.h"
#include "oscarpreferences.h"
#include "oscaruserinfo.h"
#include "oscarsocket.h"
#include "oscardebugdialog.h"
#include "aimbuddylist.h"

K_EXPORT_COMPONENT_FACTORY( kopete_oscar, KGenericFactory<OscarProtocol> );

OscarProtocol* OscarProtocol::protocolStatic_ = 0L;

OscarProtocol::OscarProtocol(QObject *parent, const char *name, const QStringList & )
				: KopeteProtocol(parent,name)
{
		if (protocolStatic_)
				kdDebug(14150) << "[OscarProtocol] ERROR: Oscar plugin already initialized" << endl;
		else
				protocolStatic_ = this;

		addAddressBookField( "messaging/aim", KopetePlugin::MakeIndexField );

		m_connected = false;
		myContact = 0L;
		engine = 0L;
		randomNewBuddyNum = 0;
		randomNewGroupNum = 0;
		mBuddyList = new AIMBuddyList();
		QObject::connect(mBuddyList, SIGNAL(groupAdded(AIMGroup *)), this, SLOT(slotReTryServerContacts(AIMGroup *)));
}

OscarProtocol::~OscarProtocol()
{
		delete oscarAwayDialog;
		delete mDebugDialog;

		protocolStatic_ = 0L;
}

// Connects to the server and logs in
void OscarProtocol::connect()
{
		kdDebug(14150) << "[OscarProtocol] Connect()" << endl;

		QString sn = getSN();
		if ( sn != i18n("(No ScreenName Set)") )
		{
				QString pw = mPrefs->password();
				if (pw.isEmpty())
				{
					slotError(i18n("Kopete is unable to attempt to signon to the " \
						"AIM network because no password was specified in the preferences."), 0);
				}
				else
				{
					kdDebug(14150) << "[OscarProtocol] Logging in as " << sn << endl;

					engine->doLogin(
						mPrefs->server(),
						mPrefs->port(), sn, pw);
					setStatusIcon("oscar_connecting");
				}
		}
		else
		{
			int retval;
			retval = KMessageBox::warningContinueCancel(
				qApp->mainWidget(),
					i18n("<qt>You have not yet specified your screen name. " \
					"You can specify your Oscar settings in the Kopete " \
					"configuration dialog<br>" \
					"Do you want to configure Oscar now?</qt>" ),
					i18n( "Oscar plugin not configured yet" ),
					KGuiItem( i18n("C&onfigure..."), "configure" ));

				// If they didn't choose cancel
				if( retval != KMessageBox::Cancel)
					mPrefs->activate();
		}
}

// Disconnects from the server
void OscarProtocol::disconnect()
{
		kdDebug(14150) << "[OscarProtocol] Disconnect() START" << endl;
		if ( !myContact )
			kdDebug(14150) << "[OscarProtocol] ERROR, myContact already got deleted before!" << endl;

		delete myContact;
		myContact = 0L;
		// unneded, slotDisconnected() gets called by the engine afterwards
		//	m_connected = false;

		kdDebug(14150) << "[OscarProtocol] Disconnect(); calling engine->doLogoff()..." << endl;
		engine->doLogoff();
		kdDebug(14150) << "[OscarProtocol] Disconnect(); logged out." << endl;
		kdDebug(14150) << "[OscarProtocol] Disconnect() END" << endl;
}

const QString OscarProtocol::protocolIcon()
{
	return "oscar_online";
}

// Tells whether we are connected
bool OscarProtocol::isConnected() const
{
		return m_connected;
}

// Called when kopete sets global away
void OscarProtocol::setAway(void)
{
		kdDebug(14150) << "[OscarProtocol] setAway()" << endl;
		QString msg;
		msg = KopeteAway::getInstance()->message();
		if (msg.isEmpty())
				msg = " ";
		// FIXME: Chris check this
		engine->sendAway(OSCAR_AWAY,msg);
}

// Called when kopete wants to set our status to online
void OscarProtocol::setAvailable(void)
{
		kdDebug(14150) << "[OscarProtocol] setAvailable()" << endl;
		if (m_connected)
		{
				// sendAway with an empty string sets you back again
				// FIXME: that's a dumbass API, write a setAvailable() method insted
				engine->sendAway(OSCAR_AWAY,"");
				m_isAway = false;
		}
		else
		{
				connect();
		}
}

// Returns true if we are away, occupied, etc
bool OscarProtocol::isAway(void) const
{
		return m_isAway;
}

// Creates the add contact page
AddContactPage *OscarProtocol::createAddContactWidget(QWidget *parent)
{
		return new OscarAddContactPage(this,parent);
}

// Adds the given contact to the protocol's list
void OscarProtocol::addNewContact(const QString &sn, const QString &alias,
				bool hasAlias, const QString &groupName, KopeteMetaContact *meta)
{
	kdDebug( 14150 ) << k_funcinfo << "sn=" << sn << ", alias=" << alias << ", group=" << groupName << endl;

	AIMGroup *group = mBuddyList->findGroup(groupName);
	if (!group)
		group = mBuddyList->addGroup(randomNewGroupNum, groupName);
	AIMBuddy *newBuddy = new AIMBuddy(randomNewBuddyNum, group->ID(), sn);
	// Check if it has an alias
	if( hasAlias )
		newBuddy->setAlias(alias);

	if (m_connected) // update serverside grouplist
		engine->sendAddGroup(group->name());

	mBuddyList->addBuddy( newBuddy );

	if (m_connected) // update serverside buddylist
		engine->sendAddBuddy(sn,group->name());

	addOldContact( newBuddy,meta );
	randomNewGroupNum++;
	randomNewBuddyNum++;
}

// Called on plugin init
void OscarProtocol::init(void)
{
		kdDebug(14150) << "[OscarProtocol] init() START" << endl;
		initActions();

		// The preferences dialog
		// (needs to be done before initEngine so we can use it to retrieve server info
		mPrefs = new OscarPreferences("oscar_protocol",this);

		initEngine();

		QObject::connect(
						KopeteContactList::contactList(),
						SIGNAL(groupAdded(KopeteGroup *)),
						this,
						SLOT(slotGroupAdded(KopeteGroup *)));

		// Set up the status bar icon
//	kdDebug(14150) << "[OscarProtocol] Setting status bar icon..." << endl;
		setStatusIcon( "oscar_offline" );

		// Instantiate the away dialog
		oscarAwayDialog = new OscarChangeStatus( engine );

		// The debug dialog
		mDebugDialog = new OscarDebugDialog();
		engine->setDebugDialog(mDebugDialog);

		// Protocol error
		QObject::connect(engine,SIGNAL(protocolError(QString, int)),
						this, SLOT(slotError(QString, int)));
		// Got IM
		QObject::connect(engine, SIGNAL(gotIM(QString,QString,bool)),
						this, SLOT(slotGotIM(QString,QString,bool)));
		// Got Config (Buddy List)
		QObject::connect(engine, SIGNAL(gotConfig(AIMBuddyList &)),
						this, SLOT(slotGotConfig(AIMBuddyList &)));
		// Got my user info
		QObject::connect(engine, SIGNAL(gotMyUserInfo(UserInfo)),
						this, SLOT(slotGotMyUserInfo(UserInfo)));
		// Status changed (I think my own status)
		QObject::connect(engine, SIGNAL(statusChanged(int)),
						this, SLOT(slotStatusChanged(int)));
		// Got warning
		QObject::connect(engine, SIGNAL(gotWarning(int,QString)),
						this, SLOT(slotGotWarning(int,QString)));
		// Got direct IM request
		QObject::connect(engine, SIGNAL(gotDirectIMRequest(QString)),
						this, SLOT(slotGotDirectIMRequest(QString)));
		// We have officially become idle
		QObject::connect(&mIdleMgr, SIGNAL(timeout()),
						this, SLOT(slotIdleTimeout()));
		// We have officially become un-idle
		QObject::connect(&mIdleMgr, SIGNAL(activity()),
						this, SLOT(slotIdleActivity()));
		// Preferences have been saved
		QObject::connect(mPrefs, SIGNAL(saved()),
						this, SLOT(slotPreferencesSaved()));

		// Autoconnect if is selected in config
		if ( mPrefs->autoConnect() )
		{
				kdDebug(14150) << "[OscarProtocol] Autoconnect, calling Connect()" << endl;
				connect();
		}

		slotPreferencesSaved();

		//set user profile in engine
		engine->setMyProfile( mPrefs->profile() );

		kdDebug(14150) << "[OscarProtocol] init() END" << endl;
}

// Called when plugin unloads
bool OscarProtocol::unload(void)
{
		kdDebug(14150) << "[OscarProtocol] unload() START" << endl;

		kdDebug(14150) << "[OscarProtocol] unload(), disconnecting..." << endl;
		disconnect();
		kdDebug(14150) << "[OscarProtocol] unload(), disconnected" << endl;

		if (engine)
		{
				kdDebug(14150) << "[OscarProtocol] unload(), deleting engine..." << endl;
				delete engine;
				engine = 0L;
				kdDebug(14150) << "[OscarProtocol] unload(), engine gone" << endl;
		}

		kdDebug(14150) << "[OscarProtocol] unload() END" << endl;
		return KopeteProtocol::unload();
}

// Initializes the actions
void OscarProtocol::initActions(void)
{
		kdDebug(14150) << "[OscarProtocol] initActions() START" << endl;
		actionGoOnline = new KAction ( i18n("Online"), "oscar_online", 0,
						this, SLOT(slotGoOnline()), this, "actionOscarConnect" );
		actionGoOffline = new KAction ( i18n("Offline"), "oscar_offline", 0,
						this, SLOT(slotGoOffline()), this, "actionOscarConnect" );
		actionGoAway = new KAction ( i18n("Away"), "oscar_away", 0,
						this, SLOT(slotGoAway()), this, "actionOscarConnect" );
		actionEditOwnInfo	= KopeteStdAction::contactInfo( this,
						SLOT(slotEditOwnInfo()), this, "actionInfo" );

		actionShowDebug = new KAction( i18n("Show Debug"), "wizard", 0,
						this, SLOT(slotShowDebugDialog()), this, "actionInfo");

		actionStatusMenu = 0L;
//	kdDebug(14150) << "[OscarProtocol] initActions() END" << endl;
}

KActionMenu *OscarProtocol::protocolActions(void)
{
	if ( !actionStatusMenu )
	{
		actionStatusMenu = new KActionMenu( displayName(), this );
		actionStatusMenu->popupMenu()->insertTitle( getSN() );
		actionStatusMenu->insert(actionGoOnline);
		actionStatusMenu->insert(actionGoOffline);
		actionStatusMenu->insert(actionGoAway);
		actionStatusMenu->popupMenu()->insertSeparator();
		actionStatusMenu->insert(actionEditOwnInfo);
		// TODO add action for displaying debug dialog
		actionStatusMenu->popupMenu()->insertSeparator();
		actionStatusMenu->insert(actionShowDebug);
	}

	return actionStatusMenu;
}

void OscarProtocol::initEngine()
{
	kdDebug( 14150 ) << "[OscarProtocol] initEngine() START" << endl;
	QByteArray cook;
	cook.duplicate("01234567",8);
	engine = new OscarSocket( mPrefs->server(), cook );
	kdDebug( 14150 ) << "[OscarProtocol] initEngine() END" << endl;
}

// Called when our status on the server has changed.
// Also called when disconnected
void OscarProtocol::slotStatusChanged(int status)
{
		kdDebug(14150) << "[OscarProtocol] slotStatusChanged(), status="
									 << status << "." << endl;
		switch(status)
		{
		case OSCAR_ONLINE:
		{
				slotOnline();
				// Update our away status
				m_isAway = false;
				break;
		}
		case OSCAR_AWAY:
		{
				setStatusIcon("oscar_away");
				m_isAway = true;
				break;
		}
		case OSCAR_OFFLINE:
		{
				slotDisconnected();
				break;
		}
		}

		// update our own record of our status
		// we need to check that it's valid,
		// because this Contact gets deleted when offline
		if ( myContact )
			myContact->mStatus = status;

}

// Called when we go online
void OscarProtocol::slotOnline(void)
{
		kdDebug(14150) << "[OscarProtocol] slotOnline(), engine says we are online now"<< endl;

		// This should NEVER happen
		if ( myContact )
		{
				kdDebug(14150) << "[OscarProtocol] ERROR, wanted to create myContact but "
											 << "there already is one!" << endl;
		}
		else
		{
				AIMBuddy *buddy = new AIMBuddy(randomNewBuddyNum, 0, getSN());
				randomNewBuddyNum++;
				mBuddyList->addBuddy(buddy);
				myContact = new OscarContact(getSN(), this, 0);
		}
		m_connected = true;
		setStatusIcon("oscar_online");
		kdDebug(14150) << "[OscarProtocol] slotOnline() END" << endl;;
}

// Gets the screen name from the config file
QString OscarProtocol::getSN(void) const
{
		return mPrefs->screenName();
}

// Displays an error dialog with the given text
void OscarProtocol::slotError(QString errmsg, int errorCode)
{
		kdDebug(14150) << "[OscarProtocol] slotError(), errmsg=" <<
				errmsg << ", errorCode=" << errorCode << "." << endl;

		//TODO: somebody add a comment about these two error-types
		if (errorCode == 1 || errorCode == 5)
				slotDisconnected();

		KMessageBox::error(qApp->mainWidget(), errmsg);
}

// Called when we want to go online
void OscarProtocol::slotGoOnline(void)
{
		kdDebug(14150) << "[OscarProtocol] slotGoOnline(); m_connected="
									 << m_connected << endl;

		if ( m_connected )
				setAvailable();
		else
				connect();
}

// Called when we get disconnected
void OscarProtocol::slotDisconnected(void)
{
		kdDebug(14150) << "[OscarProtocol] slotDisconnected() and function info is: " << k_funcinfo << endl;
		setStatusIcon("oscar_offline");
		m_connected = false;
}

// Called when we want to go away
void OscarProtocol::slotGoAway(void)
{
		kdDebug(14150) << "[OscarProtocol] slotGoAway()" << endl;
		if ( m_connected )
		{
				// Show the dialog, which takes care of the
				// setting of the away message
				oscarAwayDialog->show();
		}
}

// Called when we want to go offline
void OscarProtocol::slotGoOffline(void)
{
		kdDebug(14150) << "[OscarProtocol] slotGoOffline()" << endl;
		if ( m_connected )
				disconnect();

		// We want to at least look like we're offline, even if we weren't online
		slotDisconnected();
}

// Called when we want to show the debug dialog
void OscarProtocol::slotShowDebugDialog(){
		if(mDebugDialog){
				mDebugDialog->show();
		}
}

// Called when we have gotten an IM
void OscarProtocol::slotGotIM(QString, QString sender, bool)
{
	kdDebug(14150) << "[OscarProtocol] slotGotIM(); got a buddy for the list, sender=" << sender << endl;
	//basically, all this does is add the person to your buddy list
	if (!mBuddyList->findBuddy(sender))
		addNewContact(sender, QString(), false, i18n("Unknown"));
}

// Called when we retrieve the buddy list from the server
void OscarProtocol::slotGotConfig(AIMBuddyList &conf)
{
	kdDebug(14150) << "[OscarProtocol] slotGotConfig()" << endl;
	//save server side contact list
	*mBuddyList += conf;
	QValueList<AIMBuddy * > buddyList = conf.buddies().values();
	for ( QValueList<AIMBuddy * >::Iterator it = buddyList.begin(); it != buddyList.end(); ++it)
	{
		if ((*it))
		{
			kdDebug(14150) << "[OscarProtocol] Calling addOldContact on " << (*it)->screenname() << "." << endl;
			addOldContact((*it));
		}
	}
}

// See aimbuddylist.h under 'signals:' for an explination of this
void OscarProtocol::slotReTryServerContacts(AIMGroup *group)
{
	kdDebug(14150) << "[OscarProtocol] slotReTryServerContacts iteration... mGroupQueue.count() is " << mGroupQueue.count() << endl;
	// Process the queue
	int i = 0;
	for (AIMBuddy *it = mGroupQueue.at(i); it != 0L; it = mGroupQueue.at(i))
	{
		if (!it)
		{
			i++;
			continue;
		}
		if (mBuddyList->findGroup(it->groupID())) // Success, group now exists, add contact
		{
			mGroupQueue.remove(i);
			addOldContact(it);
		}
		i++;
	}
}

// Adds a contact that we already know about to the list
void OscarProtocol::addOldContact(AIMBuddy *bud,KopeteMetaContact *meta)
{
		bool temporary = false;
		AIMGroup *group = mBuddyList->findGroup(bud->groupID());
		if (!group && bud)
		{
			kdDebug(14150) << "[OscarProtocol] Adding AIMBuddy 'bug' to the groupQueue to check later once we know about the group." << endl;
			mGroupQueue.append(bud);
			return;
		}

		mBuddyList->addBuddy(bud);
		if ( !mBuddyList->findBuddy(bud->screenname()) ) return;

		if (group->name().isNull())
			temporary = true;

		kdDebug(14150) << "[OscarPortocol] addOldContact(); groupName is " << group->name() << endl;

		KopeteMetaContact *m = KopeteContactList::contactList()->findContact( pluginId(), QString::null, bud->screenname() );

		kdDebug(14150) << "[OscarProtocol] addOldContact(), KopeteMetaContact m=" << m << endl;

		if( m )
		{
				// Existing contact, update data that is ALREADY in the LOCAL database for AIM
				if (m->isTemporary())
						m->setTemporary(false);
		}
		else
		{
				// New contact
				kdDebug(14150) << "[OscarProtocol] Adding old contact " << bud->screenname() << " ..." << endl;

				if (meta)
						m=meta;
				else
				{
					m=new KopeteMetaContact();
					if(!temporary)
						m->addToGroup(KopeteContactList::contactList()->getGroup(group->name()));
				}

				if (temporary)
					m->setTemporary(true);

				new OscarContact( bud->screenname(), this, m );
				if (!meta)
					KopeteContactList::contactList()->addMetaContact(m);
		}
}

// Returns pointer to the logged in user's info
KopeteContact *OscarProtocol::myself(void) const
{
		kdDebug(14150) << "OscarProtocol::myself with displayname= " <<
			( (myContact == 0L) ? QString().fromLatin1("NOT INITIALISED") : myContact->displayName() ) << endl;
		return myContact;
}

// This will be called when Kopete reads the contact list
void OscarProtocol::deserializeContact( KopeteMetaContact *metaContact, const QMap<QString, QString> &serializedData,
				const QMap<QString, QString> & /* addressBookData */ )
{
		QString contactId   = serializedData[ "contactId" ];
		QString displayName = serializedData[ "displayName" ];
		addNewContact( contactId, displayName, !displayName.isEmpty(), metaContact->groups().first()->displayName(), metaContact );
}

// Called when a group is added
void OscarProtocol::slotGroupAdded(KopeteGroup *group)
{
		QString groupName=group->displayName();
		AIMGroup *aGroup = mBuddyList->findGroup(groupName);
		if (!aGroup)
		{
			aGroup = mBuddyList->addGroup(randomNewGroupNum, groupName);
			randomNewGroupNum++;
		} else return;
		kdDebug(14150) << "[OscarProtocol] addGroup() being called" << endl;
		if (m_connected)
			engine->sendAddGroup(groupName);
}

// Called when we want to return the active instance of the protocol
OscarProtocol *OscarProtocol::protocol(void)
{
		return protocolStatic_;
}

// Called when we have been warned
void OscarProtocol::slotGotWarning(int newlevel, QString warner)
{
		if (userInfo.evil < newlevel) //this is not a natural decrease in level
		{
				QString message = i18n("You have been warned %1. Your new warning level is %2%.").arg(
								(warner.isNull()) ? i18n("anonymously") : i18n("...warned by...", "by %1").arg(warner)).arg(newlevel);
				KMessageBox::sorry(0L,message);
		}
		userInfo.evil = newlevel;
}

// Called when we have received my own user info
void OscarProtocol::slotGotMyUserInfo(UserInfo u)
{
		userInfo = u;
}

// Called when we want to edit our own profile information
void OscarProtocol::slotEditOwnInfo(void)
{
		OscarUserInfo *oscaruserinfo =
				new OscarUserInfo(getSN(), getSN(), this, engine->getMyProfile());
		oscaruserinfo->exec(); // This is a modal dialog
		// FIXED: This was a double delete, AFAIK
		//delete oscaruserinfo;

}

// Saves the user profile in the preferences
void OscarProtocol::setUserProfile( QString prof )
{
		engine->setMyProfile(prof);
		mPrefs->setProfile(prof);
}

/** Called when we get a request for a direct IM session with @sn */
void OscarProtocol::slotGotDirectIMRequest(QString sn)
{
		QString title = i18n("Direct IM session request");
		QString message = i18n("%1 has requested a direct IM session with you.  Direct IM sessions allow the remote user to see your IP address, which can lead to security problems if you don't trust him/her.  Do you want to establish a direct connection to %2?").arg(sn).arg(sn);
		int result = KMessageBox::questionYesNo(qApp->mainWidget(), message, title);
		if (result == KMessageBox::Yes)
		{
				engine->sendDirectIMAccept(sn);
		}
		else if (result == KMessageBox::No)
		{
				engine->sendDirectIMDeny(sn);
		}
}

/** Called when there is no activity for a certain amount of time  */
void OscarProtocol::slotIdleTimeout(void)
{
		kdDebug(14150) << k_funcinfo << "got an idle timeout, setting idle time with server" << endl;
		//idleTimeout() gives a value in minutes, engine wants seconds
		engine->sendIdleTime( mPrefs->idleTimeout()*60 );
}

/** Called when there is mouse/keyboard activity */
void OscarProtocol::slotIdleActivity(void)
{
		kdDebug(14150) << k_funcinfo << "got some activity, setting idle time with server to 0" << endl;
		engine->sendIdleTime( 0 );
}

/** Called when preferences are saved */
void OscarProtocol::slotPreferencesSaved(void)
{
		kdDebug(14150) << k_funcinfo << "preferences were saved" << endl;
		// Set idle time timeout if it is in the config
		mIdleMgr.setTimeout( mPrefs->idleTimeout()*60 );

		// Start monitoring idle time if config tells us to
		if ( mPrefs->reportIdle() )
				mIdleMgr.start();
		else
				mIdleMgr.stop();
}

#include "oscarprotocol.moc"

/*
 * Local variables:
 * c-indentation-style: k&r
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * End:
 */

// vim: set noet ts=4 sts=4 sw=4:

