/*
	oscarprotocol.h  -  Oscar Protocol Plugin

	Copyright (c) 2002 by Tom Linsky <twl6@po.cwru.edu>

	Kopete    (c) 2002 by the Kopete developers  <kopete-devel@kde.org>

	*************************************************************************
	*                                                                       *
	* This program is free software; you can redistribute it and/or modify  *
	* it under the terms of the GNU General Public License as published by  *
	* the Free Software Foundation; either version 2 of the License, or     *
	* (at your option) any later version.                                   *
	*                                                                       *
	*************************************************************************
	*/

#ifndef OSCARPROTOCOL_H
#define OSCARPROTOCOL_H


#include <qwidget.h>
#include <qmovie.h>
#include <kaction.h>
#include <kpopupmenu.h>
#include <kopeteprotocol.h>
#include "kopetestdaction.h"
#include "xautolock.h"
#include "oscarsocket.h"

#define TAIM_OFFLINE		0
#define TAIM_ONLINE		1
#define TAIM_AWAY			2

class AIMBuddy;
class AIMGroup;
class AIMBuddyList;

/**
 * Oscar protocol plugin for Kopete!
 * @author Tom Linsky <twl6@paranoia.STUDENT.CWRU.Edu>
 */

class OscarContact;
class KopeteMetaContact;
class KopeteGroup;
class OscarPreferences;
class OscarChangeStatus;
class OscarDebugDialog;

class AIMBuddyList;
class AIMBuddy;

class OscarProtocol : public KopeteProtocol
{
		Q_OBJECT

public: 
		OscarProtocol(QObject *parent, const char *name,
						const QStringList &args);
		~OscarProtocol();
		/** Tells whether we are connected */
		virtual bool isConnected() const;
		/** Returns true if we are away, occupied, etc */
		virtual bool isAway(void) const;
		/**
		 * Called when kopete wants to set our status
		 * to online or similar
		 */
		virtual void setAvailable(void);
		/**
		 * This will be called when kopete-main wants
		 * to set the status to away
		 */
		virtual void setAway(void);
		/** Creates the add contact page */
		virtual AddContactPage *createAddContactWidget(QWidget *parent);
		/** Adds the given contact to the protocol's list */
		void addNewContact(const QString &sn,
						const QString &alias,	bool hasAlias,
						const QString &group, KopeteMetaContact *meta=0l);
		/** Called when plugin unloads */
		virtual bool unload(void);
		/** Called on plugin init */
		virtual void init(void);
		/** Returns pointer to the logged in user's info */
		virtual KopeteContact *myself(void) const;

		/**
		 * restores the contact list
		 */
		virtual void deserializeContact( KopeteMetaContact *metaContact,
						const QMap<QString, QString> &serializedData, const QMap<QString, QString> &addressBookData );

		/**
		 * Initializes the engine.  Write a class that inherits
		 * ProtocolSocket, inherit this class, and the only
		 * function you have to re-implement is this one
		 */
		virtual void initEngine(void);
		/**
		 * Adds a contact that we already know about to
		 * the list
		 */
		void addOldContact(AIMBuddy *bud, KopeteMetaContact *meta=0l);
		/** Returns a pointer to the buddy list */
		AIMBuddyList *buddyList() { return mBuddyList; };
		/** Returns the permit/deny setting */
		int permitStatus() { return 0/*FIXME: buddies.permitStatus*/; };
		/** Returns the active instance of the protocol */
		static OscarProtocol *protocol(void);
		/* Protocol menu shown from the status bar icon */
		KActionMenu *protocolActions(void);
		/** Saves the user profile in the preferences and sends it to the server */
		void setUserProfile( QString prof );

		virtual const QString protocolIcon();

		/** The engine used to communicate with the server */
		OscarSocket *engine;

		KAction* actionGoOnline;
		KAction* actionGoOffline;
		KAction* actionGoAway;
		KAction * actionEditOwnInfo; // Edits the user's profile
		KAction *actionShowDebug;

		KActionMenu* actionStatusMenu;

		int randomNewBuddyNum;
		int randomNewGroupNum;

private:
		/** Tells whether we are connected */
		bool m_connected;
		/** Tells if we are away or not */
		bool m_isAway;
		/** MY contact info */
		OscarContact *myContact;
		AIMBuddyList *mBuddyList;
		/** Our own screen name */
//	QString mName;
		/** Preferences dialog */
		OscarPreferences * mPrefs;
		/** The active instance of oscarprotocol */
		static OscarProtocol *protocolStatic_;
		/** The currently logged in user's information */
		UserInfo userInfo;
		/** Away Dialog */
		OscarChangeStatus *oscarAwayDialog;
		/** Debug Dialog */
		OscarDebugDialog *mDebugDialog;
		/** Idleness manager */
		XAutoLock mIdleMgr;

		private:
		/** Initializes the icons */
		void initIcons(void);
		/** Initializes the actions */
		void initActions(void);
		/** Gets the screen name from the config file */
		QString getSN(void) const;

		// See aimbuddylist.h under 'signals:' for an explination of this. This is 'the queue'
		QPtrList<AIMBuddy> mGroupQueue;

public slots:
		/** Displays an error dialog with the given text */
		void slotError(QString errmsg, int errorCode);
		/** Called when we want to go offline */
		void slotGoOffline(void);
		/** Called when we have gotten an IM */
		void slotGotIM(QString, QString, bool);
		/** Called when we retrieve the buddy list from the server */
		void slotGotConfig(AIMBuddyList &);

		/** Show the debug dialog */
		void slotShowDebugDialog();

		/** Disconnects from the server */
		virtual void disconnect();
		/** Connects to the server and logs in */
		virtual void connect();

		// See aimbuddylist.h under 'signals:' for an explination of this
		void slotReTryServerContacts(AIMGroup *);

private slots:
		/** Called when we want to go away */
		void slotGoAway(void);
		/** Called when we go online */
		void slotOnline(void);
		/** Called when we want to go online */
		void slotGoOnline(void);
		/** Called when we get disconnected */
		void slotDisconnected(void);
		/** Called when a group is added */
		void slotGroupAdded(KopeteGroup*);
		/** Called when our status on the server has changed
		 * This is called when disconnected too
		 */
		void slotStatusChanged(int);
		/** Called when we have been warned */
		void slotGotWarning(int newlevel, QString warner);
		/** Called when we have our own info */
		void slotGotMyUserInfo(UserInfo u);
		/** Called when we want to edit our own
		 * profile information
		 */
		void slotEditOwnInfo(void);
		/** Called when we get a request for a direct IM session with @sn */
		void slotGotDirectIMRequest(QString sn);
		/** Called when there is no activity for a certain amount of time  */
		void slotIdleTimeout(void);
		/** Called when there is mouse/keyboard activity */
		void slotIdleActivity(void);
		/** Called when preferences are saved */
		void slotPreferencesSaved(void);
};

#endif
/*
 * Local variables:
 * mode: c++
 * c-indentation-style: k&r
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * End:
 */
